with Messages, Ada.Integer_Text_IO,Ada.Numerics.Float_Random;
use Messages, Ada.Integer_Text_IO,Ada.Numerics.Float_Random;


package body Fou2_Ops is
   ScalZero: constant Scalar := Scal(Zero);
   ScalOne: constant Scalar := Scal(One);
   ScalTwo: constant Scalar := Scal(Two);
   ScalHalf: constant Scalar := Scal(Half);
   InvGold: constant Scalar := Golden-ScalOne;
   NewCutoff: constant Rep:= Rep(0.01);
   Trunc: constant Boolean := IsNumeric(ScalZero);


   procedure CopySign(H1: in Fourier2; H2: in out Fourier2) is
   begin
      CopySign(H1.COS,H2.COS);
      CopySign(H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end CopySign;


   procedure NProj(H1: in Fourier2; H2: out Fourier2) is
   begin
      EvenNProj(H1.COS,H2.COS);
      OddNProj(H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end NProj;


   procedure NProj(H1: in out Fourier2) is
      H2: Fourier2;
   begin
      EvenNProj(H1.COS,H2.COS);
      OddNProj(H1.SIN,H2.SIN);
      H2.R:=H1.R;
      SetZero(H1.R,H1);
      Copy(H2,H1);
   end NProj;


   procedure MProj(H1: in out Fourier2) is
      H2: Fourier2;
   begin
      EvenMProj(H1.COS,H2.COS);
      OddMProj(H1.SIN,H2.SIN);
      H2.R:=H1.R;
      SetZero(H1.R,H1);
      Copy(H2,H1);
   end MProj;


   procedure MNProj(H1: in out Fourier2) is
      H2: Fourier2;
   begin
      EvenMNProj(H1.COS,H2.COS);
      OddMNProj(H1.SIN,H2.SIN);
      H2.R:=H1.R;
      SetZero(H1.R,H1);
      Copy(H2,H1);
   end MNProj;


   function Norm(H: Fourier2) return Error is
   begin
      return Norm(H.COS)+Norm(H.SIN);
   end Norm;


   function SupNorm(H: Fourier2) return Error is
   begin
      return SupNorm(H.COS)+SupNorm(H.SIN);
   end SupNorm;


   function Norm(F: Scalar_Array; R3: Radius; H: Fourier2) return Error is
   begin
      return Norm(F,R3,H.COS)+Norm(F,R3,H.SIN);
   end Norm;


   procedure ShowLogNorms(H: in Fourier2) is
   begin
      Put("H.COS:"); New_Line;
      ShowLogNorms(H.COS);
      Put("H.SIN:"); New_Line;
      ShowLogNorms(H.SIN);
   end ShowLogNorms;


   function Norm(R: Weights; H: Fourier2) return Error is
   begin
      return Norm(R,H.COS)+Norm(R,H.SIN);
   end Norm;


   function SupNorm(R: Weights; H: Fourier2) return Error is
   begin
      return SupNorm(R,H.COS)+SupNorm(R,H.SIN);
   end SupNorm;


   function ShortNorm(R: Weights; H: Fourier2) return Error is
      HH: Fourier2;
   begin
      Copy(H,HH);
      SetComponent(0,0,1,Scal(0.0),HH.COS);
      SetComponent(0,0,3,Scal(0.0),HH.COS);
      return Norm(R,HH.COS)+Norm(R,HH.SIN);
   end ShortNorm;


   procedure Write(F: in File_Type;H: in Fourier2) is
   begin
      Put(F,0,H.COS);
      Put(F,1,H.SIN);
   end Write;


   procedure PiTranslateX(H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      PiTranslateX(H1.COS,H2.COS);
      PiTranslateX(H1.SIN,H2.SIN);
   end PiTranslateX;


   procedure PiTranslateY(H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      PiTranslateY(H1.COS,H2.COS);
      PiTranslateY(H1.SIN,H2.SIN);
   end PiTranslateY;


   procedure PiOver2TranslateY(H1: in out Fourier2) is
      C,S: Taylor;
      H2: Fourier2;
   begin
      SetZero(H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            C:=Component(0,M,N,H1);
            if N  mod 2=0 then
               if N mod 4=0 then
                  AddComponent(0,M,N,C,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*C,H2);
               end if;
            else
               if (N-1) mod 4=0 then
                  AddComponent(1,M,N,C,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*C,H2);
               end if;
            end if;
         end loop;
         for N in Nmin(M,1) .. Nmax(M) loop
            S:=Scal(-1.0)*Component(1,M,N,H1);
            if N  mod 2=0 then
               if N mod 4=0 then
                  AddComponent(1,M,N,S,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*S,H2);
               end if;
            else
               if (N-1) mod 4=0 then
                  AddComponent(0,M,N,S,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*S,H2);
               end if;
            end if;
         end loop;
      end loop;
      H1:=H2;
   end PiOver2TranslateY;


   procedure PiOver2TranslateX(H1: in out Fourier2) is
      C,S: Taylor;
      H2: Fourier2;
   begin
      SetZero(H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            C:=Component(0,M,N,H1);
            if M  mod 2=0 then
               if M mod 4=0 then
                  AddComponent(0,M,N,C,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*C,H2);
               end if;
            else
               if (M-1) mod 4=0 then
                  AddComponent(1,M,N,C,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*C,H2);
               end if;
            end if;
         end loop;
         for N in Nmin(M,1) .. Nmax(M) loop
            S:=Scal(-1.0)*Component(1,M,N,H1);
            if M  mod 2=0 then
               if M mod 4=0 then
                  AddComponent(1,M,N,S,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*S,H2);
               end if;
            else
               if (M-1) mod 4=0 then
                  AddComponent(0,M,N,S,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*S,H2);
               end if;
            end if;
         end loop;
      end loop;
      H1:=H2;
   end PiOver2TranslateX;


   procedure MinusPiOver2TranslateX(H1: in out Fourier2) is
      C,S: Taylor;
      H2: Fourier2;
   begin
      SetZero(H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            C:=Scal(-1.0)*Component(0,M,N,H1);
            if M  mod 2=0 then
               if M mod 4=0 then
                  AddComponent(0,M,N,C,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*C,H2);
               end if;
            else
               if (M-1) mod 4=0 then
                  AddComponent(1,M,N,C,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*C,H2);
               end if;
            end if;
         end loop;
         for N in Nmin(M,1) .. Nmax(M) loop
            S:=Component(1,M,N,H1);
            if M  mod 2=0 then
               if M mod 4=0 then
                  AddComponent(1,M,N,S,H2);
               else
                  AddComponent(1,M,N,Scal(-1.0)*S,H2);
               end if;
            else
               if (M-1) mod 4=0 then
                  AddComponent(0,M,N,S,H2);
               else
                  AddComponent(0,M,N,Scal(-1.0)*S,H2);
               end if;
            end if;
         end loop;
      end loop;
      H1:=H2;
   end MinusPiOver2TranslateX;


   procedure Show(H: in Fourier2; Cut: in Rep := Zero) is
   begin
      Show(H.COS,Cut);
      Show(H.SIN,Cut);
      ShowR(Rho(H));
   end Show;


   procedure Cshow(F: in File_Type; H: in Fourier2) is
   begin
      Put(F,"r=0.0;");
      New_Line(F);
      Cshow(F,0,H.COS);
      Cshow(F,1,H.SIN);
   end Cshow;


   procedure SetRho(R: in Weights; H: in out Fourier2) is
   begin
      H.R := R;
   end SetRho;


   function Deg(J: Parity; M: Freq1; N: Freq2; H: Fourier2) return Integer is
   begin
      if J=0 then
         return Deg(M,N,H.COS);
      else
         return Deg(M,N,H.SIN);
      end if;
   end Deg;


   procedure CutDeg(R: in Weights; Cut: in Rep; H: in out Fourier2) is
   begin
      if Cut>Zero then
         declare
            F: Scalar_Array;
            Pc: Taylor;
         begin
            if not Trunc then
               Message("Fou2_Ops.CutDeg",Numeric_Only);
            end if;
            NormFactors(R,F);
            for M in Freq1 loop
               for N in Nmin(M,0) .. Nmax(M) loop
                  Pc:=Component(0,M,N,H);
                  CutDeg(R(3),Cut/Center(F(M,N)),Pc);
                  SetComponent(0,M,N,Pc,H);
               end loop;

               for N in Nmin(M,1) .. Nmax(M) loop
                  Pc:=Component(1,M,N,H);
                  CutDeg(R(3),Cut/Center(F(M,N)),Pc);
                  SetComponent(1,M,N,Pc,H);
               end loop;

            end loop;
         end;
      end if;
   end CutDeg;


   procedure NumValQ0(H: in Fourier2; P: out Taylor) is
   begin
      SetZero(P);
      NumValQ0(H.COS,P);
   end NumValQ0;


   function EffDeg(J: Parity; M,N: integer; H: Fourier2) return Integer is
   begin
      if J=0 then
         return EffDeg(M,N,H.COS);
      else
         return EffDeg(M,N,H.SIN);
      end if;
   end EffDeg;


   procedure Get(F: in File_Type; R: in Weights; H: out Fourier2) is
   begin
      Get(F,0,R,H.COS);
      Get(F,1,R,H.SIN);
      H.R:=R;
   end Get;


   procedure GetEven(F: in File_Type; R: in Weights; H: out Fourier2) is
   begin
      Get(F,0,R,H.COS);
      SetZero(1,R,H.SIN);
      H.R:=R;
   end GetEven;


   procedure Put(F: in File_Type; R: in Weights; H: in Fourier2) is
   begin
      Put(F,0,H.COS);
      Put(F,1,H.SIN);
   end Put;

   procedure Read(Name: in String; R: in Weights; H: out Fourier2) is
      F: File_Type;
   begin
      Open(F,In_File,Name);
      Get(F,R,H);
      Close(F);

   exception
      when Name_Error =>
         XE:=1;
         return;
         raise Messages.Leaving;
   end Read;


   procedure ReadEven(Name: in String; R: in Weights; H: out Fourier2) is
      F: File_Type;
   begin
      Open(F,In_File,Name);
      Get(F,0,R,H.COS);
      SetZero(1,R,H.SIN);
      H.R:=R;
      Close(F);
   end ReadEven;


   function EffDeg(M,N: integer; H: Fourier2) return Integer is
   begin
      return Max(EffDeg(M,N,H.COS),EffDeg(M,N,H.SIN));
   end EffDeg;


   procedure EffDeg(J: in Parity; M,N: in Integer; H: in Fourier2; Dk: out Integer) is
   begin
      if J=0 then
         Dk:=EffDeg(M,N,H.COS);
      else
         Dk:=EffDeg(M,N,H.SIN);
      end if;
   end EffDeg;


   procedure NumDqDot(R: in Weights; S1,S2: in Scalar; H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      NumDqDot(R,S1,S2,H1.COS,H2.SIN);
      NumDqDot(R,S1,S2,H1.SIN,H2.COS);
   end NumDqDot;


   procedure NumDer2(R2: in Radius; H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      NumDer2(R2,H1.COS,H2.SIN);
      NumDer2(R2,H1.SIN,H2.COS);
   end NumDer2;


   procedure NumDer1(R1: in Radius; H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      NumDer1(R1,H1.COS,H2.SIN);
      NumDer1(R1,H1.SIN,H2.COS);
   end NumDer1;


   function NumDqDot(R: in Weights; S1,S2: in Scalar; H1: in Fourier2) return Fourier2 is
      H2: Fourier2;
   begin
      SetZero(H1.R,H2);
      NumDqDot(R,S1,S2,H1,H2);
      return H2;
   end NumDqDot;


   procedure EffDeg(M,N: in Integer; H: in Fourier2; Dk: out Integer) is
      Dks,Dkc: Integer;
   begin
      Dkc:=EffDeg(M,N,H.COS);
      Dks:=EffDeg(M,N,H.SIN);
      Dk:=Max(Dkc,Dks);
   end EffDeg;


   procedure EffDeg(J: in Parity; M: in Integer; H: in Fourier2; Dn,Dk: out Integer) is
   begin
      if J=0 then
         EffDeg(M,H.COS,Dn,Dk);
      else
         EffDeg(M,H.SIN,Dn,Dk);
      end if;
   end EffDeg;


   function Coeff(J: Parity; M,N: Integer; K: Natural; H: Fourier2) return Scalar is
   begin
      if J=0 then
         return Coeff(M,N,K,H.COS);
      else
         return Coeff(M,N,K,H.SIN);
      end if;
   end Coeff;


   function Coeff(J: Parity; M,N: Integer; H: Fourier2) return Taylor is
   begin
      if J=0 then
         return Coeff(M,N,H.COS);
      else
         return Coeff(M,N,H.SIN);
      end if;
   end Coeff;


   procedure EffDeg(M: in Integer; H: in Fourier2; Dn,Dk: out Integer) is
      Dkc,Dks,Dnc,Dns: Integer;
   begin
      EffDeg(M,H.COS,Dnc,Dkc);
      EffDeg(M,H.SIN,Dns,Dks);
      Dk:=Max(Dkc,Dks);
      Dn:=Max(Dnc,Dns);
   end EffDeg;


   procedure EffDeg(J: in Parity; H: in Fourier2; Dm,Dn,Dk: out Integer) is
   begin
      if J=0 then
         EffDeg(H.COS,Dm,Dn,Dk);
      else
         EffDeg(H.SIN,Dm,Dn,Dk);
      end if;
   end EffDeg;


   procedure EffDeg(H: in Fourier2; Dm,Dn,Dk: out Integer) is
      Dmc,Dms,Dkc,Dks,Dnc,Dns: Integer;
   begin
      EffDeg(H.COS,Dmc,Dnc,Dkc);
      EffDeg(H.SIN,Dms,Dns,Dks);
      Dm:=Max(Dmc,Dms);
      Dk:=Max(Dkc,Dks);
      Dn:=Max(Dnc,Dns);
   end EffDeg;


   procedure SetZero(R: in Weights; H: out Fourier2) is
   begin
      SetZero(0,R,H.COS);
      SetZero(1,R,H.SIN);
      H.R:=R;
   end SetZero;


   -- We do not need 'zeroaverage' the sinus terms
   procedure ZeroAverage(H: in out Fourier2) is
   begin
      ZeroAverage(H.COS);
   end ZeroAverage;


   procedure Iminus(L: in Cutoffs; H1: in Fourier2; H2: out Fourier2; Collect: in Boolean := True) is
   begin
      Iminus(L,H1.COS,H2.COS,Collect);
      Iminus(L,H1.SIN,H2.SIN,Collect);
      H2.R:=H1.R;
   end Iminus;


   procedure Iminus(L: in Cutoffs; H: in out Fourier2) is
   begin
      Iminus(L,H.COS);
      Iminus(L,H.SIN);
   end Iminus;


   function Iminus(L: Cutoffs; H: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(H.R,Hr);
      Iminus(L,H,Hr);
      return Hr;
   end Iminus;


   procedure Iplus(L: in Cutoffs; H1: in Fourier2; H2: out Fourier2; Collect: in Boolean := True) is
   begin
      Iplus(L,H1.COS,H2.COS,Collect);
      Iplus(L,H1.SIN,H2.SIN,Collect);
      H2.R:=H1.R;
   end Iplus;


   procedure Iplus(L: in Cutoffs; H: in out Fourier2) is
   begin
      Iplus(L,H.COS);
      Iplus(L,H.SIN);
   end Iplus;


   function Iplus(L: Cutoffs; H: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(H.R,Hr);
      Iplus(L,H,Hr);
      return Hr;
   end Iplus;


   -- Basic procedures and functions
   procedure Copy(H1: in Fourier2; H2: out Fourier2) is
   begin
      Copy(H1.COS,H2.COS);
      Copy(H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end Copy;


   procedure Eliminate(H1: in Fourier2; H2: out Fourier2) is
   begin
      EvenEliminate(H1.COS,H2.COS);
      OddEliminate(H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end Eliminate;


   procedure Neg(H1: in Fourier2; H2: out Fourier2) is
   begin
      Neg(H1.COS,H2.COS);
      Neg(H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end Neg;


   function "-"(H: Fourier2) return Fourier2 is
      Ht: Fourier2;
   begin
      SetZero(H.R,Ht);
      Neg(H,Ht);
      return Ht;
   end "-";


   procedure Mult(S: in Scalar; H: in out Fourier2) is
   begin
      Mult(S,H.COS);
      Mult(S,H.SIN);
   end Mult;


   procedure ShortMult(S: in Scalar; H: in out Fourier2) is
      S1: Scalar;
      S3: Scalar;
   begin
      S1:=Component(0,0,0,1,H);
      S3:=Component(0,0,0,3,H);
      Mult(S,H.COS);
      Mult(S,H.SIN);
      SetComponent(0,0,0,1,S1,H);
      SetComponent(0,0,0,3,S3,H);
   end ShortMult;


   procedure ShortMult(S: in Scalar; H1: in Fourier2; H2: out Fourier2) is
      S1: Scalar;
      S3: Scalar;
   begin
      SetZero(H1.R,H2);
      S1:=Component(0,0,0,1,H1);
      S3:=Component(0,0,0,3,H1);
      Copy(H1,H2);
      Mult(S,H2.COS);
      Mult(S,H2.SIN);
      SetComponent(0,0,0,1,S1,H2);
      SetComponent(0,0,0,3,S3,H2);
   end ShortMult;


   procedure Prod(S: in Scalar; H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      Prod(S,H1.COS,H2.COS);
      Prod(S,H1.SIN,H2.SIN);
      H2.R:=H1.R;
   end Prod;


   procedure Prod(H1: in Monom; H2: in Fourier2; H3: out Fourier2; D: in Degrees := Dmax) is
   begin
      SetZero(H2.R,H3);
      if GetParity(H1)=0
      then
         Prod(H1,H2.COS,H3.COS);
         Prod(H1,H2.SIN,H3.SIN);
      else
         if GetParity(H1)=1
         then
            Prod(H1,H2.COS,H3.SIN);
            Prod(H1,H2.SIN,H3.COS);
         end if;
      end if;
   end Prod;


   procedure MultAdd(H1: in Monom; H2: in Fourier2; H3: in out Fourier2; D: in Degrees := Dmax) is
   begin
      if (GetParity(H1)=0)
      then
         MultAdd(H1,H2.COS,H3.COS);
         MultAdd(H1,H2.SIN,H3.SIN);
      else
         MultAdd(H1,H2.COS,H3.SIN);
         MultAdd(H1,H2.SIN,H3.COS);
      end if;
   end MultAdd;


   procedure MultAdd(J: in Parity; S: in Scalar; M,N: in Integer; P: in Taylor; H2: in Fourier2; H3: in out Fourier2; D: in Degrees := Dmax) is
   begin
      if Deg(P)>=0 then
         if (J=0) then
            MultAdd(S,M,N,P,H2.COS,H3.COS);
            MultAdd(S,M,N,P,H2.SIN,H3.SIN);
         else
            MultAdd(S,M,N,P,H2.COS,H3.SIN);
            MultAdd(S,M,N,P,H2.SIN,H3.COS);
         end if;
      end if;
   end MultAdd;


   function "*"(S: Scalar; H: Fourier2) return Fourier2 is
      Ht: Fourier2;
   begin
      SetZero(H.R,Ht);
      Prod(S,H,Ht);
      return Ht;
   end "*";


   procedure Add(H1: in Fourier2; H2: in out Fourier2) is
   begin
      Add(H1.COS,H2.COS);
      Add(H1.SIN,H2.SIN);
      H2.R := Min(H1.R,H2.R);
   end Add;


   procedure Sum(H1,H2: in Fourier2; H3: out Fourier2) is
   begin
      Sum(H1.COS,H2.COS,H3.COS);
      Sum(H1.SIN,H2.SIN,H3.SIN);
      H3.R := Min(H1.R,H2.R);
   end Sum;


   function "+"(H1: Fourier2; H2: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(Min(H1.R,H2.R),Hr);
      Sum(H1,H2,Hr);
      return Hr;
   end "+";


   procedure Sub(H1: in Fourier2; H2: in out Fourier2) is
   begin
      Sub(H1.COS,H2.COS);
      Sub(H1.SIN,H2.SIN);
   end Sub;


   procedure Diff(H1,H2: in Fourier2; H3: out Fourier2) is
   begin
      SetZero(Min(H1.R,H2.R),H3);
      Diff(H1.COS,H2.COS,H3.COS);
      Diff(H1.SIN,H2.SIN,H3.SIN);
   end Diff;


   function "-"(H1,H2: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(Min(H1.R,H2.R),Hr);
      Diff(H1,H2,Hr);
      return Hr;
   end "-";


   procedure MultAdd(S: in Scalar; H1: in Fourier2; H2: in out Fourier2) is
   begin
      MultAdd(S,H1.COS,H2.COS);
      MultAdd(S,H1.SIN,H2.SIN);
   end MultAdd;


   procedure MultAdd(R: in Weights; P: in Taylor; H1: in Fourier2; H2: in out Fourier2; D: in Degrees := Dmax) is
   begin
      MultAdd(R,P,H1.COS,H2.COS,D);
      MultAdd(R,P,H1.SIN,H2.SIN,D);
   end MultAdd;


   procedure Prod(H1,H2: in Fourier2; H3: out Fourier2; D: in Degrees :=Dmax) is
      F1: Fourier;
      F2: Fourier;
   begin
      SetZero(0,Min(H1.R,H2.R),F1);
      SetZero(1,Min(H1.R,H2.R),F2);

      SetZero(Min(Rho(H1.COS),Rho(H2.COS)),H3);
      Prod(Min(Rho(H1.COS),Rho(H2.COS)),H1.COS,H2.COS,H3.COS,D);
      Prod(Min(Rho(H1.SIN),Rho(H2.SIN)),H1.SIN,H2.SIN,F1,D);
      Add(0,F1,H3.COS);
      Prod(Min(Rho(H1.COS),Rho(H2.SIN)),H1.COS,H2.SIN,H3.SIN,D);
      Prod(Min(Rho(H1.SIN),Rho(H2.COS)),H1.SIN,H2.COS,F2,D);
      Add(1,F2,H3.SIN);
      H3.R := Min(H1.R,H2.R);
   end Prod;


   procedure Prod(R: in Weights; H1,H2: in Fourier2; H3: out Fourier2; D: in Degrees := Dmax) is
      G1: Fourier;
      G2: Fourier;
   begin
      SetZero(0,R,G1);
      SetZero(1,R,G2);
      SetZero(R,H3);
      Prod(R,H1.COS,H2.COS,H3.COS);
      Prod(R,H1.SIN,H2.SIN,G1);
      Add(G1,H3.COS);
      Prod(R,H1.COS,H2.SIN,H3.SIN);
      Prod(R,H1.SIN,H2.COS,G2);
      Add(G2,H3.SIN);
      H3.R:=R;
   end Prod;


   function Max(S1: Integer; S2: Integer) return Integer is
   begin
      if S1 > S2 then return S1;
      else return S2;
      end if;
   end Max;


   function "*"(H1,H2: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(Min(H1.R,H2.R),Hr);
      Prod(H1,H2,Hr);
      return Hr;
   end "*";


   procedure Scale(Z,S: in Scalar; H1: in Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      Scale(Z,S,H1.COS,H2.COS);
      Scale(Z,S,H1.SIN,H2.SIN);
   end Scale;


   function Scale(Z,S: Scalar; H: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(H.R,Hr);
      Scale(Z,S,H,Hr);
      return Hr;
   end Scale;


   procedure Der3(R3: in Radius; H1: in Fourier2; H2: out Fourier2) is
   begin
      Der3(R3,H1.COS,H2.COS);
      Der3(R3,H1.SIN,H2.SIN);
      H2.R := H1.R;
      H2.R(3) := R3;
   end Der3;


   function Der3(R3: Radius; H: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      SetZero(H.R,Hr);
      Der3(R3,H,Hr);
      return Hr;
   end Der3;


   procedure DerDer3(R3: in Radius; H1: in Fourier2; H2: out Fourier2) is
   begin
      DerDer3(R3,H1.COS,H2.COS);
      DerDer3(R3,H1.SIN,H2.SIN);
      H2.R := H1.R;
      H2.R(3) := R3;
   end DerDer3;


   function DerDer3(R3: Radius; H: Fourier2) return Fourier2 is
      Hr: Fourier2;
   begin
      DerDer3(R3,H,Hr);
      return Hr;
   end DerDer3;


   procedure Random(Seed: in out Long_Integer; R: in Weights; H: out Fourier2; L: in Cutoffs) is
   begin
      SetZero(R,H);
      RandomOdd(Seed,0,R,H.COS,L);
      NextRandom(Seed);
      RandomOdd(Seed,1,R,H.SIN,L);
   end Random;


   procedure AddRandom(Seed: in out Long_Integer; J: in Parity; M: in Freq1; N: in Freq2; K: in Degree; H: in out Fourier2) is
   begin
      NextRandom(Seed);
      AddComponent(J,M,N,K,Scal((One-(Two*Rep(Seed))/Rep(2147483647))*Rho(H)(3)**(-K))/NormFactor(Rho(H),M,N),H);
   end AddRandom;


   procedure RandomOld(Seed: in out Long_Integer; R: in Weights; H: out Fourier2) is
   begin
      Random(Seed,0,R,H.COS);
      NextRandom(Seed);
      Random(Seed,1,R,H.SIN);
      H.R:=R;
   end RandomOld;


   procedure NextRandom(I: in out Long_Integer) is
      Hi,Lo: Long_Integer;
   begin
      Hi := I/127773;
      Lo := I-127773*Hi;
      I := 16807*Lo-2836*Hi;
      if I < 0 then
         I := I+2147483647;
      end if;
   end NextRandom;


   procedure Gig(L: in Cutoffs; H: in out Fourier2) is
      H2: Fourier2;
   begin
      SetZero(H.R,H2);
      Gig(L,H,H2);
      H:=H2;
   end Gig;


   procedure Gig(L: in Cutoffs; H1: in Fourier2; H2: out Fourier2) is
   begin
      Gig(L,H1.SIN,H2.SIN);
      Gig(L,H1.COS,H2.COS);
      H2.R:=H1.R;
   end Gig;


   procedure Ig(L: in Cutoffs; H: in out Fourier2) is
      H2: Fourier2;
   begin
      SetZero(H.R,H2);
      Ig(L,H,H2);
      H:=H2;
   end Ig;


   procedure Ig(L: in Cutoffs; H1: in Fourier2; H2: out Fourier2) is
   begin
      Ig(L,H1.SIN,H2.SIN);
      Ig(L,H1.COS,H2.COS);
      H2.R:=H1.R;
   end Ig;


   procedure Small_Gig(L: in Cutoffs; H1: in Fourier2; H2: out Fourier2) is
   begin
      Small_Gig(L,H1.SIN,H2.SIN);
      Small_Gig(L,H1.COS,H2.COS);
      H2.R:=H1.R;
   end Small_Gig;


   procedure Dig(L: in Cutoffs; H: in out Fourier2) is
   begin
      Digg(L,H.COS);
      Digg(L,H.SIN);
   end Dig;


   procedure Dig(L: in Cutoffs; H1: in  Fourier2; H2: out Fourier2) is
   begin
      SetZero(H1.R,H2);
      Digg(L,H1.COS,H2.SIN);
      Digg(L,H1.SIN,H2.COS);
      H2.R:=H1.R;
   end Dig;


   procedure ZcAdd(I: in Index3; H1: in Fourier2; H2: in out Fourier2) is
   begin
      ZcAdd(I,H1.SIN,H2.SIN);
      ZcAdd(I,H1.COS,H2.COS);
   end ZcAdd;


   procedure AddComponent(J: in Parity;M: in Freq1; N: in Freq2; P: in Taylor; H: in out Fourier2) is
      R:Weights;
   begin
      R:=H.R;
      if J=0 then
         AddComponent(M,N,P,H.COS);
      else
         AddComponent(M,N,P,H.SIN);
      end if;
      H.R:=R;
   end AddComponent;


   procedure AddComponent(M: in Freq1; N: in Freq2; Pc: in Taylor; Ps: in Taylor; H: in out Fourier2) is
      R:Weights;
   begin
      R:=H.R;
      AddComponent(M,N,Ps,H.SIN);
      AddComponent(M,N,Pc,H.COS);
      H.R:=R;
   end AddComponent;


   procedure AddComponent(J: in Parity; M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier2) is
   begin
      if J=0 then
         AddComponent(M,N,K,S,H.COS);
      else
         AddComponent(M,N,K,S,H.SIN);
      end if;
   end AddComponent;


   procedure AddComponent(M: in Freq1; N: in Freq2; K: in Degree; S1,S2: in Scalar; H: in out Fourier2) is
      R:Weights;
   begin
      R:=H.R;
      AddComponent(M,N,K,S2,H.SIN);
      AddComponent(M,N,K,S1,H.COS);
      H.R:=R;
   end AddComponent;


   function AddComponent(J: Parity; M: Freq1; N: Freq2; K: Degree; S: Scalar; H: Fourier2) return Fourier2 is
      H2: Fourier2;
   begin
      H2:=H;
      AddComponent(J,M,N,K,S,H2);
      return H2;
   end AddComponent;


   function AddComponent(M: Freq1; N: Freq2; K: Degree; S1: Scalar; S2: Scalar; H: Fourier2) return Fourier2 is
      H2: Fourier2;
   begin
      H2:=H;
      AddComponent(M,N,K,S1,S2,H2);
      return H2;
   end AddComponent;


   procedure SetComponent(J: in Parity; M: in Freq1; N: in Freq2; P: in Taylor; H: in out Fourier2) is
      R:Weights;
   begin
      R:=H.R;
      if InRange(M,N,J) then
         if J=0 then
            SetComponentNoCheck(M,N,P,H.COS);
         else
            SetComponentNoCheck(M,N,P,H.SIN);
         end if;
      else
         Message("Fourier_Ops.SetComponent",Index_Error);
      end if;
      H.R:=R;
   end SetComponent;


   procedure SetComponent(J: in Parity; M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier2) is
      R:Weights;
   begin
      R:=H.R;
      if InRange(M,N,J) then
         if J=0 then
            SetComponentNoCheck(M,N,K,S,H.COS);
         else
            SetComponentNoCheck(M,N,K,S,H.SIN);
         end if;
      else
         Message("Fourier_Ops.SetComponent",Index_Error);
      end if;
      H.R:=R;
   end SetComponent;


   procedure NumInv(Err: in Radius; H1: in Fourier2; H2: out Fourier2) is
      --  iterate h2 -> h2*[2-h1*h2]
      --  let h2(n) - n'th iteration, be h2(n)*h1=I+e(n), then h2(n+1)*h1=h2(n)*(2-h1*h2(n))*h1=
      --=2*h2(n)*h1-I-2*e(n)-e(n)^2=I-e(n)^2 and e(n+1)=e(n)^2
      Er: Rep:=1.0;
      S: Scalar;
      F: Scalar_Array;
      Hw1,Hw2: Fourier2;
      I,MaxII: Integer;
   begin
      MaxII:=5000;
      SetZero(H1.R,Hw1);
      SetZero(H1.R,Hw2);
      if not Trunc then
         Message("Fou2_Ops.NumInv",Numeric_Only);
      end if;
      S :=Component(0,0,0,H1.COS);
      if abs(Center(S)) > 1.0E-3 then
         Prod(-ScalOne/(S*S),H1,H2);
         AddComponent(0,0,0,0,ScalTwo/S,H2);
         NormFactors(H1.R,F);
         I:=0;
         while Er > Rep(Err) loop
            I:=I+1;
            Prod(H1.R,H1,H2,Hw1);
            AddComponent(0,0,0,0,-ScalOne,Hw1);
            Er := Sup(Norm(F,H1.R(3),Hw1));
            AddComponent(0,0,0,0,-ScalOne,Hw1);
            Prod(H1.R,H2,Hw1,Hw2);
            Neg(Hw2,H2);
            if (I>MaxII) then
               if (Er<1.0E-3) then
                  Put("  Exiting before required accuracy, Error in NumInv="); Put(Current_Output,Er);
                  New_Line;
                  return;
               else
                  Put("  Error in NumInv="); Put(Current_Output,Er);
                  New_Line;
                  Message("  Fou2_Ops.NumInv.Long.",Giving_Up);
               end if;
            end if;
            if (Er>Rep(1.0E+250)) then
               Message("  Fou2_Ops.NumInv.Blow_UP",Giving_Up);
            end if;
         end loop;
      else
         Message("  Fou2_Ops.NumInv",Division_By_Zero);
      end if;
   end NumInv;


   procedure Split(H: in Fourier2; Hc,He: out Fourier2) is
   begin
      Hc.R:=H.R;
      He.R:=H.R;
      Split(H.COS,Hc.COS,He.COS);
      Split(H.SIN,Hc.SIN,He.SIN);
   end Split;


   procedure CompOmega12(R: in Weights; H,Hq: in Fourier2; Hr: out Fourier2) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      Hr1,Hr2,Hw1,Hw2: Fourier2;
      P: Taylor;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier2; Hc2,Hs2: in out Fourier2) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
         Rr: Weights;
      begin
         Rr:=Min(Hc1.R,Hc2.R);
         SetZero(Rr,Hw1);
         SetZero(Rr,Hw2);
         Prod(Rr,Hs1,Hs2,Hw2);
         Prod(Rr,Hc1,Hs2,Hw1);
         Prod(Rr,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(Rr,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm,Dn,Dk: Integer;
      M: Freq1;
      Rr: Weights;
      Ht,Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmss,Hmnc,Hmns: Fourier2;
   begin
      SetZero(R,Ht);
      SetZero(R,Hc1);
      SetZero(R,Hs1);
      SetZero(R,Hc2);
      SetZero(R,Hs2);
      SetZero(R,Hmc);
      SetZero(R,Hms);
      SetZero(R,Hmss);
      SetZero(R,Hmnc);
      SetZero(R,Hmns);
      Copy(H,Ht);
      Rr := R;
      SetZero(Rr,Hr1);
      SetZero(Rr,Hr2);
      SetZero(Rr,Hr);
      CosSin(Hq,Hc1,Hs1); -- Hc1=Cos(Hq), Hs1=Sin(Hq);
      CosSin(-InvGold*Hq,Hc2,Hs2);  -- Hc2=Cos(-InvGold*Hq), Hc2=Sin(-InvGold*Hq),
      SetZero(Rr,Hmc);
      SetComponent(0,0,0,0,ScalOne,Hmc); --Cos(0)=1;
      SetZero(Rr,Hms); --Sin(0)=0;
      EffDeg(Ht,Dm,Dn,Dk);
      --CosSin(N)=CosSin(X+Hq);
      for Mabs in 0 .. Dm loop
         if Mabs>0 then
            Inc(Rr,Hc1,Hs1,Hmc,Hms); -- CosSin(Hq+N) from CosSin(InvGold*Hq) and CosSin(InvGold*N)
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc); --Cos(H)=Cos(Hq+N)
               Copy(Hms,Hmns); --Sin(H)=Sin(Hq+N)
               EffDeg(M,Ht,Dn,Dk);

               for N  in  IMin(Nmin(M,0),Nmin(M,1)) .. Dn loop
                  if N /= 0 then
                     Inc(Rr,Hc2,Hs2,Hmnc,Hmns); --CosSin(N)= CosSin(InvGold*InvGold*Hq+N) from CosSin(InvGold*InvGold*Hq) and CosSin(N)
                  end if;
                  if N>=Nmin(M,0) then
                     P:=Component(0,M,N,Ht); --take the scaled taylor of thus developed Cos(M,N) of composed Ht
                     MultAdd(0,ScalOne,M,N,P,Hmnc,Hr1);
                     MultAdd(1,-ScalOne,M,N,P,Hmns,Hr1);
                  end if;
                  if N>=Nmin(M,1) then
                     if (M/=0 or N/=0) then
                        P:=Component(1,M,N,Ht); --take the scaled taylor of thus developed Sin(M,N) of composed Ht
                        MultAdd(0,ScalOne,M,N,P,Hmns,Hr2);
                        MultAdd(1,ScalOne,M,N,P,Hmnc,Hr2);
                     end if;
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
      Hr:=Hr1+Hr2;
   end CompOmega12;


   procedure CompOmega12(R: in Weights; H1,H2,Hq: in Fourier2; Hr1,Hr2: out Fourier2) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      Hr11,Hr12,Hr21,Hr22,Hw1,Hw2: Fourier2;
      P: Taylor;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier2; Hc2,Hs2: in out Fourier2) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
         Rr: Weights;
      begin
         Rr:=Min(Hc1.R,Hc2.R);
         SetZero(Rr,Hw1);
         SetZero(Rr,Hw2);
         Prod(Rr,Hs1,Hs2,Hw2);
         Prod(Rr,Hc1,Hs2,Hw1);
         Prod(Rr,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(Rr,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm1,Dn1,Dk1,Dm2,Dn2,Dk2: Integer;
      M: Freq1;
      Rr: Weights;
      Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmss,Hmnc,Hmns: Fourier2;
   begin
      SetZero(R,Hc1);
      SetZero(R,Hs1);
      SetZero(R,Hc2);
      SetZero(R,Hs2);
      SetZero(R,Hmc);
      SetZero(R,Hms);
      SetZero(R,Hmss);
      SetZero(R,Hmnc);
      SetZero(R,Hmns);
      Rr := R;
      SetZero(Rr,Hr11);
      SetZero(Rr,Hr12);
      SetZero(Rr,Hr21);
      SetZero(Rr,Hr22);
      CosSin(Hq,Hc1,Hs1); --FINE, Hc1=Cos(Hq), Hs1=Sin(Hq);
      CosSin(-InvGold*Hq,Hc2,Hs2);  --FINE, Hc2=Cos(-InvGold*Hq), Hc2=Sin(-InvGold*Hq),
      SetZero(Rr,Hmc);
      SetComponent(0,0,0,0,ScalOne,Hmc); --Cos(0)=1;
      SetZero(Rr,Hms); --Sin(0)=0;
      EffDeg(H1,Dm1,Dn1,Dk1);
      EffDeg(H2,Dm2,Dn2,Dk2);
      --CosSin(N)=CosSin(X+Hq);
      for Mabs in 0 .. Imax(Dm1,Dm2) loop
         if Mabs>0 then
            Inc(Rr,Hc1,Hs1,Hmc,Hms); -- CosSin(Hq+N) from CosSin(InvGold*Hq) and CosSin(InvGold*N)
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc); --Cos(H)=Cos(Hq+N)
               Copy(Hms,Hmns); --Sin(H)=Sin(Hq+N)
               EffDeg(M,H1,Dn1,Dk1);
               EffDeg(M,H2,Dn2,Dk2);
               for N  in  IMin(Nmin(M,0),Nmin(M,1)) .. Imax(Dn1,Dn2) loop
                  if N /= 0 then
                     Inc(Rr,Hc2,Hs2,Hmnc,Hmns); --CosSin(N)= CosSin(InvGold*InvGold*Hq+N) from CosSin(InvGold*InvGold*Hq) and CosSin(N)
                  end if;
                  if N>=Nmin(M,0) then
                     P:=Component(0,M,N,H1);
                     MultAdd(0,ScalOne,M,N,P,Hmnc,Hr11);
                     MultAdd(1,-ScalOne,M,N,P,Hmns,Hr11);
                     P:=Component(0,M,N,H2);
                     MultAdd(0,ScalOne,M,N,P,Hmnc,Hr21);
                     MultAdd(1,-ScalOne,M,N,P,Hmns,Hr21);
                  end if;
                  if N>=Nmin(M,1) then
                     if (M/=0 or N/=0) then
                        P:=Component(1,M,N,H1);
                        MultAdd(0,ScalOne,M,N,P,Hmns,Hr12);
                        MultAdd(1,ScalOne,M,N,P,Hmnc,Hr12);
                        P:=Component(1,M,N,H2);
                        MultAdd(0,ScalOne,M,N,P,Hmns,Hr22);
                        MultAdd(1,ScalOne,M,N,P,Hmnc,Hr22);
                     end if;
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
      Hr1:=Hr11+Hr12;
      Hr2:=Hr21+Hr22;
   end CompOmega12;


   procedure CompOmegaTmu0(R: in Weights; H,Hq: in Fourier2; Hr: out Fourier2) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      S0: constant Scalar := Golden*Golden*Golden;
      S2: constant Scalar := -InvGold;
      InvS2: constant Scalar := -Golden;
      Hr1,Hr2,Hw1,Hw2: Fourier2;
      P: Taylor;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier2; Hc2,Hs2: in out Fourier2) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
         Rr: Weights;
      begin
         Rr:=Min(Hc1.R,Hc2.R);
         SetZero(Rr,Hw1);
         SetZero(Rr,Hw2);
         Prod(Rr,Hs1,Hs2,Hw2);
         Prod(Rr,Hc1,Hs2,Hw1);
         Prod(Rr,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(Rr,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm,Dn,Dk: Integer;
      M: Freq1;
      Rr: Weights;
      Test,Ht,Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmss,Hmnc,Hmns: Fourier2;
   begin
      SetZero(R,Ht);
      SetZero(R,Hc1);
      SetZero(R,Hs1);
      SetZero(R,Hc2);
      SetZero(R,Hs2);
      SetZero(R,Hmc);
      SetZero(R,Hms);
      SetZero(R,Hmss);
      SetZero(R,Hmnc);
      SetZero(R,Hmns);
      SetZero(R,Test);
      Copy(H,Ht);
      Rr := R;
      SetZero(Rr,Hr1);
      SetZero(Rr,Hr2);
      SetZero(Rr,Hr);
      CosSin((-InvGold)*Hq,Hc1,Hs1); -- Hc1=Cos(InvGold*Hq), Hs1=Sin(InvGold*Hq);
      CosSin(InvGold*InvGold*Hq,Hc2,Hs2);  -- Hc2=Cos(Sqr(InvGold)*Hq), Hc2=Sin(Sqr(InvGold)*Hq),
      SetZero(Rr,Hmc);
      SetComponent(0,0,0,0,ScalOne,Hmc); --Cos(0)=1;
      SetZero(Rr,Hms); --Sin(0)=0;
      EffDeg(Ht,Dm,Dn,Dk);
      --CosSin(N)=CosSin(X+Hq);
      for Mabs in 0 .. Dm loop
         if Mabs>0 then
            Inc(Rr,Hc1,Hs1,Hmc,Hms); -- CosSin(Hq+N) from CosSin(InvGold*Hq) and CosSin(InvGold*N)
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc); --Cos(H)=Cos(Hq+N)
               Copy(Hms,Hmns); --Sin(H)=Sin(Hq+N)
               EffDeg(M,Ht,Dn,Dk);
               for N  in  IMin(Nmin(M,0),Nmin(M,1)) .. Dn loop
                  if N /= 0 then
                     Inc(Rr,Hc2,Hs2,Hmnc,Hmns); --CosSin(N)= CosSin(InvGold*InvGold*Hq+N) from CosSin(InvGold*InvGold*Hq) and CosSin(N)
                  end if;
                  if N>=Nmin(M,0) then
                     Scale(ScalOne,S2,Component(0,M,N,Ht),P); --take the scaled taylor of thus developed Cos(M,N) of composed Ht
                     MultAdd(0,S0,N,M+N,P,Hmnc,Hr1);
                     MultAdd(1,Scal(-1.0)*S0,N,M+N,P,Hmns,Hr1);
                  end if;
                  if N>=Nmin(M,1) then
                     Scale(ScalOne,S2,Component(1,M,N,Ht),P); --take the scaled taylor of thus developed Sin(M,N) of composed Ht
                     MultAdd(0,S0,N,M+N,P,Hmns,Hr2);
                     MultAdd(1,S0,N,M+N,P,Hmnc,Hr2);
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
      Hr:=Hr1+Hr2;
   end CompOmegaTmu0;


   procedure Rotation(R: in Weights; S1,S2: in Scalar; H,Hq: in Fourier2; Hr: out Fourier2) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      Hr1,Hr2,Hw1,Hw2: Fourier2;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier2; Hc2,Hs2: in out Fourier2) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
         Rr: Weights;
      begin
         Rr:=Min(Hc1.R,Hc2.R);
         SetZero(Rr,Hw1);
         SetZero(Rr,Hw2);
         Prod(Rr,Hs1,Hs2,Hw2);
         Prod(Rr,Hc1,Hs2,Hw1);
         Prod(Rr,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(Rr,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm,Dn,Dk,Repeat: Integer;
      M: Freq1;
      Rd2: Rep;
      Rr: Weights;
      Ht,Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmss,Hmnc,Hmns: Fourier2;
   begin
      SetZero(R,Ht);
      SetZero(R,Hc1);
      SetZero(R,Hs1);
      SetZero(R,Hc2);
      SetZero(R,Hs2);
      SetZero(R,Hmc);
      SetZero(R,Hms);
      SetZero(R,Hmss);
      SetZero(R,Hmnc);
      SetZero(R,Hmns);
      Split(H,Ht,Hw1);
      Rr := R;
      if not Trunc then
         Repeat := 2;
         while Repeat>0 loop
            Rd2 := Sup(Norm(Rr,Hq));                                       ---- delta2
            if H.R(2)>=Rr(2)+Rd2 then
               Repeat := 0;
            else
               Repeat := Repeat-1;
               Put_Line("Fou2_Ops.CompOmega12, Warning: Shrinking domain");
               Put("Rho(H)(2) = "); Put(Current_Output,H.R(2),5,3); New_Line;
               Put("   ||Hq|| = "); Put(Current_Output,Rd2,5,3);    New_Line;
               Put("old    R2 = "); Put(Current_Output,Rr(2),5,3);  New_Line;
               Rr(2) := H.R(2)-Rd2;
               Put("new    R2 = "); Put(Current_Output,Rr(2),5,3);  New_Line;
               if Repeat=0 then
                  Message("Fou2_Ops.CompOmega12",Domain_Violation);
               end if;
            end if;
         end loop;
         MultAdd(ScalOne+Scal(Rd2)/Scal(Rr(2)),Hw1,Ht);
      end if;
      SetZero(Rr,Hr1);
      SetZero(Rr,Hr2);
      SetZero(Rr,Hr);
      CosSinX(S1*Hq,Hc1,Hs1);
      CosSinY(S2*Hq,Hc2,Hs2);
      SetZero(Rr,Hmc);
      SetComponent(0,0,0,0,ScalOne,Hmc);
      SetZero(Rr,Hms);
      EffDeg(Ht,Dm,Dn,Dk);
      for Mabs in 0 .. Dm loop
         if Mabs>0 then
            Inc(Rr,Hc1,Hs1,Hmc,Hms);
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc);
               Copy(Hms,Hmns);
               EffDeg(M,Ht,Dn,Dk);
               for N  in  IMin(Nmin(M,0),Nmin(M,1)) .. Dn loop
                  if N > 0 then
                     Inc(Rr,Hc2,Hs2,Hmnc,Hmns);
                  end if;
                  if N>=Nmin(M,0) then
                     MultAdd(Rr,Component(0,M,N,Ht),Hmnc,Hr1);
                  end if;
                  if N>=Nmin(M,1) then
                     if (M/=0 or N/=0) then
                        MultAdd(Rr,Component(1,M,N,Ht),Hmns,Hr2);
                     end if;
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
      Hr:=Hr1+Hr2;
   end Rotation;


   function NumInv(Err: Radius; H1: Fourier2) return Fourier2 is
      H2: Fourier2;
   begin
      SetZero(H1.R,H2);
      NumInv(Err,H1,H2);
      return H2;
   end;


   procedure CosSin(H: in Fourier2; Hc: out Fourier2; Hs: out Fourier2) is
      Hcc,Hcs,Hsc,Hss: Fourier;
   begin
      SetZero(H.R,Hs);
      SetZero(H.R,Hc);
      CosSin(H.COS,Hcc,Hsc); --Hcc=Cos(H.COS) and Hsc=Sin(H.COS)
      CosSin(H.SIN,Hcs,Hss); --Hcs=Cos(H.SIN) and Hss=Sin(H.SIN)
      Hc.COS:=Hcc*Hcs;
      Hc.SIN:=-Hsc*Hss;
      Hs.COS:=Hsc*Hcs;
      Hs.SIN:=Hcc*Hss;
   end;


   procedure CosSin_New(H: in Fourier2; Hc,Hs: out Fourier2) is
      Iter: constant Integer := 20;
      Small: constant Rep := NewCutoff;
      K: Natural;
      E,Efc,Efs: Error;
      Hw0c,Hw1c,Hw2c: Fourier;
      Hw0s,Hw1s,Hw2s: Fourier;
      Hcc,Hcs,Hsc,Hss: Fourier;
   begin
      K :=0;
      E := Norm(H);

      while Sup(E)>Small loop
         K := K+1;
         E := Half*E;
      end loop;
      Prod(Scal(Half**K),H.COS,Hw0c);                              ---- divide H by 2, K times
      Copy(Hw0c,Hsc);
      SetZero(0,H.R,Hcc);
      AddComponent(0,0,0,ScalOne,Hcc);
      Copy(Hw0c,Hw1c);
      Efc := Exp(Scal(Sup(E)))*E**(2*Iter+1);
      Prod(Scal(Half**K),H.SIN,Hw0s);                              ---- divide H by 2, K times
      Copy(Hw0s,Hss);
      SetZero(0,H.R,Hcs);
      AddComponent(0,0,0,ScalOne,Hcs);
      Copy(Hw0s,Hw1s);
      Efs := Exp(Scal(Sup(E)))*E**(2*Iter+1);
      for I in 1 .. Iter loop                                 ---- Taylor polynomials
         Prod(H.R,Hw0c,Hw1c,Hw2c);
         Mult(Inv(Scal(-2*I)),Hw2c);
         Add(Hw2c,Hcc);
         Prod(H.R,Hw0c,Hw2c,Hw1c);
         Mult(Inv(Scal(2*I+1)),Hw1c);
         Add(Hw1c,Hsc);
         Efc := Efc/(2*I*(2*I+1));
         Prod(H.R,Hw0s,Hw1s,Hw2s);
         Mult(Inv(Scal(-2*I)),Hw2s);
         Add(Hw2s,Hcs);
         Prod(H.R,Hw0s,Hw2s,Hw1s);
         Mult(Inv(Scal(2*I+1)),Hw1s);
         Add(Hw1s,Hss);
         Efs := Efs/(2*I*(2*I+1));
      end loop;
      AddComponent(0,0,0,Scal(Efc),Hcc);                        ---- Remainder for Cos(H)
      Efc := Efc*E/(2*Iter+2);
      AddComponent(0,0,0,Scal(Efc),Hsc);                        ---- Remainder for Sin(H)
      AddComponent(0,0,0,Scal(Efs),Hcs);                        ---- Remainder for Cos(H)
      Efs := Efs*E/(2*Iter+2);
      AddComponent(0,0,0,Scal(Efc),Hss);                        ---- Remainder for Sin(H)
      for I in 1 .. K loop                                    ---- double angle formula, K times
         Prod(H.R,Hsc,Hsc,Hw1c);
         Prod(H.R,Hsc,Hcc,Hw2c);
         Prod(ScalTwo,Hw2c,Hsc);
         Prod(H.R,Hcc,Hcc,Hw2c);
         Diff(Hw2c,Hw1c,Hcc);
         Prod(H.R,Hss,Hss,Hw1s);
         Prod(H.R,Hss,Hcs,Hw2s);
         Prod(ScalTwo,Hw2s,Hss);
         Prod(H.R,Hcs,Hcs,Hw2s);
         Diff(Hw2s,Hw1s,Hcs);
      end loop;
      Prod(H.R,Hcc,Hcs,Hc.COS);
      Prod(H.R,Hsc,Hss,Hc.SIN);
      Hc.SIN:=-Hc.SIN;
      Prod(H.R,Hsc,Hcs,Hs.COS);
      Prod(H.R,Hcc,Hss,Hs.SIN);
   end CosSin_New;


   procedure CosSinX(H: in Fourier2; Hc,Hs: out  Fourier2) is
      Hwc,Hws,Hcc: Fourier2;
   begin
      SetZero(H.R,Hwc);
      SetZero(H.R,Hws);
      SetZero(H.R,Hcc);
      CosSin(H,Hwc,Hws);
      Prod(Mon(0,1,0,0,ScalOne),Hwc,Hc);
      MultAdd(Mon(1,1,0,0,-ScalOne),Hws,Hc);
      Prod(Mon(0,1,0,0,ScalOne),Hws,Hs);
      MultAdd(Mon(1,1,0,0,ScalOne),Hwc,Hs);
   end CosSinX;


   procedure CosSinY(H: in Fourier2; Hc,Hs: out Fourier2) is
      Hwc,Hws: Fourier2;
   begin
      SetZero(H.R,Hws);
      SetZero(H.R,Hwc);
      CosSin(H,Hwc,Hws);
      Prod(Mon(0,0,1,0,ScalOne),Hwc,Hc);
      MultAdd(Mon(1,0,1,0,-ScalOne),Hws,Hc);
      Prod(Mon(0,0,1,0,ScalOne),Hws,Hs);
      MultAdd(Mon(1,0,1,0,ScalOne),Hwc,Hs);
   end CosSinY;


   function Component(J: Parity; M,N: Integer; H: Fourier2) return Taylor is
   begin
      if J=0 then
         return Component(M,N,H.COS);
      else
         return Component(M,N,H.SIN);
      end if;
   end Component;


   function Component(J: Parity; M,N: Integer; K: Natural; H: Fourier2) return Scalar is -- coefficient in front of nu=[M,N], k=K
   begin
      if InRange(M,N,J) then
         if J=0 then
            return Component(M,N,K,H.COS);
         else return Component(M,N,K,H.SIN);
         end if;
      else
         return ScalZero;
      end if;
   end Component;


   function NumVal(X,Y,Z: Scalar; H: Fourier2) return Scalar is
   begin
      return  NumVal(X,Y,Z,H.COS)+NumVal(X,Y,Z,H.SIN);
   end;


   procedure Comp3(H,H3: in Fourier2; Hr: out Fourier2) is
      Hw: Fourier2;
   begin
      SetZero(H3.R,Hw);
      if not Trunc and then Sup(Norm(H3))>H.R(3) then
         Message("Fourier_Ops.Comp3",Domain_Violation);
         Put(" H.R(3) =");
         Put(Current_Output,H.R(3),5,4); New_Line;
         Put("  ||H3|| =");
         Put(Current_Output,Sup(Norm(H3)),5,4); New_Line;
      end if;
      SetZero(H3.R,Hr);
      ZcAdd(Pdeg,H,Hr);
      for I in reverse 0 .. Pdeg-1 loop
         Prod(H3.R,Hr,H3,Hw);
         ZcAdd(I,H,Hw);  -- H2[M,N,0] := H2[M,N,0]+H1[M,N,I] ...
         Copy(Hw,Hr);
      end loop;
   end Comp3;


   function Comp3(H,H3: Fourier2) return Fourier2 is
      Hw, Hr: Fourier2;
   begin
      SetZero(H.R,Hw);
      SetZero(H.R,Hr);
      if not Trunc and then Sup(Norm(H3))>H.R(3) then
         Message("Fourier_Ops.Comp3",Domain_Violation);
         Put(" H.R(3) =");
         Put(Current_Output,H.R(3),5,4); New_Line;
         Put("  ||H3|| =");
         Put(Current_Output,Sup(Norm(H3)),5,4); New_Line;
      end if;
      SetZero(Min(H.R,H3.R),Hr);
      Hr.R(3) := H3.R(3);
      ZcAdd(Pdeg,H,Hr);
      for I in reverse 0 .. Pdeg-1 loop
         Prod(Hr.R,Hr,H3,Hw);
         ZcAdd(I,H,Hw);
         Copy(Hw,Hr);
      end loop;
      return Hr;
   end Comp3;


   function Rho(H: Fourier2) return Weights is
   begin
      return H.R;
   end Rho;


   procedure NumInvComp3(Err: in Radius; H1,H2: in Fourier2; H3: out Fourier2) is
      --  iterate h3 -> h3-[h2(..,h3)-h1]/(D_3 h2)(..,h3)
      Er: Rep:=1.0;
      R: Weights;
      F: Scalar_Array;
      H,Hd,Hd2,Hdd,Hd3,Ht,Hc: Fourier2;
      P:Taylor;
      I,MaxII: Integer;
   begin
      MaxII:=50000;
      R:= Min(H1.R,H2.R);
      SetZero(R,H);
      SetZero(R,Hd);
      SetZero(R,Hd2);
      SetZero(R,Hdd);
      SetZero(R,Hd3);
      SetZero(R,Ht);
      SetZero(R,Hc);
      if not Trunc then
         Message("Fou2_Ops.NumInvComp3",Numeric_Only);
      end if;
      NormFactors(R,F);
      SetZero(P);
      SetZero(R,H3);
      InvComp(R(3),Err,Component(0,0,0,H1),Component(0,0,0,H2),P); -- solves C2(P)=C1
      SetComponent(0,0,0,P,H3);
      Der3(R(3),H2,H);
      NumInv(Err,H,Hd2);
      I:=0;
      while (Er>Rep(Err)) loop
         I:=I+1;
         Comp3(H2,H3,H);
         Sub(H1,H);
         Der3(R(3),H2,Hd2);
         Comp3(Hd2,H3,Hd);
         NumInv(Err,Hd,Hdd);
         Prod(R,H,Hdd,Hd3);
         Sub(Hd3,H3);
         Er:= Sup(Norm(R,Hd3));
         if (I>MaxII) then
            if (Er<1.0E-3) then
               Put("  Exiting before required accuracy, Error in NumInvComp3="); Put(Current_Output,Er);
               New_Line;
               return;
            else
               Put("  Error in NumInvComp3="); Put(Current_Output,Er);
               New_Line;
               Message("  Fou2_Ops.NumInvComp3.Long.",Giving_Up);
            end if;
         end if;
         if (Er>Rep(1.0E+250)) then
            Message("  Fou2_Ops.NumInvComp3.Blow_UP",Giving_Up);
         end if;
      end loop;
   end NumInvComp3;


   procedure NumFixComp3(Err: in Radius; H: in Fourier2; Hz: out Fourier2) is
      H1,H2: Fourier2;
   begin
      SetZero(H.R,H1);
      SetZero(H.R,H2);
      Copy(H,H2);
      AddComponent(0,0,0,1,ScalOne,H2);
      SetZero(Rho(H),H1);
      SetComponent(0,0,0,1,ScalOne,H1);
      NumInvComp3(Err,H1,H2,Hz);
      AddComponent(0,0,0,1,-ScalOne,Hz);
   end NumFixComp3;


   function NumFixComp3(Err: Radius; H: Fourier2) return Fourier2 is
      Hz: Fourier2;
   begin
      SetZero(H.R,Hz);
      NumFixComp3(Err,H,Hz);
      return Hz;
   end NumFixComp3;


   procedure NumInormalize(R: in Weights; H: in out Fourier2) is
      C,S: Scalar;
   begin
      if not Trunc then
         Message("Fou2_Ops.NumInormalize",Numeric_Only);
      end if;
      C := Component(0,0,0,1,H);
      S := NumIprod(R,H,H);
      if Center(C)>Zero then
         Mult(Inv(Sqrt(S)),H);
      else
         Mult(-Inv(Sqrt(S)),H);
      end if;
   end NumInormalize;


   function NumIprod(R: Weights; H1,H2: Fourier2) return Scalar is
      S: Scalar;
      F: Scalar_Array;
      P1,P2: Taylor;
   begin
      if not Trunc then
         Message("Fou2_Ops.NumIprod",Numeric_Only);
      end if;
      NormFactors(R,F);
      S := ScalZero;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            P1:=Component(0,M,N,H1);
            P2:=Component(0,M,N,H2);
            if (Deg(P1)/=-1 and Deg(P2)/=-1) then
               S := S+NumIprod(R(3),P1,P2)*F(M,N)*F(M,N);
            end if;
         end loop;
         for N in Nmin(M,1) .. Nmax(M) loop
            P1:=Component(1,M,N,H1);
            P2:=Component(1,M,N,H2);
            if (Deg(P1)/=-1 and Deg(P2)/=-1) then
               S := S+NumIprod(R(3),P1,P2)*F(M,N)*F(M,N);
            end if;
         end loop;
      end loop;
      return S;
   end NumIprod;


end Fou2_Ops;



































