with Messages, Ada.Integer_Text_IO;
use Messages, Ada.Integer_Text_IO;

package body Fourier_Ops is


   ScalZero: constant Scalar := Scal(Zero);
   ScalOne: constant Scalar := Scal(One);
   ScalTwo: constant Scalar := Scal(Two);
   ScalHalf: constant Scalar := Scal(Half);
   InvGold: constant Scalar := Golden-ScalOne;
   Trunc: constant Boolean := IsNumeric(ScalZero);
   NewCutoff: constant Rep:= Rep(0.01);


   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;


   function Nmin(M: in Integer; J: in Parity) return Integer is
   begin
      -- for even part if M<0 Nmin=1, M>=0 Nmin=0
      -- for odd part if M<1 Nmin=1, M>=1 Nmin=0
      if M < J then
         return 1;
      else
         return 0;
      end if;
   end Nmin;


   procedure CopySign(H1: in Fourier; H2: in out Fourier) is
   begin
      for M in Freq1 loop
         for N in IMax(Nmin(M,H1.J),Nmin(M,H2.J)) .. Nmax(M) loop
            CopySign(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end CopySign;


   procedure EvenNProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop   --change
            if abs(N) mod 2=0 then
               Copy(H1.C(M,N),H2.C(M,N));
            end if;

         end loop;
      end loop;
   end EvenNProj;


   procedure EvenMProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         if abs(M) mod 2=0 then
            for N in Nmin(M,H1.J) .. Nmax(M) loop   --change
               Copy(H1.C(M,N),H2.C(M,N));
            end loop;
         end if;
      end loop;
   end EvenMProj;


   procedure EvenMNProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop   --change
            if abs(M+N) mod 2=0 then
               Copy(H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
   end EvenMNProj;


   procedure OddNProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop    --change
            if abs(N) mod 2=1 then
               Copy(H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
   end OddNProj;


   procedure OddMProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         if abs(M) mod 2=1 then
            for N in Nmin(M,H1.J) .. Nmax(M) loop  --change
               Copy(H1.C(M,N),H2.C(M,N));
            end loop;
         end if;
      end loop;
   end OddMProj;


   procedure OddMNProj(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop --change
            if  abs(M+N) mod 2=1 then
               Copy(H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
   end OddMNProj;


   procedure PiTranslateX(H1: in Fourier; H2:  out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if M mod 2=0 then
               Copy(H1.C(M,N),H2.C(M,N));
            else
               Copy(Scal(-1.0)*H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
   end PiTranslateX;


   procedure PiTranslateY(H1: in Fourier; H2:  out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if N mod 2=0 then
               Copy(H1.C(M,N),H2.C(M,N));
            else
               Copy(Scal(-1.0)*H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
   end PiTranslateY;


   function Nmax(M: in Integer) return Integer is
   begin
      return IMin(Size2,Qdeg-Abs(M));
   end Nmax;


   function InRange(M: in Integer; N: in Integer; J: in Parity) return Boolean is
   begin
      return M in Freq1 and then Nmin(M,J) <= N and then N <= Nmax(M);
   end InRange;


   procedure MaxDegs(Degs: out Degrees; Count: out Natural) is
      D: Integer;
   begin
      Count := 0;
      for M in Freq1 loop
         for N in Freq2 loop
            D := QPdeg-Abs(M)-N;
            if D >= 0 and then InRange(M,N,0) then
               Degs(M,N) := IMin(D,Pdeg);
               Count := Count+Degs(M,N)+1;
            else
               Degs(M,N) := -1;
            end if;
         end loop;
      end loop;
   end MaxDegs;


   function Deg(M: Freq1; N: Freq2; H: Fourier) return Integer is
   begin
      return Deg(H.C(M,N));
   end Deg;


   function EffDeg(M,N: integer; H: Fourier) return Integer is
   begin
      if InRange(M,N,H.J) then
         return EffDeg(H.C(M,N));
      end if;
      return -1;
   end EffDeg;


   procedure EffDeg(M,N: in Integer; H: in Fourier; Dk: out Integer) is
   begin
      if Inrange(M,N,H.J) then
         Dk := EffDeg(H.C(M,N));
      else
         Dk := -1;
      end if;
   end EffDeg;


   procedure EffDeg(M: in Integer; H: in Fourier; Dn,Dk: out Integer) is
      K: Integer;
   begin
      Dn := -1;
      Dk := -1;
      for N in Nmin(M,H.J) .. Nmax(M) loop
         K := EffDeg(H.C(M,N));
         if K>=0 then
            Dn := Imax(N,Dn);
            Dk := Imax(K,Dk);
         end if;
      end loop;
   end EffDeg;


   procedure EffDeg(H: in Fourier; Dm,Dn,Dk: out Integer) is
      K: Integer;
   begin
      Dm := -1;
      Dn := -1;
      Dk := -1;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            K := EffDeg(H.C(M,N));
            if K>=0 then
               Dm := Imax(Abs(M),Dm);
               Dn := Imax(N,Dn);
               Dk := Imax(K,Dk);
            end if;
         end loop;
      end loop;
   end EffDeg;


   procedure ShowR(R: in Weights) is
   begin
      Put(Current_Output," R=(");
      Put(Current_Output,R(1),5,3);
      Put(Current_Output,R(2),5,3);
      Put(Current_Output,R(3),5,3);
      Put_Line(" )");
   end ShowR;


   procedure ShowJR(H: in Fourier) is
   begin
      Put(Current_Output,"J=");
      Put(Current_Output,H.J,1,10);
      ShowR(H.R);
   end ShowJR;


   procedure ShowIndices(J: in Parity; M: in Freq1; N: in Freq2; K: in Index3) is
   begin
      if J=0 then
         Put(Current_Output,"cos(");
      else
         Put(Current_Output,"sin(");
      end if;
      Put(Current_Output,M,3,10);
      Put(Current_Output,"x+");
      Put(Current_Output,N,2,10);
      Put(Current_Output,"y) z**");
      Put(Current_Output,K,2,10);
      Put(Current_Output,"  ");
   end ShowIndices;


   procedure CheckPar(H: in Fourier; M: in Integer :=0; N: in Integer :=0) is
   begin
      if H.J=1 and then M=0 and then N=0 and then not Center0(H.C(0,0)) then
         Message("Fourier_Ops.CheckPar",Parity_Error);
      end if;
   end CheckPar;


   function Par(H: Fourier) return Parity is
   begin
      CheckPar(H);
      return H.J;
   end Par;


   procedure SetPar(J: in Parity; H: in out Fourier) is
   begin
      H.J := J;
      if J=1 then
         ResetCenter(H.C(0,0));
      end if;
   end SetPar;


   function Rho(H: Fourier) return Weights is
   begin
      return H.R;
   end Rho;


   procedure SetRho(R: in Weights; H: in out Fourier) is
   begin
      H.R := R;
   end SetRho;


   function Min(R1,R2: Weights) return Weights is
   begin
      return (Min(R1(1),R2(1)),Min(R1(2),R2(2)),Min(R1(3),R2(3)));
   end Min;


   procedure Component(M,N: in Integer; H: in Fourier; P: out Taylor) is    --taylor in front of the f. c. w/ nu=[M,N]
   begin
      if InRange(M,N,H.J) then
         Copy(H.C(M,N),P);
      else
         SetZero(P);
      end if;
   end Component;


   function Component(M,N: Integer; H: Fourier) return Taylor is
      P: Taylor;
   begin
      Component(M,N,H,P);
      return P;
   end Component;


   function Component(M,N: Integer; K: Natural; H: Fourier) return Scalar is -- coefficient in front of nu=[M,N], k=K
   begin
      if InRange(M,N,H.J) then
         return Component(K,H.C(M,N));
      else
         return ScalZero;
      end if;
   end Component;


   procedure SetComponent(M: in Freq1; N: in Freq2; P: in Taylor; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         Copy(P,H.C(M,N));
         CheckPar(H,M,N);
      else
         Message("Fourier_Ops.SetComponent",Index_Error);
      end if;
   end SetComponent;


   procedure SetComponentNoCheck(M: in Freq1; N: in Freq2; P: in Taylor; H: in out Fourier) is
   --SetComponent without the parity check
   begin
      if InRange(M,N,H.J) then
         Copy(P,H.C(M,N));
      else
         Message("Fourier_Ops.SetComponentNew",Index_Error);
      end if;
   end SetComponentNoCheck;


   procedure AddComponent(M: in Freq1; N: in Freq2; P: in Taylor; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         Add(P,H.C(M,N));
      else
         Message("Fourier_Ops.AddComponent",Index_Error);
      end if;
   end AddComponent;


   procedure MultComponent(M: in Freq1; N: in Freq2; S: in Scalar; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         Mult(S,H.C(M,N));
      else
         Message("Fourier_Ops.MultComponent",Index_Error);
      end if;
   end MultComponent;


   procedure SetComponent(M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         SetComponent(K,S,H.C(M,N));
         CheckPar(H,M,N);
      else
         Message("Fourier_Ops.SetComponent",Index_Error);
      end if;
   end SetComponent;


   procedure SetComponentNoCheck(M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier) is
      --SetComponent without the parity check
   begin
      if InRange(M,N,H.J) then
         SetComponent(K,S,H.C(M,N));
      else
         Message("Fourier_Ops.SetComponentNew",Index_Error);
      end if;
   end SetComponentNoCheck;


   procedure AddComponent(M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         AddComponent(K,S,H.C(M,N));
         CheckPar(H,M,N);      --change
      else
         Message("Fourier_Ops.AddComponent",Index_Error);
      end if;
   end AddComponent;


   procedure AddComponentNoCheck(M: in Freq1; N: in Freq2; K: in Degree; S: in Scalar; H: in out Fourier) is
   begin
      if InRange(M,N,H.J) then
         AddComponent(K,S,H.C(M,N));
      else
         Message("Fourier_Ops.AddComponent",Index_Error);
      end if;
   end AddComponentNoCheck;


   function NormFactor(R: Weights; M,N: Integer) return Scalar is
   begin
      if N>0 or else (N=0 and M>=0) then
         if N <= Nmax(M) and then R = SavedFactors.R then
            return SavedFactors.C(M,N);
         else
            return Exp(Abs(Scal(R(1))*(Scal(M)*InvGold+Scal(N))))*Cosh(Abs(Scal(R(2)))*(Scal(M)-Scal(N)*InvGold));
         end if;
      else
         return NormFactor(R,-M,-N);
      end if;
   end NormFactor;


   procedure NormFactors(R: in Weights; F: out Scalar_Array) is
      Fm1,Fn1,Fm2,Fn2,Pm1,Pm2,Pmn1,Pmn2: Scalar;
   begin
      if R=SavedFactors.R then
         F := SavedFactors.C;
      else
         Fm1 := Exp(Scal(R(1))*InvGold);
         Fn1 := Exp(Scal(R(1)));
         Fm2 := Exp(Scal(R(2)));                                    ---- start nonnegative m's
         Fn2 := Exp(Scal(-R(2))*InvGold);
         Pm1 := ScalOne;
         Pm2 := ScalOne;
         for M in 0 .. Size1 loop
            Pmn1 := Pm1;
            Pmn2 := Pm2;
            for N in 0 .. Nmax(M) loop
               F(M,N) := ExpAbsLog(Pmn1)*ScalHalf*(Pmn2+Inv(Pmn2));
               Pmn1 := Pmn1*Fn1;
               Pmn2 := Pmn2*Fn2;
            end loop;
            Pm1 := Pm1*Fm1;
            Pm2 := Pm2*Fm2;
         end loop;
         Fm1 := Inv(Fm1);                                            ---- now negative m's
         Fm2 := Inv(Fm2);
         Pm1 := Fm1;
         Pm2 := Fm2;
         for M in reverse -Size1 .. -1 loop
            Pmn1 := Pm1;
            Pmn2 := Pm2;
            for N in 0 .. Nmax(M) loop
               F(M,N) := ExpAbsLog(Pmn1)*ScalHalf*(Pmn2+Inv(Pmn2));
               Pmn1 := Pmn1*Fn1;
               Pmn2 := Pmn2*Fn2;
            end loop;
            Pm1 := Pm1*Fm1;
            Pm2 := Pm2*Fm2;
         end loop;
      end if;
   end NormFactors;


   procedure UpdateSavedFactors(R: in Weights) is
   begin
      if R /= SavedFactors.R then
         NormFactors(R,SavedFactors.C);
         SavedFactors.R := R;
      end if;
   end UpdateSavedFactors;


   function Coeff(M,N: Integer; H: Fourier) return Taylor is
      S: Scalar;
      P: Taylor;
   begin
      Component(M,N,H,P);
      if Trunc then return P; end if;
      Put("Strange in Coeff");
      New_Line;
      UpdateSavedFactors(H.R);
      ResetErr(P);
      if M=0 and N=0 then
         for Mh in Freq1 loop
            for Nh in Nmin(Mh,H.J) .. Nmax(Mh) loop
               S := ScalHalf*Inv(NormFactor(H.R,Mh,Nh));
               MultAdd(S,ResetCenter(H.C(Mh,Nh)),P);
            end loop;
         end loop;
      else
         for Mh in Freq1 loop
            for Nh in Nmin(Mh,H.J) .. Nmax(Mh) loop
               S := ScalHalf*(Inv(NormFactor(H.R,Mh+M,Nh+N))+Inv(NormFactor(H.R,Mh-M,Nh-N)));
               MultAdd(S,ResetCenter(H.C(Mh,Nh)),P);
            end loop;
         end loop;
      end if;
      return P;
   end Coeff;


   function Coeff(M,N: Integer; K: Natural; H: Fourier) return Scalar is
   begin
      return Component(K,Component(M,N,H));
   end Coeff;


   procedure SetZero(J: in Parity; R: in Weights; H: out Fourier) is
   begin
      for M in Freq1 loop
         for N in Nmin(M,J) .. Nmax(M) loop
            SetZero(H.C(M,N));
         end loop;
      end loop;
      H.J := J;
      H.R := R;
   end SetZero;


   procedure ZeroAverage(H: in out Fourier) is
   begin
      ResetCenter(H.C(0,0));
   end ZeroAverage;


   procedure Split(H: in Fourier; Hc,He: out Fourier) is
   begin
      Hc.J := H.J;
      Hc.R := H.R;
      He.J := H.J;
      He.R := H.R;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop

            Split(H.C(M,N),Hc.C(M,N),He.C(M,N));
         end loop;
      end loop;
      if H.J=1 then
         SetZero(Hc.C(0,0));
      end if;
   end Split;


   function Center0(H: Fourier) return Boolean is
   begin
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            if not Center0(H.C(M,N)) then
               return False;
            end if;
         end loop;
      end loop;
      return True;
   end Center0;


   function Err0(H: Fourier) return Boolean is
   begin
      if not Trunc then
         for M in Freq1 loop
            for N in Nmin(M,H.J) .. Nmax(M) loop
               if not Err0(H.C(M,N)) then
                  return False;
               end if;
            end loop;
         end loop;
      end if;
      return True;
   end Err0;


   procedure ErrConvert(H: in Fourier; He: out Fourier) is
   begin
      He.J := H.J;
      He.R := H.R;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            ErrConvert(H.C(M,N),He.C(M,N));
         end loop;
      end loop;
   end ErrConvert;


   function ErrConvert(H: Fourier) return Fourier is
      He: Fourier;
   begin
      ErrConvert(H,He);
      return He;
   end ErrConvert;


   procedure Get(F: in File_Type; J: in Parity; R: in Weights; H: out Fourier) is
      Sz: Integer;
      P: Taylor;
   begin
      ---- old numerical version
      SetZero(J,R,H);
      Get(F,Sz);
      SetZero(P);
      for M in -Sz .. Sz loop
         for N in Nmin(M,J) .. Sz loop
            if InRange(M,N,J) then
               Get(F,H.C(M,N));
            else
               Get(F,P);
            end if;
         end loop;
      end loop;
   end Get;


   procedure Put(F: in File_Type; J: in Parity; H: in Fourier) is
      Sz: Integer;
      P0: Taylor;
   begin                                                    ---- old numerical version
      Sz := IMax(Size1,Size2);
      SetZero(P0);
      Put(F,Sz,3);
      New_Line(F);
      for M in -Sz .. Sz loop
         for N in Nmin(M,J) .. Sz loop
            if InRange(M,N,J) then
               Put(F,H.C(M,N));
            else
               Put(F,P0);
            end if;
         end loop;
      end loop;
   end Put;


   procedure Read(Name: in String; J: in Parity; R: in Weights; H: out Fourier) is
      F: File_Type;
   begin
      Open(F,In_File,Name);
      Get(F,J,R,H);
      Close(F);
   end Read;


   procedure Write(Name: in String; J: in Parity :=0 ; H: in Fourier) is
      F: File_Type;
   begin
      Create(F,Out_File,Name);
      Put(F,J,H);
      Close(F);
   end Write;


   procedure ShortPut(I: in Integer) is
      K,D: Integer;
   begin
      K := Abs(I);
      if I<0 then D := 2; else D := 1; end if;
      while K>9 loop
         K := K/10;
         D := D+1;
      end loop;
      Put(Current_Output,I,D,10);
   end ShortPut;


   procedure ShortPut(F: in File_Type; I: in Integer) is
      K,D: Integer;
   begin
      K := Abs(I);
      if I<0 then D := 2; else D := 1; end if;
      while K>9 loop
         K := K/10;
         D := D+1;
      end loop;
      Put(F,I,D,10);
   end ShortPut;


   procedure Cshow(F: in File_Type; J: in Parity; H: in Fourier) is
   begin
      if J=0 then
         for M in Freq1 loop
            for N in Nmin(M,H.J) .. Nmax(M) loop
               Cshow(F,H.C(M,N));
               if (Deg(H.C(M,N))>=0) then
                  Put(F,"r=r+s*cos(");ShortPut(F,M);Put(F,"*q1+");ShortPut(F,N);Put(F,"*q2);");
                  New_Line(F);
               end if;
            end loop;
         end loop;
      end if;
      if J=1 then
         for M in Freq1 loop
            for N in Nmin(M,H.J) .. Nmax(M) loop
               Cshow(F,H.C(M,N));
               if (Deg(H.C(M,N))>=0) then
                  Put(F,"r=r+s*sin(");ShortPut(F,M);Put(F,"*q1+");ShortPut(F,N);Put(F,"*q2);");
                  New_Line(F);
               end if;
            end loop;
         end loop;
      end if;
   end Cshow;


   procedure Show(H: in Fourier; Cut: in Rep := Zero) is
      C: Scalar;
   begin
      ShowJR(H);
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            for K in Index3 loop
               C := Component(M,N,K,H);
               if SupAbs(C) > Cut then
                  ShowIndices(H.J,M,N,K);
                  Show(C);
               end if;
            end loop;
         end loop;
      end loop;
      New_Line;
   end Show;




   procedure Show(H1,H2: in Fourier; Cut: in Rep := Zero) is
      C1,C2: Scalar;
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.Show",Not_Implemented);
      end if;
      Put("1st "); ShowJR(H1);
      Put("2nd "); ShowJR(H2);
      if not Trunc then
         Put("Showing centers only:");
      end if;
      New_Line;
      for M in Freq1 loop
         for N in IMin(Nmin(M,H1.J),Nmin(M,H2.J)) .. Nmax(M) loop
            for K in Index3 loop
               C1 := Component(M,N,K,H1);
               C2 := Component(M,N,K,H2);
               if SupAbs(C1) > Cut or SupAbs(C2) > Cut then
                  ShowIndices(H1.J,M,N,K);
                  Put(Current_Output,C1);
                  Put(Current_Output,"   ");
                  Put(Current_Output,C2);
                  New_Line;
               end if;
            end loop;
         end loop;
      end loop;
      New_Line;
   end Show;


   function Norm(F: Scalar_Array; R3: Radius; H: Fourier) return Error is
      E: Error;
   begin
      E := Zero_Error;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            E := E+F(M,N)*Norm(R3,H.C(M,N));
         end loop;
      end loop;
      return E;
   end Norm;


   function SupNorm(F: Scalar_Array; R3: Radius; H: Fourier) return Error is
      E,Er: Error;
   begin
      Er := Zero_Error;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            E := F(M,N)*SupNorm(R3,H.C(M,N));
            if Sup(E)>Sup(Er) then Er:=E; end if;
         end loop;
      end loop;
      return Er;
   end SupNorm;


   function Norm(R: Weights; H: Fourier) return Error is
      F: Scalar_Array;
   begin
      NormFactors(R,F);
      return Norm(F,R(3),H);
   end Norm;


   function SupNorm(R: Weights; H: Fourier) return Error is
      F: Scalar_Array;
   begin
      NormFactors(R,F);
      return SupNorm(F,R(3),H);
   end SupNorm;


   function Norm(H: Fourier) return Error is
      F: Scalar_Array;
   begin
      NormFactors(H.R,F);
      return Norm(F,H.R(3),H);
   end Norm;


   function SupNorm(H: Fourier) return Error is
      F: Scalar_Array;
   begin
      NormFactors(H.R,F);
      return SupNorm(F,H.R(3),H);
   end SupNorm;


   function MaxAbsComponent(H: Fourier) return Rep is
      R: Rep;
   begin
      R := Zero;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            MaxAbsComponent(H.C(M,N),R);
         end loop;
      end loop;
      return R;
   end MaxAbsComponent;


   procedure ShowNorms(H: in Fourier; Cut: in Rep := Zero) is
      FK,Emnk,E: Error;
      C: Scalar;
      F: Scalar_Array;
   begin
      NormFactors(H.R,F);
      ShowJR(H);
      if not Trunc then
         Put_Line("Showing centers and norms only:");
      end if;
      E := Zero_Error;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            FK := Unit_Error;
            for K in Index3 loop
               C := Component(M,N,K,H);
               Emnk := F(M,N)*FK*C;
               if Sup(Emnk) > Cut then
                  ShowIndices(H.J,M,N,K);
                  Put(Current_Output,C);
                  Put("   contributes ");
                  Put(Current_Output,Sup(Emnk),3,3);
                  New_Line;
               end if;
               E := E+Emnk;
               FK := H.R(3)*FK;
            end loop;
         end loop;
      end loop;
      Put("Norm: ");
      Put(Current_Output,Sup(E));
      New_Line;
   end ShowNorms;


   procedure ShowLogNorms(H: in Fourier) is
      Fuzz: constant Rep := 1.0E-200;
      Dm,Dn,Dk,I: Integer;
      C: Rep;
      F: Scalar_Array;
   begin
      New_Line;
      ShowR(H.R);
      New_Line;
      EffDeg(H,Dm,Dn,Dk);
      NormFactors(H.R,F);
      Put_Line("Showing Floor(-Log(Norm(Component(M,N,H)))), or 0 if negative");
      New_Line;
      Put("  \ M");
      for M in -Dm .. Dm loop
         Put(Current_Output,M,3,10);
      end loop;
      New_Line;
      Put("  N \_");
      for M in -Dm .. Dm loop
         Put("___");
      end loop;
      New_Line;
      for N in 0 .. Dn loop
         Put(Current_Output,N,3,10);
         Put(" |");
         for M in -Dm .. Dm loop
            if Nmin(M,H.J)>N or Nmax(M)<N then
               Put("   ");
            else
               C := Sup(Norm(H.R(3),H.C(M,N))*F(M,N));
               I := Integer(Rep'Floor(-Log(C+Fuzz)/Log(1.0E+1)));
               if I<100 then
                  Put(Current_Output,Imax(I,0),3,10);
               else
                  Put("  *");
               end if;
            end if;
         end loop;
         New_Line;
      end loop;
      New_Line;
   end ShowLogNorms;


   procedure Copy(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Copy(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end Copy;


   procedure OddEliminate(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            OddEliminate(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end OddEliminate;


   procedure EvenEliminate(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            EvenEliminate(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end EvenEliminate;


   procedure Neg(H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Neg(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end Neg;


   function "-"(H: Fourier) return Fourier is
      Ht: Fourier;
   begin
      Neg(H,Ht);
      return Ht;
   end "-";


   procedure Mult(S: in Scalar; H: in out Fourier) is
   begin
      if S = ScalZero then
         SetZero(H.J,H.R,H);
      else
         for M in Freq1 loop
            for N in Nmin(M,H.J) .. Nmax(M) loop
               Mult(S,H.C(M,N));
            end loop;
         end loop;
      end if;
   end Mult;


   procedure Prod(S: in Scalar; H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      if S = ScalZero then
         null;
      else
         for M in Freq1 loop
            for N in Nmin(M,H1.J) .. Nmax(M) loop
               Prod(S,H1.C(M,N),H2.C(M,N));
            end loop;
         end loop;
      end if;
   end Prod;


   function "*"(S: Scalar; H: Fourier) return Fourier is
      Ht: Fourier;
   begin
      Prod(S,H,Ht);
      return Ht;
   end "*";


   procedure Add(H1: in Fourier; H2: in out Fourier) is
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.Add",Parity_Error);
      end if;
      H2.R := Min(H1.R,H2.R);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Add(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end Add;


   procedure Add(J: in Parity; H1: in Fourier; H2: in out Fourier) is
   begin
      Add(H1,H2);
      H2.J:=J;
   end Add;


   procedure Sum(H1,H2: in Fourier; H3: out Fourier) is
   begin
      SetZero(H1.J,Min(H1.R,H2.R),H3);
      if H1.J /= H2.J then
         Message("Fourier_Ops.Sum",Parity_Error);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Sum(H1.C(M,N),H2.C(M,N),H3.C(M,N));
         end loop;
      end loop;
   end Sum;


   function "+"(H1,H2: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Sum(H1,H2,Hr);
      return Hr;
   end "+";

   procedure Sub(H1: in Fourier; H2: in out Fourier) is
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.Sub",Parity_Error);
      end if;
      H2.R := Min(H1.R,H2.R);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Sub(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end Sub;


   procedure Diff(H1,H2: in Fourier; H3: out Fourier) is
   begin
      SetZero(H1.J,Min(H1.R,H2.R),H3);
      if H1.J /= H2.J then
         Message("Fourier_Ops.Diff",Parity_Error);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Diff(H1.C(M,N),H2.C(M,N),H3.C(M,N));
         end loop;
      end loop;
   end Diff;


   function "-"(H1,H2: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Diff(H1,H2,Hr);
      return Hr;
   end "-";


   procedure MultAdd(S: in Scalar; H1: in Fourier; H2: in out Fourier) is
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.MultAdd",Parity_Error);
      end if;
      if S /= ScalZero then
         H2.R := Min(H1.R,H2.R);
         for M in Freq1 loop
            for N in Nmin(M,H1.J) .. Nmax(M) loop
               MultAdd(S,H1.C(M,N),H2.C(M,N));
            end loop;
         end loop;
      end if;
   end MultAdd;


   procedure MultAdd(R: in Weights; P: in Taylor; H1: in Fourier; H2: in out Fourier; D: in Degrees := Dmax) is
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.MultAdd",Parity_Error);
      end if;
      if Deg(P) >= 0 then
         H2.R := Min(R,Min(H1.R,H2.R));
         for M in Freq1 loop
            for N in Nmin(M,H1.J) .. Nmax(M) loop
               MultAdd(H2.R(3),P,H1.C(M,N),H2.C(M,N),Dmax(M,N));
            end loop;
         end loop;
      end if;
   end MultAdd;


   procedure ZcAdd(I: in Index3; H1: in Fourier; H2: in out Fourier) is
   begin
      if H1.J /= H2.J then
         Message("Fourier_Ops.ZcAdd",Parity_Error);
      end if;
      if not Trunc and then H1.R /= H2.R then
         Message("Fourier_Ops.ZcAdd",Domain_Violation);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            AddComponent(0,Component(I,H1.C(M,N)),H2.C(M,N));
         end loop;
      end loop;
   end ZcAdd;


   procedure CollectErr(H: in out Fourier) is
      F: Scalar_Array;
   begin
      NormFactors(H.R,F);
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            MultAdd(F(M,N),ResetCenter(H.C(M,N)),H.C(0,0));
            ResetErr(H.C(M,N));
         end loop;
      end loop;
   end CollectErr;


   procedure Iminus(L: in Cutoffs; H1: in Fourier; H2: out Fourier; Collect: in Boolean := True) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            ResetCenter(H1.C(M,N),H2.C(M,N),L(M,N)+1,Pdeg);
         end loop;
      end loop;
      if Collect and not Trunc then
         CollectErr(H2);
      end if;
   end Iminus;


   procedure Iminus(L: in Cutoffs; H: in out Fourier) is
   begin
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            ResetCenter(H.C(M,N),L(M,N)+1,Pdeg);
         end loop;
      end loop;
      if not Trunc then
         CollectErr(H);
      end if;
   end Iminus;


   function Iminus(L: Cutoffs; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Iminus(L,H,Hr);
      return Hr;
   end Iminus;


   procedure Iplus(L: in Cutoffs; H1: in Fourier; H2: out Fourier; Collect: in Boolean := True) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            ResetCenter(H1.C(M,N),H2.C(M,N),0,L(M,N));
         end loop;
      end loop;
      if Collect and not Trunc then
         CollectErr(H2);
      end if;
   end Iplus;


   procedure Iplus(L: in Cutoffs; H: in out Fourier) is
   begin
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            ResetCenter(H.C(M,N),0,L(M,N));
         end loop;
      end loop;
      if not Trunc then
         CollectErr(H);
      end if;
   end Iplus;


   function Iplus(L: Cutoffs; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Iplus(L,H,Hr);
      return Hr;
   end Iplus;


   procedure Normalize(M,N: in Integer; M3,N3: out Integer; Flip,Simple: out Boolean) is
   begin
      if N<0 or else (N=0 and M<0) then
         M3 := -M;
         N3 := -N;
         Flip := True;
      else
         M3 := M;
         N3 := N;
         Flip := False;
      end if;
      Simple := M3 in Freq1 and then N3 <= Nmax(M3);
   end Normalize;


   procedure Decompose(M,N: in Integer; M1,N1,M2,N2: out Integer) is
      Flip,Simple: Boolean;
      P,Q,Mt,Nt: Integer;
   begin
      if Abs(M)*Size2 > Abs(N)*Size1 then
         P := Abs(M);
         Q := Size1;
      else
         P := Abs(N);
         Q := Size2;
      end if;
      if (Abs(M)+Abs(N))*Q > P*Qdeg then
         P := Abs(M)+Abs(N);
         Q := Qdeg;
      end if;
      Mt := (M*Q)/P;
      Nt := (N*Q)/P;
      M2 := M-Mt;
      N2 := N-Nt;
      Normalize(Mt,Nt,M1,N1,Flip,Simple);
      if not Simple then
         Message("Fourier_Ops.Decompose",Index_Error);
      end if;
   end Decompose;


   function LowFreq(M,N: Integer) return Boolean is
      Flip,Simple: Boolean;
      M3,N3: Integer;
   begin
      Normalize(4*M,4*N,M3,N3,Flip,Simple);
      return Simple;
   end LowFreq;


   function Mon(J: Parity; M: Freq1; N: Freq2; K: Index3; C: Scalar) return Monom is
   begin
      if J=1 and then M=0 and then N=0 and then not Center0(C) then
         Message("Fourier_Ops.Mon",Parity_Error);
      end if;
      return (J,M,N,K,C);
   end Mon;


   function GetParity(H: Monom) return Integer is
   begin
      return H.J;
   end GetParity;


   function GetParity(H: Fourier) return Integer is
   begin
      return H.J;
   end GetParity;


   procedure Prod(H1: in Monom; H2: in Fourier; H3: out Fourier; D: in Degrees := Dmax) is
   begin
      SetZero((H1.J+H2.J) mod 2,H2.R,H3);
      MultAdd(H1,H2,H3,D);
   end Prod;


   procedure MultAdd(S: in Scalar; M,N: in Integer; P1: in Taylor; H2: in Fourier; H3: in out Fourier; D: in Degrees := Dmax) is
      J1: constant Parity := (H2.J+H3.J) mod 2;
      Flip,Simp: Boolean;
      M1,M3,N1,N3: Integer;
      Cp,Cm1,Cm2: Scalar;
   begin
      if Deg(P1)>=0 then
         Normalize(M,N,M1,N1,Flip,Simp);           --- (M,N) need not be inrange
         if Flip and then J1=1 then
            Cp := -Scal(0.5)*S;
         else
            Cp := Scal(0.5)*S;
         end if;
         Cm1 := Cp;
         Cm2 := Cp;
         if J1 < H2.J then
            Cm1 := -Cp;
         elsif J1 > H2.J then
            Cm2 := -Cp;
         elsif J1 = 1 then
            Cp := -Cp;
         end if;

         for M2 in Freq1 loop
            for N2 in Nmin(M2,H2.J) .. Nmax(M2) loop
               if Deg(H2.C(M2,N2))>=0 then

                  Normalize(M1+M2,N1+N2,M3,N3,Flip,Simp);
                  if InRange(M3,N3,H3.J) then                                             --- frequency sum
                     MultAdd(H3.R(3),Cp*P1,H2.C(M2,N2),H3.C(M3,N3),D(M3,N3));
                  end if;

                  Normalize(M1-M2,N1-N2,M3,N3,Flip,Simp);
                  if InRange(M3,N3,H3.J) then                                             --- frequency difference
                     if not Flip then                                                 --- 1st minus 2nd frequency
                        MultAdd(H3.R(3),Cm1*P1,H2.C(M2,N2),H3.C(M3,N3),D(M3,N3));
                     else                                                             --- 2nd minus 1st frequency
                        MultAdd(H3.R(3),Cm2*P1,H2.C(M2,N2),H3.C(M3,N3),D(M3,N3));
                     end if;
                  end if;
               end if;
            end loop;
         end loop;
         if H3.J=1 then
            SetZero(H3.C(0,0));
         end if;
      end if;
   end MultAdd;


   procedure MultAdd(H1: in Monom; H2: in Fourier; H3: in out Fourier; D: in Degrees:=Dmax) is
      Cp,Cm1,Cm2: Scalar;
   begin
      if (H1.J+H2.J) mod 2 /= H3.J then
         Message("Fourier_Ops.MonMultAdd",Parity_Error);
      end if;
      if H1.C /= ScalZero then
         Cp := H1.C*ScalHalf;
         Cm1 := Cp;
         Cm2 := Cp;
         if H1.J < H2.J then
            Cm1 := -Cp;
         elsif H1.J > H2.J then
            Cm2 := -Cp;
         elsif H1.J = 1 then
            Cp := -Cp;
         end if;
         for M2 in Freq1 loop
            for N2 in Nmin(M2,H2.J) .. Nmax(M2) loop
               if InRange(H1.M+M2,H1.N+N2,H3.J) then
                  XMultAdd(H3.R(3),H1.K,Cp,H2.C(M2,N2),H3.C(H1.M+M2,H1.N+N2),D(H1.M+M2,H1.N+N2));
               end if;
               if InRange(H1.M-M2,H1.N-N2,H3.J) then
                  XMultAdd(H3.R(3),H1.K,Cm1,H2.C(M2,N2),H3.C(H1.M-M2,H1.N-N2),D(H1.M-M2,H1.N-N2));
               elsif InRange(M2-H1.M,N2-H1.N,H3.J) then
                  XMultAdd(H3.R(3),H1.K,Cm2,H2.C(M2,N2),H3.C(M2-H1.M,N2-H1.N),D(M2-H1.M,N2-H1.N));
               end if;
            end loop;
         end loop;
      end if;
   end MultAdd;


   procedure Prod(R: in Weights; H1,H2: in Fourier; H3: out Fourier; D: in Degrees := Dmax) is
   begin
      SetZero((H1.J+H2.J) mod 2,R,H3);
      for M1 in Freq1 loop
         for N1 in Nmin(M1,H1.J) .. Nmax(M1) loop
            for M2 in Freq1 loop
               for N2 in Nmin(M2,H2.J) .. Nmax(M2) loop
                  if InRange(M1+M2,N1+N2,H3.J) then
                     if H1.J = 1 and H2.J = 1 then
                        MultSub(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M1+M2,N1+N2),D(M1+M2,N1+N2) );
                     else
                        MultAdd(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M1+M2,N1+N2), D(M1+M2,N1+N2));
                     end if;
                  end if;
                  if InRange(M1-M2,N1-N2,H3.J) then
                     if H1.J = 0 and H2.J = 1 then
                        MultSub(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M1-M2,N1-N2), D(M1-M2,N1-N2));
                     else
                        MultAdd(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M1-M2,N1-N2), D(M1-M2,N1-N2) );
                     end if;
                  elsif InRange(M2-M1,N2-N1,H3.J) then
                     if H1.J = 1 and H2.J = 0 then
                        MultSub(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M2-M1,N2-N1), D(M2-M1,N2-N1) );
                     else
                        MultAdd(R(3),H1.C(M1,N1),H2.C(M2,N2),H3.C(M2-M1,N2-N1), D(M2-M1,N2-N1));
                     end if;
                  end if;
               end loop;
            end loop;
         end loop;
      end loop;
      Mult(ScalHalf,H3);
   end Prod;


   function "*"(H1,H2: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Prod(Min(H1.R,H2.R),H1,H2,Hr);
      return Hr;
   end "*";


   procedure Scale(Z,S: in Scalar; H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      H2.R(3) := Inf(Scal(H2.R(3))/Abs(S));
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Scale(Z,S,H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end Scale;


   function Scale(Z,S: Scalar; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Scale(Z,S,H,Hr);
      return Hr;
   end Scale;


   procedure Der3(R3: in Radius; H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            Der(H1.R(3),R3,H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
      H2.R := H1.R;
      H2.R(3) := R3;
   end Der3;


   function Der3(R3: Radius; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Der3(R3,H,Hr);
      return Hr;
   end Der3;


   procedure DerDer3(R3: in Radius; H1: in Fourier; H2: out Fourier) is
   begin
      SetZero(H1.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            DerDer(H1.R(3),R3,H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
      H2.R(3) := R3;
   end DerDer3;


   function DerDer3(R3: Radius; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      DerDer3(R3,H,Hr);
      return Hr;
   end DerDer3;


   procedure Comp(R3: in Radius; P: in Taylor; H1: in Fourier; H2: out Fourier) is
      D: Integer;
   begin
      if H1.J /= 0 then
         Message("Fourier_Ops.Comp",Not_Implemented);
      end if;
      if not Trunc and then Sup(Norm(H1))>R3 then
         Message("Fourier_Ops.Comp",Domain_Violation);
      end if;
      D := Deg(P);
      if D<=0 then
         SetZero(0,H1.R,H2);
         if D = 0 then
            SetComponent(0,0,0,Component(0,P),H2);
         end if;
      else
         Prod(Component(D,P),H1,H2);
         AddComponent(0,0,0,Component(D-1,P),H2);
         if D>1 then
            declare
               Hw: Fourier;
            begin
               for K in reverse 0 .. D-2 loop
                  Prod(H1.R,H1,H2,Hw);
                  AddComponent(0,0,0,Component(K,P),Hw);
                  Copy(Hw,H2);
               end loop;
            end;
         end if;
      end if;
   end Comp;


   function Comp(R3: Radius; P: Taylor; H: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Comp(R3,P,H,Hr);
      return Hr;
   end Comp;


   procedure Comp3(H,H3: in Fourier; Hr: out Fourier) is
      Hw: Fourier;
   begin
      if H3.J /= 0 then
         Message("Fourier_Ops.Comp3",Not_Implemented);
      end if;
      if not Trunc and then Sup(Norm(H3))>H.R(3) then
         Message("Fourier_Ops.Comp3",Domain_Violation);
         Put("  H.R(3) =");
         Put(Current_Output,H.R(3),5,4); New_Line;
         Put("  ||H3|| =");
         Put(Current_Output,Sup(Norm(H3)),5,4); New_Line;
      end if;
      SetZero(H.J,Min(H.R,H3.R),Hr);
      Hr.R(3) := H3.R(3);
      ZcAdd(Pdeg,H,Hr);
      for I in reverse 0 .. Pdeg-1 loop
         Prod(Hr.R,Hr,H3,Hw);
         ZcAdd(I,H,Hw);
         Copy(Hw,Hr);
      end loop;
   end Comp3;


   function Comp3(H,H3: Fourier) return Fourier is
      Hr: Fourier;
   begin
      Comp3(H,H3,Hr);
      return Hr;
   end Comp3;


   procedure CosSin(H: in Fourier; Hc,Hs: out Fourier) is
      Iter: constant Integer := 20;
      Small: constant Rep := NewCutoff;
      K: Natural;
      E,Ef: Error;
      Hw0,Hw1,Hw2: Fourier;
   begin
      K := 0;
      E := Norm(H);
      while Sup(E)>Small loop
         K := K+1;
         E := Half*E;
      end loop;
      Prod(Scal(Half**K),H,Hw0);                              ---- divide H by 2, K times
      Copy(Hw0,Hs);
      SetZero(0,H.R,Hc);
      AddComponent(0,0,0,ScalOne,Hc);
      Copy(Hw0,Hw1);
      Ef := Exp(Scal(Sup(E)))*E**(2*Iter+1);
      for I in 1 .. Iter loop                                 ---- Taylor polynomials
         Prod(H.R,Hw0,Hw1,Hw2);
         Mult(Inv(Scal(-2*I)),Hw2);
         Add(Hw2,Hc);
         Prod(H.R,Hw0,Hw2,Hw1);
         Mult(Inv(Scal(2*I+1)),Hw1);
         Add(Hw1,Hs);
         Ef := Ef/(2*I*(2*I+1));
      end loop;
      AddComponent(0,0,0,Scal(Ef),Hc);                        ---- Remainder for Cos(H)
      Ef := Ef*E/(2*Iter+2);
      if Hs.J=0 then
         AddComponent(0,0,0,Scal(Ef),Hs);                        ---- Remainder for Sin(H)
      end if;
      for I in 1 .. K loop                                    ---- double angle formula, K times
         Prod(H.R,Hs,Hs,Hw1);
         Prod(H.R,Hs,Hc,Hw2);
         Prod(ScalTwo,Hw2,Hs);
         Prod(H.R,Hc,Hc,Hw2);
         Diff(Hw2,Hw1,Hc);
      end loop;
      Hc.J:=0;
      if H.J=0
      then
         Hs.J:=0;
      else
         Hs.J:=1;
      end if;
   end CosSin;


   procedure CosSinX(H: in Fourier; Hc,Hs: out Fourier) is
      Hwc,Hws: Fourier;
   begin
      CosSin(H,Hwc,Hws); --Hwc.J=1, Hws=1 or 0
      Prod(Mon(0,1,0,0,ScalOne),Hwc,Hc);
      MultAdd(Mon(1,1,0,0,-ScalOne),Hws,Hc);
      Prod(Mon(0,1,0,0,ScalOne),Hws,Hs);
      MultAdd(Mon(1,1,0,0,ScalOne),Hwc,Hs);
   end CosSinX;


   procedure CosSinY(H: in Fourier; Hc,Hs: out Fourier) is
      Hwc,Hws: Fourier;
   begin
      CosSin(H,Hwc,Hws);
      Prod(Mon(0,0,1,0,ScalOne),Hwc,Hc);
      MultAdd(Mon(1,0,1,0,-ScalOne),Hws,Hc);
      Prod(Mon(0,0,1,0,ScalOne),Hws,Hs);
      MultAdd(Mon(1,0,1,0,ScalOne),Hwc,Hs);
   end CosSinY;


   procedure CompOmega12(R: in Weights; H,Hq: in Fourier; Hr: out Fourier) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      Hw1,Hw2: Fourier;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier; Hc2,Hs2: in out Fourier) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
      begin
         Prod(R,Hs1,Hs2,Hw2);
         Prod(R,Hc1,Hs2,Hw1);
         Prod(R,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(R,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm,Dn,Dk,Repeat: Integer;
      M: Freq1;
      Rd2: Rep;
      Rr: Weights;
      Ht,Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmnc,Hmns: Fourier;
   begin
      Split(H,Ht,Hw1);
      Rr := R;
      if not Trunc then
         Repeat := 2;
         while Repeat>0 loop
            Rd2 := Sup(Norm(Rr,Hq));                                      ---- delta2
            if H.R(2)>=Rr(2)+Rd2 then
               Repeat := 0;
            else
               Repeat := Repeat-1;
               Put_Line("Fourier_Ops.CompOmega12, Warning: Shrinking domain");
               Put("Rho(H)(2) = "); Put(Current_Output,H.R(2),5,3); New_Line;
               Put("   ||Hq|| = "); Put(Current_Output,Rd2,5,3);    New_Line;
               Put("old    R2 = "); Put(Current_Output,Rr(2),5,3);  New_Line;
               Rr(2) := H.R(2)-Rd2;
               Put("new    R2 = "); Put(Current_Output,Rr(2),5,3);  New_Line;
               if Repeat=0 then
                  Message("Fourier_Ops.CompOmega12",Domain_Violation);
               end if;
            end if;
         end loop;
         MultAdd(ScalOne+Scal(Rd2)/Scal(Rr(2)),Hw1,Ht);
      end if;
      SetZero((H.J+Hq.J) mod 2,Rr,Hr);
      CosSinX(Hq,Hc1,Hs1); --Hc1=COS(X+Hq), Hs1=SIN(X+Hq)
      CosSinY(-InvGold*Hq,Hc2,Hs2); --Hc2=COS(Y-InvTheta*Hq), Hs2=SIN(Y-InvTheta*Hq)
      SetZero(0,Rr,Hmc);
      SetComponent(0,0,0,ScalOne,Hmc);
      SetZero(1,Rr,Hms);
      EffDeg(Ht,Dm,Dn,Dk);
      for Mabs in 0 .. Dm loop
         if Mabs>0 then
            Inc(Rr,Hc1,Hs1,Hmc,Hms); --Hmc=COS(X+Hq), Hms=SIN(X+Hq)
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc);
               Copy(Hms,Hmns);
               EffDeg(M,Ht,Dn,Dk);
               for N in Nmin(M,H.J) .. Dn loop
                  if N > 0 then
                     Inc(Rr,Hc2,Hs2,Hmnc,Hmns); --Hmnc=COS(Y-....
                  end if;
                  if H.J=0 then
                     MultAdd(Rr,Ht.C(M,N),Hmnc,Hr);
                  else
                     MultAdd(Rr,Ht.C(M,N),Hmns,Hr);
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
   end CompOmega12;


   procedure CompOmega12(R: in Weights; H1,H2,Hq: in Fourier; Hr1,Hr2: out Fourier) is
      -- R is desired domain for Hr, Rho(Hr)(2) may be chosen smaller
      Hw1,Hw2: Fourier;
      procedure Inc(R: in Weights; Hc1,Hs1: in Fourier; Hc2,Hs2: in out Fourier) is
         -- make Hc2 := cos(H2+H1), Hs2 := sin(H2+H1)
         -- from Hc1=cos(H1), Hs1=sin(H1), Hc2=cos(H2), Hs2=sin(H2)
         -- uses Hw1, Hw2
      begin
         Prod(R,Hs1,Hs2,Hw2);
         Prod(R,Hc1,Hs2,Hw1);
         Prod(R,Hs1,Hc2,Hs2);
         Add(Hw1,Hs2);
         Prod(R,Hc1,Hc2,Hw1);
         Diff(Hw1,Hw2,Hc2);
      end Inc;
      Dm1,Dn1,Dk1,Dm2,Dn2,Dk2: Integer;
      M: Freq1;
      Hc1,Hs1,Hc2,Hs2,Hmc,Hms,Hmnc,Hmns: Fourier;
   begin
      if Hq.J /= 1 or not Trunc then
         Message("Fourier_Ops.CompOmega12",Not_Implemented);
      end if;
      SetZero(H1.J,R,Hr1);
      SetZero(H2.J,R,Hr2);
      CosSinX(Hq,Hc1,Hs1);
      CosSinY(-InvGold*Hq,Hc2,Hs2);
      SetZero(0,R,Hmc);
      SetComponent(0,0,0,ScalOne,Hmc);
      SetZero(1,R,Hms);
      EffDeg(H1,Dm1,Dn1,Dk1);
      EffDeg(H2,Dm2,Dn2,Dk2);
      for Mabs in 0 .. Imax(Dm1,Dm2) loop
         if Mabs>0 then
            Inc(R,Hc1,Hs1,Hmc,Hms);
         end if;
         for I in 0 .. 1 loop
            M := (2*I-1)*Mabs;
            Hms := -Hms;
            if M<0 or else I=1 then
               Copy(Hmc,Hmnc);
               Copy(Hms,Hmns);
               EffDeg(M,H1,Dn1,Dk1);
               EffDeg(M,H2,Dn2,Dk2);
               for N in IMin(Nmin(M,H1.J),Nmin(M,H2.J)) .. Imax(Dn1,Dn2) loop
                  if N > 0 then
                     Inc(R,Hc2,Hs2,Hmnc,Hmns);
                  end if;
                  if H1.J=0 then
                     MultAdd(R,H1.C(M,N),Hmnc,Hr1);
                  else
                     MultAdd(R,H1.C(M,N),Hmns,Hr1);
                  end if;
                  if H2.J=0 then
                     MultAdd(R,H2.C(M,N),Hmnc,Hr2);
                  else
                     MultAdd(R,H2.C(M,N),Hmns,Hr2);
                  end if;
               end loop;
            end if;
         end loop;
      end loop;
   end CompOmega12;


   procedure NumValQ0(H: in Fourier; P: out Taylor) is
   begin
      if not Trunc then
         Message("Fourier_Ops.NumValQ0",Numeric_Only);
      end if;
      if H.J = 1 then
         SetZero(P);
      else
         Copy(H.C(0,0),P);
         for M in Freq1 loop
            for N in Nmin(M,1) .. Nmax(M) loop
               Add(H.C(M,N),P);
            end loop;
         end loop;
      end if;
   end NumValQ0;


   function NumVal(X,Y,Z: Scalar; H: Fourier) return Scalar is
      MX,NY: Rep;
      S: Scalar;
   begin
      if not Trunc then
         Message("Fourier_Ops.NumVal",Numeric_Only);
      end if;
      if H.J = 0 then
         S := Val(H.R(3),Z,H.C(0,0));
      else
         S := ScalZero;
      end if;
      for M in Freq1 loop
         MX := Rep(M)*Center(X);
         for N in Nmin(M,H.J) .. Nmax(M) loop
            NY := Rep(N)*Center(Y);
            if H.J = 0 then
               S := S+Val(H.R(3),Z,H.C(M,N))*Scal(Cos(MX+NY));
            else
               S := S+Val(H.R(3),Z,H.C(M,N))*Scal(Sin(MX+NY));
            end if;
         end loop;
      end loop;
      return S;
   end NumVal;


   procedure NumInv(Err: in Radius; H1: in Fourier; H2: out Fourier) is
      --  iterate h2 -> h2*[2-h1*h2]
      E: Error;
      S: Scalar;
      F: Scalar_Array;
      Hw1,Hw2: Fourier;
   begin
      if not Trunc then
         Message("Fourier_Ops.NumInv",Numeric_Only);
      end if;
      S := Component(0,0,0,H1);
      if S /= ScalZero then
         Prod(-ScalOne/(S*S),H1,H2);
         AddComponent(0,0,0,ScalTwo/S,H2);
         NormFactors(H1.R,F);
         for I in 1 .. 20 loop
            Prod(H1.R,H1,H2,Hw1);
            AddComponent(0,0,0,-ScalOne,Hw1);
            E := Norm(F,H1.R(3),Hw1);
            AddComponent(0,0,0,-ScalOne,Hw1);
            Prod(H1.R,H2,Hw1,Hw2);
            Neg(Hw2,H2);
            if Sup(E) <= Err then
               return;
            end if;
         end loop;
         Message("Fourier_Ops.NumInv",Giving_Up);
      else
         Message("Fourier_Ops.NumInv",Division_By_Zero);
      end if;
   end NumInv;


   procedure NumInvComp3(Err: in Radius; H1,H2: in Fourier; H3: out Fourier) is
      --  iterate h3 -> h3-[h2(..,h3)-h1]/(D_3 h2)(..,h3)
      E: Error;
      R: Weights;
      F: Scalar_Array;
      H,Hd,Hd2,Hd3: Fourier;
   begin
      if not Trunc then
         Message("Fourier_Ops.NumInvComp3",Numeric_Only);
      end if;
      if H1.J /= H2.J then
         Message("Fourier_Ops.NumInvComp3",Not_Implemented);
      end if;
      R := Min(H1.R,H2.R);                                          ---- arbitrary
      NormFactors(R,F);
      SetZero(0,R,H3);
      InvComp(R(3),Err,H1.C(0,0),H2.C(0,0),H3.C(0,0));
      Der3(R(3),H2,H);
      NumInv(Err,H,Hd2);
      for I in 1 .. 100 loop
         Comp3(H2,H3,H);
         Sub(H1,H);
         Comp3(Hd2,H3,Hd);
         Prod(R,H,Hd,Hd3);
         Sub(Hd3,H3);
         E := Norm(F,R(3),Hd3);
         if Sup(E) <= Err then
            return;
         end if;
      end loop;
      Message("Fourier_Ops.NumInvComp3",Giving_Up);
   end NumInvComp3;


   procedure NumFixComp3(Err: in Radius; H: in Fourier; Hz: out Fourier) is
      H1,H2: Fourier;
   begin
      Copy(H,H2);
      AddComponent(0,0,1,ScalOne,H2);
      SetZero(0,Rho(H),H1);
      SetComponent(0,0,1,ScalOne,H1);
      NumInvComp3(Err,H1,H2,Hz);
      AddComponent(0,0,1,-ScalOne,Hz);
   end NumFixComp3;


   function NumFixComp3(Err: Radius; H: Fourier) return Fourier is
      Hz: Fourier;
   begin
      NumFixComp3(Err,H,Hz);
      return Hz;
   end NumFixComp3;


   procedure NumDer1(R1: in Radius; H1: in Fourier; H2: out Fourier) is
      S: Scalar;
   begin
      if not Trunc then
         Message("Fourier_Ops.NumDer1",Numeric_Only);
      end if;
      H2.J := 1-H1.J;
      H2.R := H1.R;
      H2.R(1) := R1;
      for M in Freq1 loop
         if H1.J = 0 then
            S := Scal(-M);
         else
            S := Scal(M);
         end if;
         for N in Nmin(M,H1.J).. Nmax(M) loop
            Prod(S,H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
      SetZero(H2.C(0,0));
   end NumDer1;


   procedure NumDer2(R2: in Radius; H1: in Fourier; H2: out Fourier) is
   begin
      if not Trunc then
         Message("Fourier_Ops.NumDer2",Numeric_Only);
      end if;
      H2.J := 1-H1.J;
      H2.R := H1.R;
      H2.R(2) := R2;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if H1.J = 0 then
               Prod(Scal(-N),H1.C(M,N),H2.C(M,N));
            else
               Prod(Scal(N),H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
      SetZero(H2.C(0,0));
   end NumDer2;


   procedure NumDqDot(R: in Weights; S1,S2: in Scalar; H1: in Fourier; H2: out Fourier) is
   begin
      if not Trunc then
         Message("Fourier_Ops.NumDqDot",Numeric_Only);
      end if;
      H2.J := 1-H1.J;
      H2.R := Min(R,H1.R);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if H1.J = 0 then
               Prod(S1*Scal(-M)+S2*Scal(-N),H1.C(M,N),H2.C(M,N));
            else
               Prod(S1*Scal(M)+S2*Scal(N),H1.C(M,N),H2.C(M,N));
            end if;
         end loop;
      end loop;
      SetZero(H2.C(0,0));
   end NumDqDot;


   procedure Gig(L: in Cutoffs; H: in out Fourier) is
      -- assumes that H is of type I^minus
   begin
      if (not Trunc) and then (not Err0(H)) then
         Message("Fourier_Ops.Gig",Not_Implemented);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               Mult((Scal(M)-InvGold*Scal(N))/(InvGold*Scal(M)+Scal(N)),H.C(M,N));
            end if;
         end loop;
      end loop;
   end Gig;


   procedure Gig(L: in Cutoffs; H1: in Fourier; H2: out Fourier) is
      -- assumes that H1 is of type I^minus
   begin
      SetZero(H1.J,H1.R,H2);
      if (not Trunc) and then (not Err0(H1)) then
         Message("Fourier_Ops.Gig",Not_Implemented);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               Prod((Scal(M)-InvGold*Scal(N))/(InvGold*Scal(M)+Scal(N)),H1.C(M,N),H2.C(M,N));
            else
               SetZero(H2.C(M,N));
            end if;
         end loop;
      end loop;
   end Gig;


   procedure Ig(L: in Cutoffs; H1: in Fourier; H2: out Fourier) is
      -- assumes that H1 is of type I^minus
   begin
      SetZero(H1.J,H1.R,H2);
      if (not Trunc) and then (not Err0(H1)) then
         Message("Fourier_Ops.Gig",Not_Implemented);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               Prod(ScalOne/(InvGold*Scal(M)+Scal(N)),H1.C(M,N),H2.C(M,N));
            else
               SetZero(H2.C(M,N));
            end if;
         end loop;
      end loop;
   end Ig;


   procedure Small_Gig(L: in Cutoffs; H1: in Fourier; H2: out Fourier) is
      -- assumes that H1 is of type I^minus
   begin
      SetZero(H1.J,H1.R,H2);
      if (not Trunc) and then (not Err0(H1)) then
         Message("Fourier_Ops.Gig",Not_Implemented);
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               H2.C(M,N):=H1.C(M,N);
            else
               SetZero(H2.C(M,N));
            end if;
         end loop;
      end loop;
   end Small_Gig;


   procedure Dig(L: in Cutoffs; H: in out Fourier) is
      -- assumes that H is of type I^minus
   begin
      if (not Trunc) and then (not Err0(H)) then
         Message("Fourier_Ops.Dig",Not_Implemented);
      end if;
      if H.J=0 then
         H.J := 1;
      else
         H.J:=0;
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               DerMultPart(Inv(InvGold*Scal(M)+Scal(N)),H.C(M,N));
            end if;
         end loop;
      end loop;
   end Dig;


   procedure Digg(L: in Cutoffs; H: in out Fourier) is
      -- assumes that H is of type I^minus
   begin
      if (not Trunc) and then (not Err0(H)) then
         Message("Fourier_Ops.Dig",Not_Implemented);
      end if;
      if H.J=0 then
         H.J := 1;
      else
         H.J:=0;
      end if;
      for M in Freq1 loop
         for N in Nmin(M,H.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               DerMultPart(Inv(InvGold*Scal(M)+Scal(N)),H.C(M,N));
            end if;
         end loop;
      end loop;
   end Digg;


   procedure Dig(L: in Cutoffs; H1: in Fourier; H2: out Fourier) is
      -- assumes that H1 is of type I^minus
   begin
      if (not Trunc) and then (not Err0(H1)) then
         Message("Fourier_Ops.Dig",Not_Implemented);
      end if;
      if H1.J=0 then
         H2.J := 1;
      else
         H2.J:=0;
      end if;
      SetZero(H2.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               DerProdPart(Inv(InvGold*Scal(M)+Scal(N)),H1.C(M,N),H2.C(M,N));
            else
               SetZero(H2.C(M,N));
            end if;
         end loop;
      end loop;
   end Dig;


   procedure Digg(L: in Cutoffs; H1: in Fourier; H2: out Fourier) is
      -- assumes that H1 is of type I^minus
   begin
      if (not Trunc) and then (not Err0(H1)) then
         Message("Fourier_Ops.Dig",Not_Implemented);
      end if;
      if H1.J=0 then
         H2.J := 1;
      else
         H2.J:=0;
      end if;
      SetZero(H2.J,H1.R,H2);
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            if L(M,N)>=0 then
               DerProdPart(Inv(InvGold*Scal(M)+Scal(N)),H1.C(M,N),H2.C(M,N));
            else
               SetZero(H2.C(M,N));
            end if;
         end loop;
      end loop;
   end Digg;


   procedure MinDeg(H1: in Fourier; H2: in out Fourier) is
   begin
      if not Trunc then
         Message("Fourier_Ops.MinDeg",Numeric_Only);
      end if;
      for M in Freq1 loop
         for N in IMin(Nmin(M,H1.J),Nmin(M,H2.J)) .. Nmax(M) loop
            MinDeg(H1.C(M,N),H2.C(M,N));
         end loop;
      end loop;
   end MinDeg;


   procedure CutDeg(R: in Weights; Cut: in Rep; H: in out Fourier) is
   begin
      if Cut>Zero then
         declare
            F: Scalar_Array;
         begin
            if not Trunc then
               Message("Fourier_Ops.CutDeg",Numeric_Only);
            end if;
            NormFactors(R,F);
            for M in Freq1 loop
               for N in Nmin(M,H.J) .. Nmax(M) loop
                  CutDeg(R(3),Cut/Center(F(M,N)),H.C(M,N));
               end loop;
            end loop;
         end;
      end if;
   end CutDeg;


   function NumIprod(R: Weights; H1,H2: Fourier) return Scalar is
      S: Scalar;
      F: Scalar_Array;
   begin
      if H1.J/=H2.J then
         Message("Fourier_Ops.NumIprod",Index_Error);
      end if;
      if not Trunc then
         Message("Fourier_Ops.NumIprod",Numeric_Only);
      end if;
      NormFactors(R,F);
      S := ScalZero;
      for M in Freq1 loop
         for N in Nmin(M,H1.J) .. Nmax(M) loop
            S := S+NumIprod(R(3),H1.C(M,N),H2.C(M,N))*F(M,N);
         end loop;
      end loop;
      return S;
   end NumIprod;


   procedure NumInormalize(R: in Weights; H: in out Fourier) is
      C,S: Scalar;
   begin
      if not Trunc then
         Message("Fourier_Ops.NumInormalize",Numeric_Only);
      end if;
      C := Component(0,0,1,H);
      S := NumIprod(R,H,H);
      if Center(C)>Zero then
         Mult(Inv(Sqrt(S)),H);
      else
         Mult(-Inv(Sqrt(S)),H);
      end if;
   end NumInormalize;


   procedure Random(Seed: in out Long_Integer; J: in Parity; R: in Weights; H: out Fourier) is
      F: Scalar_Array;
   begin
      SetZero(J,R,H);
      NormFactors(R,F);
      for M in Freq1 loop
         for N in Nmin(M,J) .. Nmax(M) loop
            Random(Dmax(M,N),Seed,R(3),H.C(M,N));
            Mult(Inv(F(M,N)),H.C(M,N));
            ResetErr(H.C(M,N));
         end loop;
      end loop;
   end Random;


   procedure RandomOdd(Seed: in out Long_Integer; J: in Parity; R: in Weights; H: out Fourier; L: in Cutoffs) is
      F: Scalar_Array;
      D: Degree;
   begin
      if J=0 then
         --Odd powers of z only
         SetZero(J,R,H);
         NormFactors(R,F);
         for M in Freq1 loop
            for N in Nmin(M,J) .. Nmax(M) loop
               D:=IMin(L(M,N),Dmax(M,N));
               RandomOdd(D,Seed,R(3),H.C(M,N));
               Mult(Inv(F(M,N)),H.C(M,N));
               ResetErr(H.C(M,N));
            end loop;
         end loop;
      end if;
      if J=1 then
         --Even powers of z only
         SetZero(J,R,H);
         NormFactors(R,F);
         for M in Freq1 loop
            for N in Nmin(M,J) .. Nmax(M) loop
               D:=IMin(L(M,N),Dmax(M,N));
               RandomEven(D,Seed,R(3),H.C(M,N));
               Mult(Inv(F(M,N)),H.C(M,N));
               ResetErr(H.C(M,N));
            end loop;
         end loop;
      end if;
      H.J:=J;
   end RandomOdd;


begin
   MaxDegs(Dmax,DmaxCount);
   SavedFactors.R := (Radius(Zero),Radius(Zero),Radius(One));
   for M in Freq1 loop
      for N in Freq2 loop
         SavedFactors.C(M,N) := ScalOne;
      end loop;
   end loop;

end Fourier_Ops;

