with Ada.Text_IO, Ada.Integer_Text_IO;
use Ada.Text_IO, Ada.Integer_Text_IO;

package body Linear_Ops is


   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;


   procedure ShortPut(F: in File_Type; I: in Integer) is
      K,D: Integer;
   begin
      K := Abs(I);
      if I<0 then D := 2; else D := 1; end if;
      while K>9 loop
         K := K/10;
         D := D+1;
      end loop;
      Put(F,I,D,10);
   end ShortPut;


   procedure BinRead(Name: in String; B: out Vector) is
      use Sequential_Rep_IO;
      F: Sequential_Rep_IO.File_Type;
      R: Rep;
   begin
      Open(F,In_File,Name);
      for I in B'Range loop
         R := Zero;
         if not End_Of_File(F) then
            Read(F,R);
         end if;
         B(I) := R;
      end loop;
      Close(F);
   end BinRead;


   procedure TextRead(Name: in String; B: out Vector) is
      F: File_Type;
      R: Rep;
   begin
      Open(F,In_File,Name);
      for I in B'Range loop
         R := Zero;
         if not End_Of_File(F) then
            Get(F,R);
         end if;
         B(I) := R;
      end loop;
      Close(F);
   end TextRead;


   procedure Read(Name: in String; B: out Vector; Bin: in Boolean := False) is
   begin
      if Bin then
         BinRead(Name,B);
      else
         TextRead(Name,B);
      end if;
   end Read;


   procedure BinRead(Name: in String; A: out Matrix) is
      use Sequential_Rep_IO;
      F: Sequential_Rep_IO.File_Type;
      R: Rep;
   begin
      Open(F,In_File,Name);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Zero;
            if not End_Of_File(F) then
               Read(F,R);
            end if;
            A(I,J) := R;
         end loop;
      end loop;
      Close(F);
   end BinRead;


   procedure TextRead(Name: in String; A: out Matrix) is
      F: File_Type;
      R: Rep;
   begin
      Open(F,In_File,Name);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Zero;
            if not End_Of_File(F) then
               Get(F,R);
            end if;
            A(I,J) := R;
         end loop;
      end loop;
      Close(F);
   end TextRead;


   procedure Read(Name: in String; A: out Matrix; Bin: in Boolean := False) is
   begin
      if Bin then
         BinRead(Name,A);
      else
         TextRead(Name,A);
      end if;
   end Read;


   procedure BinRead(Name: in String; W: out Sparse) is
      use Sequential_Trip_IO;
      F: Sequential_Trip_IO.File_Type;
      U: Trip;
   begin
      SetZero(W);
      Open(F,In_File,Name);
      while not End_Of_File(F) loop
         Read(F,U);
         AddComponent(U,W);
      end loop;
      Close(F);
   end BinRead;


   procedure TextRead(Name: in String; W: out Sparse) is
      F: File_Type;
      U: Trip;
   begin
      SetZero(W);
      Open(F,In_File,Name);
      while not End_Of_File(F) loop
         Get(F,U.I);
         Get(F,U.J);
         Get(F,U.R);
         AddComponent(U,W);
      end loop;
      Close(F);
   end TextRead;


   procedure Read(Name: in String; W: out Sparse; Bin: in Boolean := False) is
   begin
      if Bin then
         BinRead(Name,W);
      else
         TextRead(Name,W);
      end if;
   end Read;


   procedure BinWrite(Name: in String; B: in Vector) is
      use Sequential_Rep_IO;
      F: Sequential_Rep_IO.File_Type;
   begin
      Create(F,Out_File,Name);
      for I in B'Range loop
         Write(F,B(I));
      end loop;
      Close(F);
   end BinWrite;


   procedure TextWrite(Name: in String; B: in Vector) is
      F: File_Type;
   begin
      Create(F,Out_File,Name);
      for I in B'Range loop
         Put(F,B(I));
         New_Line(F);
      end loop;
      Close(F);
   end TextWrite;


   procedure Write(Name: in String; B: in Vector; Bin: in Boolean := False) is
   begin
      if Bin then
         BinWrite(Name,B);
      else
         TextWrite(Name,B);
      end if;
   end Write;


   procedure BinWrite(Name: in String; A: in Matrix) is
      use Sequential_Rep_IO;
      F: Sequential_Rep_IO.File_Type;
   begin
      Create(F,Out_File,Name);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Write(F,A(I,J));
         end loop;
      end loop;
      Close(F);
   end BinWrite;


   procedure TextWrite(Name: in String; A: in Matrix) is
      F: File_Type;
   begin
      Create(F,Out_File,Name);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Put(F,A(I,J));
            New_Line(F);
         end loop;
      end loop;
      Close(F);
   end TextWrite;


   procedure Write(Name: in String; A: in Matrix; Bin: in Boolean := False) is
   begin
      if Bin then
         BinWrite(Name,A);
      else
         TextWrite(Name,A);
      end if;
   end Write;


   procedure BinWrite(Name: in String; W: in Sparse) is
      use Sequential_Trip_IO;
      F: Sequential_Trip_IO.File_Type;
   begin
      Create(F,Out_File,Name);
      for K in 1 .. Length(W) loop
         Write(F,W(K));
      end loop;
      Close(F);
   end BinWrite;


   procedure TextWrite(Name: in String; W: in Sparse) is
      F: File_Type;
   begin
      Create(F,Out_File,Name);
      for K in 1 .. Length(W) loop
         ShortPut(F,W(K).I);
         Put(F," ");
         ShortPut(F,W(K).J);
         if W(K).R<Zero then
            Put(F," ");
         end if;
         Put(F,W(K).R);
         New_Line(F);
      end loop;
      Close(F);
   end TextWrite;


   procedure Write(Name: in String; W: in Sparse; Bin: in Boolean := False) is
   begin
      if Bin then
         BinWrite(Name,W);
      else
         TextWrite(Name,W);
      end if;
   end Write;


   procedure Show(P: in Pivot; L: in Integer := Integer'Last) is
   begin
      for I in P'First .. Imin(P'Last,L) loop
         Put("P(");
         ShortPut(Current_Output,I);
         Put(") = ");
         ShortPut(Current_Output,P(I));
         New_Line;
      end loop;
      New_Line;
   end Show;


   procedure Show(B: in Vector; Cut: in Rep := Zero) is
   begin
      for I in B'Range loop
         --      if Abs(B(I))>Cut then
         Put("B(");
         ShortPut(Current_Output,I);
         Put(") = ");
         Put(Current_Output,B(I));
         New_Line;
         --      end if;
      end loop;
      New_Line;
   end Show;


   procedure Show(B1,B2: in Vector; Cut: in Rep := Zero) is
   begin
      for I in B2'Range loop
         if Abs(B1(I))>Cut or Abs(B2(I))>Cut then
            Put("(");
            ShortPut(Current_Output,I);
            Put(")  ");
            Put(Current_Output,B1(I));
            Put("   ");
            Put(Current_Output,B2(I));
            New_Line;
         end if;
      end loop;
      New_Line;
   end Show;


   procedure Show(A: in Matrix; Cut: in Rep := Zero) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            if Abs(A(I,J))>Cut then
               Put("A(");
               ShortPut(Current_Output,I);
               Put(",");
               ShortPut(Current_Output,J);
               Put(") = ");
               Put(Current_Output,A(I,J));
               New_Line;
            end if;
         end loop;
      end loop;
      New_Line;
   end Show;


   procedure ShowAll(A: in Matrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Put("A(");
            ShortPut(Current_Output,I);
            Put(",");
            ShortPut(Current_Output,J);
            Put(") = ");
            Put(Current_Output,A(I,J));
            New_Line;
         end loop;
      end loop;
      New_Line;
   end ShowAll;


   procedure Show(A1,A2: in Matrix; Cut: in Rep := Zero) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            if Abs(A1(I,J))>Cut or Abs(A2(I,J))>Cut then
               Put("(");
               ShortPut(Current_Output,I);
               Put(",");
               ShortPut(Current_Output,J);
               Put(")  ");
               Put(Current_Output,A1(I,J));
               Put("   ");
               Put(Current_Output,A2(I,J));
               New_Line;
            end if;
         end loop;
      end loop;
      New_Line;
   end Show;


   procedure Show(W: in Sparse; Cut: in Rep := Zero) is
   begin
      for K in 1 .. Length(W) loop
         if Abs(W(K).R)>Cut then
            Put("W(");
            ShortPut(Current_Output,W(K).I);
            Put(",");
            ShortPut(Current_Output,W(K).J);
            Put(") = ");
            Put(Current_Output,W(K).R);
            New_Line;
         end if;
      end loop;
      New_Line;
   end Show;


   function Length(W: in Sparse) return Integer is
   begin
      return W(0).I;
   end Length;


   procedure Component(I,J: in Integer; W: in out Sparse; R: out Rep) is
      K: Integer;
   begin
      Cleanup(W);
      K := MatchIJ((I,J,Zero),W,Length(W));
      if K=0 then
         R := Zero;
      else
         R := W(K).R;
      end if;
   end Component;


   procedure AddComponent(U: in Trip; W: in out Sparse; Cut: in Rep := Zero) is
      L: Integer;
   begin
      if Abs(U.R)>Cut then
         L := Length(W)+1;
         if L>W'Last then
            Put("no more room in AddComponent");
            raise Matrix_Full;
         end if;
         W(L) := U;
         W(0).I := L;    -- new length of W
         W(0).J := 1;    -- may need cleanup
      end if;
   end AddComponent;


   procedure AddColumn(J: in Integer; V: in Vector; W: in out Sparse; Cut: in Rep := Zero) is
   begin
      for I in V'Range loop
         AddComponent((I,J,V(I)),W,Cut);
      end loop;
   end AddColumn;


   procedure Cleanup(W: in out Sparse; Force: in Boolean := False) is
   begin
      if not Force and then W(0).J=0 then return; end if;
      if W(0).I=0 then return; end if;
      declare
         Changed: Boolean;
         L: Integer;
         Keep: array(W'Range) of Boolean;
      begin
         Sort(W,CompareIJ'Access);
         Changed := False;
         for J in reverse 1 .. Length(W) loop
            if W(J).R=Zero then
               Keep(J) := False;
               Changed := True;
            elsif J>1 and then CompareIJ(W(J),W(J-1))=0 then
               W(J-1).R := W(J-1).R+W(J).R;
               Keep(J) := False;
               Changed := True;
            else
               Keep(J) := True;
            end if;
         end loop;
         if Changed then
            L := 0;
            for J in 1 .. Length(W) loop
               if Keep(J) then
                  L := L+1;
                  W(L) := W(J);
               end if;
            end loop;
            W(0).I := L;
         end if;
         W(0).J := 0;
      end;
   end Cleanup;


   procedure RowMap(W: in out Sparse; P: out Pivot; L: out Integer) is
   begin
      Cleanup(W);
      if Length(W)>0 then
         L := 1;
         P(1) := W(1).I;
         for K in 2 .. Length(W) loop
            if W(K).I>P(L) then
               L := L+1;
               P(L) := W(K).I;
            end if;
         end loop;
      else
         L := 0;
      end if;
   end RowMap;


   procedure MapIndex(P1,P2: in Pivot; W: in out Sparse) is
      -- assumes domains and ranges of P1 and P2 are ok
   begin
      for K in 1 .. Length(W) loop
         W(K).I := P1(W(K).I);
         W(K).J := P2(W(K).J);
      end loop;
      W(0).J := 1;    -- may need cleanup
   end MapIndex;


   procedure InvMapIndex(P1,P2: in Pivot; L1,L2: in Integer; W: in out Sparse) is
      -- assumes P1 and P2 are increasing
      L,K: Integer;
      U,V: Trip;
   begin
      Cleanup(W);
      L := Length(W);
      if L>0 then
         W(0).I := 0;
         K := 1;
         U := W(K);
         for I in 1 .. L1 loop
            for J in 1 .. L2 loop
               V := (P1(I),P2(J),Zero);
               while K<L and then CompareIJ(U,V)>0 loop
                  K := K+1;
                  U := W(K);
               end loop;
               if CompareIJ(U,V)=0 then
                  AddComponent((I,J,U.R),W);
               end if;
            end loop;
         end loop;
         if Length(W)<L then
            Put_Line("Linear_Ops.InvMapIndex: Warning - W got shorter");
         end if;
      end if;
   end InvMapIndex;


   function Norm1(B: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(B(I));
      end loop;
      return R;
   end Norm1;

   function Norm1(Bf,B: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(Bf(I)*B(I));
      end loop;
      return R;
   end Norm1;


   function Norm1(A: Matrix) return Rep is
      R,Rj: Rep;
   begin
      R := Zero;
      for J in A'Range(1) loop
         Rj := Zero;
         for I in A'Range(2) loop
            Rj := Rj+Abs(A(I,J));
         end loop;
         if Rj>R then
            R := Rj;
         end if;
      end loop;
      return R;
   end Norm1;


   function Norm1(Bf: Vector; A: Matrix) return Rep is
      R,Rj: Rep;
   begin
      R := Zero;
      for J in A'Range(1) loop
         Rj := Zero;
         for I in A'Range(2) loop
            Rj := Rj+Abs(Bf(I)*A(I,J));
         end loop;
         R := Max(R,Rj/Bf(J));
      end loop;
      return R;
   end Norm1;

   function Norm2(B: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+B(I)*B(I);
      end loop;
      return Sqrt(R);
   end Norm2;


   function IProd(Bf,B1,B2: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B2'Range loop
         R := R+(Bf(I)*Bf(I))*B1(I)*B2(I);
      end loop;
      return R;
   end IProd;


   procedure SetZero(B: out Vector) is
   begin
      for I in B'Range loop
         B(I) := Zero;
      end loop;
   end SetZero;


   procedure SetZero(A: out Matrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := Zero;
         end loop;
      end loop;
   end SetZero;


   procedure SetZero(W: out Sparse) is
   begin
      W(0) := (0,0,Zero);
   end SetZero;


   procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
         A(I,J) := V(I);
      end loop;
   end SetColumn;


   procedure Transpose(W: in out Sparse) is
      WKI: Integer;
   begin
      for K in 1 .. Length(W) loop
         WKI := W(K).I;
         W(K).I := W(K).J;
         W(K).J := WKI;
      end loop;
      W(0).J := 1;    -- may need cleanup
   end Transpose;


   procedure Copy(A: in Matrix; W: out Sparse; Cut: in Rep := Zero) is
   begin
      SetZero(W);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            AddComponent((I,J,A(I,J)),W,Cut);
         end loop;
      end loop;
   end Copy;


   procedure Copy(A: in Matrix; P1,P2: in Pivot; W: out Sparse; Cut: in Rep := Zero) is
   begin
      SetZero(W);
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            AddComponent((P1(I),P2(J),A(I,J)),W,Cut);
         end loop;
      end loop;
   end Copy;


   procedure Copy(W: in Sparse; A: out Matrix; Check: in Boolean := True) is
   begin
      SetZero(A);
      for K in 1 .. Length(W) loop
         if Check or else (W(K).I in A'Range(1) and W(K).J in A'Range(2)) then
            A(W(K).I,W(K).J) := A(W(K).I,W(K).J)+W(K).R;
         end if;
      end loop;
   end Copy;


   procedure Copy(W: in out Sparse; A: out Matrix; P: out Pivot; L: out Integer) is
      L1,L2: Integer;
      P1,P2: Pivot(P'Range);
   begin
      Transpose(W);
      RowMap(W,P2,L2);
      Transpose(W);
      RowMap(W,P1,L1);
      Merge(P1,P2,L1,L2,P,L);
      InvMapIndex(P,P,L,L,W);
      Copy(W,A);
      MapIndex(P,P,W);
   end Copy;


   procedure Copy(W: in out Sparse; A: out Matrix; P1,P2: out Pivot; L1,L2: out Integer) is
   begin
      Transpose(W);
      RowMap(W,P2,L2);
      Transpose(W);
      RowMap(W,P1,L1);
      InvMapIndex(P1,P2,L1,L2,W);
      Copy(W,A);
      MapIndex(P1,P2,W);
   end Copy;


   procedure Add(B1: in Vector; B2: in out Vector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)+B1(I);
      end loop;
   end Add;


   procedure Add(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
   end Add;


   procedure Add(R: in Rep; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
         A(I,I) := A(I,I)+R;
      end loop;
   end Add;


   procedure Sub(B1: in Vector; B2: in out Vector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)-B1(I);
      end loop;
   end Sub;


   procedure Sub(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
   end Sub;


   procedure Mult(R: in Rep; B: in out Vector) is
   begin
      for I in B'Range loop
         B(I) := R*B(I);
      end loop;
   end Mult;


   procedure Mult(R: in Rep; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := R*A(I,J);
         end loop;
      end loop;
   end Mult;


   procedure Mult(A1: in Matrix; A2: in out Matrix) is
      R: Rep;
      B: Vector(A2'Range(1));
   begin
      for J in B'Range loop
         for K in B'Range loop
            B(K) := A2(K,J);
         end loop;
         for I in B'Range loop
            R := Zero;
            for K in B'Range loop
               R := R+A1(I,K)*B(K);
            end loop;
            A2(I,J) := R;
         end loop;
      end loop;
   end Mult;


   procedure DMult(Bf: in Vector; B: in out Vector) is
   begin
      for I in B'Range loop
         B(I) := B(I)*Bf(I);
      end loop;
   end DMult;


   procedure InvDMult(Bf: in Vector; B: in out Vector) is
   begin
      for I in B'Range loop
         B(I) := B(I)/Bf(I);
      end loop;
   end InvDMult;


   procedure Cutoff(Rf: in Rep; Bf,B1: in Vector; N1,N2: out Integer; B2: out Vector) is
   begin
      N1 := 0;
      N2 := 0;
      for I in B1'Range loop
         if Abs(B1(I))>Rf*Bf(I) then
            B2(I) := B1(I);
            N1 := N1+1;
            N2 := N2+1;
         else
            B2(I) := Zero;
            if B1(I) /= Zero then
               N1 := N1+1;
            end if;
         end if;
      end loop;
   end Cutoff;


   function "-"(B: Vector) return Vector is
      Bt: Vector(B'Range);
   begin
      for I in B'Range loop
         Bt(I) := -B(I);
      end loop;
      return Bt;
   end "-";


   function "+"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)+B1(I);
      end loop;
      return B;
   end "+";


   function "+"(A1,A2: in Matrix) return Matrix is
      A: Matrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range loop
         for J in A'Range loop
            A(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
      return A;
   end "+";


   function "-"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)-B1(I);
      end loop;
      return B;
   end "-";


   function "-"(A1,A2: in Matrix) return Matrix is
      A: Matrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range loop
         for J in A'Range loop
            A(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
      return A;
   end "-";


   function "*"(B1,B2: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B2'Range loop
         R := R+B1(I)*B2(I);
      end loop;
      return R;
   end "*";


   function "*"(R: Rep; B: Vector) return Vector is
      Br: Vector(B'Range);
   begin
      for I in B'Range loop
         Br(I) := R*B(I);
      end loop;
      return Br;
   end "*";


   function "*"(R: Rep; A: Matrix) return Matrix is
      Ar: Matrix(A'Range(1),A'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Ar(I,J) := R*A(I,J);
         end loop;
      end loop;
      return Ar;
   end "*";


   function "*"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B2'Range loop
         B(I) := B1(I)*B2(I);
      end loop;
      return B;
   end "*";


   function "*"(A: Matrix; B: Vector) return Vector is
      R: Rep;
      Br: Vector(A'Range(1));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+A(I,J)*B(J);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(B: Vector; A: Matrix) return Vector is
      R: Rep;
      Br: Vector(A'Range(2));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+B(J)*A(J,I);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";

   function "*"(A1,A2: Matrix) return Matrix is
      R: Rep;
      A: Matrix(A1'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Zero;
            for K in A2'Range(1) loop
               R := R+A1(I,K)*A2(K,J);
            end loop;
            A(I,J) := R;
         end loop;
      end loop;
      return A;
   end "*";


   function "*"(W: Sparse; B: Vector) return Vector is
      Br: Vector(B'Range);
   begin
      SetZero(Br);
      for K in 1 .. Length(W) loop
         Br(W(K).I) := Br(W(K).I)+W(K).R*B(W(K).J);
      end loop;
      return Br;
   end "*";


   function "*"(B: Vector; W: Sparse) return Vector is
      Br: Vector(B'Range);
   begin
      SetZero(Br);
      for K in 1 .. Length(W) loop
         Br(W(K).J) := Br(W(K).J)+B(W(K).I)*W(K).R;
      end loop;
      return Br;
   end "*";


   procedure Balanc(A: in out Matrix) is
      -- translated from numerical recipes
      NearOne: constant Rep := 0.95;
      Radix: constant Rep := Two;
      Done: Boolean;
      N0,N: Integer;
      S,R,G,F,C,Sqrdx: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Sqrdx := Radix*Radix;
      loop
         Done := True;
         for I in N0 .. N loop
            C := Zero;
            R := Zero;
            for J in N0 .. N loop
               if J /= I then
                  C := C+Abs(A(J,I));
                  R := R+Abs(A(I,J));
               end if;
            end loop;
            if C /= Zero and then R /= Zero then
               G := R/Radix;
               F := One;
               S := C+R;
               while C < G loop
                  F := F*Radix;
                  C := C*Sqrdx;
               end loop;
               G := R*Radix;
               while C > G loop
                  F := F/Radix;
                  C := C/Sqrdx;
               end loop;
               if (C+R)/F < NearOne*S then
                  Done := False;
                  G := One/F;
                  for J in N0 .. N loop A(I,J) := A(I,J)*G; end loop;
                  for J in N0 .. N loop A(J,I) := A(J,I)*F; end loop;
               end if;
            end if;
         end loop;
         exit when Done;
      end loop;
   end Balanc;


   procedure Decomp(A: in out Matrix; Cond: out Rep; Pvt: out Pivot; Work: out Vector) is
      -- translated from numerical recipes
      N0,N,M: Integer;
      T,Anorm,Ynorm,Znorm: Rep;
   begin
      N0 := Work'First;
      N := Work'Last;

      Pvt(N) := N0;

      if N = N0 then
         Cond := One;
         if A(N0,N0) = Zero then
            Put("Singular matrix in Decomp");
            raise Matrix_Singular;
         end if;
         return;
      end if;
      Anorm := Zero;
      for J in N0 .. N loop
         T := Zero;
         for I in N0 .. N loop
            T := T+Abs(A(I,J));
         end loop;
         if T > Anorm then
            Anorm := T;
         end if;
      end loop;
      for K in N0 .. N-1 loop
         M := K;
         for I in K+1 .. N loop
            if Abs(A(I,K)) > Abs(A(M,K)) then
               M := I;
            end if;
         end loop;
         Pvt(K) := M;
         if M /= K then
            Pvt(N) := -Pvt(N);
         end if;

         T := A(M,K);
         A(M,K) := A(K,K);
         A(K,K) := T;
         if T /= Zero then
            for I in K+1 .. N loop
               A(I,K) := -A(I,K)/T;
            end loop;
            for J in K+1 .. N loop
               T := A(M,J);
               A(M,J) := A(K,J);
               A(K,J) := T;
               if T /= Zero then
                  for I in K+1 .. N loop
                     A(I,J) := A(I,J)+A(I,K)*T;
                  end loop;
               end if;
            end loop;
         end if;
      end loop;
      for K in N0 .. N loop
         T := Zero;
         for I in N0 .. K-1 loop
            T := T+A(I,K)*Work(I);
         end loop;
         if A(K,K) = Zero then
            Put("Singular matrix in Decomp");
            raise Matrix_Singular;
            return;
         end if;
         if T >= Zero then
            Work(K) := -(T+One)/A(K,K);
         else
            Work(K) := -(T-One)/A(K,K);
         end if;
      end loop;
      for K in reverse N0 .. N-1 loop
         T := Zero;
         for I in K+1 .. N loop
            T := T+A(I,K)*Work(K);
         end loop;
         Work(K) := T;
         M := Pvt(K);
         if M /= K then
            T := Work(M);
            Work(M) := Work(K);
            Work(K) := T;
         end if;
      end loop;
      Ynorm := Zero;
      for I in N0 .. N loop
         Ynorm := Ynorm+Abs(Work(I));
      end loop;
      Solve(A,Work,Pvt);
      Znorm := Zero;
      for I in N0 .. N loop
         Znorm := Znorm+Abs(Work(I));
      end loop;
      Cond := Anorm*Znorm/Ynorm;
      if Cond < One then
         Cond := One;
      end if;
   end Decomp;


   procedure Solve(A: in Matrix; B: in out Vector; Pvt: in Pivot) is
      -- solve A*B2=B1
      -- input:   A     triangularized matrix obtained from Decomp
      -- Pvt   pivot vector obtained from Decomp
      -- B  vector B1
      -- output:  B     vector B2
      N0,N,M: Integer;
      T: Rep;
   begin
      N0 := B'First;
      N := B'Last;

      for K in N0 .. N-1 loop
         M := Pvt(K);
         T := B(M);
         B(M) := B(K);
         B(K) := T;
         for I in K+1 .. N loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;

      for K in reverse N0+1 .. N loop
         B(K) := B(K)/A(K,K);
         T := -B(K);
         for I in N0 .. K-1 loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;

      B(N0) := B(N0)/A(N0,N0);
   end Solve;


   procedure Solve(W: in Sparse; B: in Vector; X: in out Vector; Rsq: out Rep; Eps: in Rep := 1.0E-5) is
      -- translated procedure Sparse from numerical recipes
      Vsize: Integer;
      Rp,Gg,Gam,Eps2,Dgg,Bsq,Anum,Aden: Rep;
      G,H,Xi,Xj: Vector(B'Range);
   begin
      Vsize := B'Last-B'First+1;
      Eps2 := Rep(Vsize)*Eps*Eps;
      for Irst in 1 .. 3 loop    -- restart at most 3 times
         Xi := W*X;
         Rp := Zero;
         Bsq := Zero;
         for J in B'Range loop
            Bsq := Bsq+B(J)*B(J);
            Xi(J) := Xi(J)-B(J);
            Rp := Rp+Xi(J)*Xi(J);
         end loop;
         G := Xi*W;
         for J in B'Range loop
            G(J) := -G(J);
            H(J) := G(J);
         end loop;
         for Chances in reverse 1 .. 100*Vsize loop
            Xi := W*H;
            Anum := Zero;
            Aden := Zero;
            for J in B'Range loop
               Anum := Anum+G(J)*H(J);
               Aden := Aden+Xi(J)*Xi(J);
            end loop;
            if Aden = Zero then
               Put("Singular matrix in Solve");
               raise Matrix_Singular;
            end if;
            Anum := Anum/Aden;
            for J in B'Range loop
               Xi(J) := X(J);
               X(J) := X(J)+Anum*H(J);
            end loop;
            Xj := W*X;
            Rsq := Zero;
            for J in B'Range loop
               Xj(J) := Xj(J)-B(J);
               Rsq := Rsq+Xj(J)*Xj(J);
            end loop;
            if Rsq=Rp or Rsq<=Bsq*Eps2 then return; end if;
            if Rsq>Rp then
               X := Xi;
               exit when True;               --- restart
            end if;
            Rp := Rsq;
            Xi := Xj*W;
            Gg := Zero;
            Dgg := Zero;
            for J in B'Range loop
               Gg := Gg+G(J)*G(J);
               Dgg := Dgg+(Xi(J)+G(J))*Xi(J);
            end loop;
            if Gg=Zero then return; end if;
            Gam := Dgg/Gg;
            for J in B'Range loop
               G(J) := -Xi(J);
               H(J) := G(J)+Gam*H(J);
            end loop;
            if Chances=1 then
               Put("Too many iterations in Solve");
               raise Getting_Tired;
            end if;
         end loop;
      end loop;
   end Solve;


   procedure DiagFirst(W: in out Sparse) is
      -- assumes diagonal of W is clean
      U: Trip;
   begin
      for K in 1 .. Length(W) loop
         while W(K).I=W(K).J and then W(K).I /= K loop
            U := W(K);
            W(K) := W(U.I);
            W(U.I) := U;
         end loop;
      end loop;
      W(0).J := 1;    -- may need cleanup
   end DiagFirst;


   procedure NeuSolve(W: in Sparse; B: in Vector; X: in out Vector; Eps: in Rep := 1.0E-8) is
      Enough: constant Integer := 100;
      N: Integer;
      S,Err: Rep;
      U: Vector(B'Range);
   begin
      Put("Linear_Ops.NeuSolve: ");
      N := 1;
      loop
         U := B;
         for K in B'Last+1 .. Length(W) loop
            U(W(K).I) := U(W(K).I)-W(K).R*X(W(K).J);
         end loop;
         Err := Zero;
         for I in B'Range loop
            S := X(I);
            X(I) := U(I)/W(I).R;
            Err := Err+Abs(X(I)-S);
         end loop;
         exit when N=Enough or Err<Eps;
         N := N+1;
      end loop;
      ShortPut(Current_Output,N);
      Put(" iterations, Err =");
      Put(Current_Output,Err,5,4);
      New_Line;
   end NeuSolve;


   procedure BCG(W: in Sparse; B: in Vector; X: in out Vector; Eps: in Rep := 1.0E-8) is
      I,Vsize: Integer;
      Rho,Rhold,Alpha,Beta,Err: Rep;
      R,Rt,P,Pt,Q,Qt: Vector(B'Range);
   begin
      Put("Linear_Ops.BCG: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Rt := R;
      I := 1;
      Rho := Zero;
      Alpha := Zero;
      loop
         Rhold := Rho;
         Rho := (Rt*R);
         if I=1 then
            P := R;
            Pt := Rt;
         else
            Beta := Rho/Rhold;
            for J in B'Range loop
               P(J) := Beta*P(J)+R(J);
               Pt(J) := Beta*Pt(J)+Rt(J);
            end loop;
         end if;
         Q := W*P;
         Qt := Pt*W;
         Alpha := Rho/(Pt*Q);
         Err := Zero;
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J);
            R(J) := R(J)-Alpha*Q(J);
            Rt(J) := Rt(J)-Alpha*Qt(J);
            Err := Err+Abs(R(J));
         end loop;
         exit when I=50*Vsize or Err<Eps;
         I := I+1;
      end loop;
      ShortPut(Current_Output,I);
      Put(" iterations, Err =");
      Put(Current_Output,Err,5,4);
      New_Line;
   end BCG;


   procedure BCGstab(W: in Sparse; B: in Vector; X: in out Vector; Eps: in Rep := 1.0E-8) is
      I,Vsize: Integer;
      Rho,Rhold,Alpha,Beta,Gamma,Err: Rep;
      R,Rt,P,Q,T: Vector(B'Range);
   begin
      Put("Linear_Ops.BCGstab: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Rt := R;
      I := 1;
      Rho := Zero;
      Alpha := Zero;
      Gamma := Zero;
      loop
         Rhold := Rho;
         Rho := (Rt*R);
         if I=1 then
            P := R;
         else
            Beta := (Rho/Rhold)*(Alpha/Gamma);
            for J in B'Range loop
               P(J) := R(J)+Beta*(P(J)-Gamma*Q(J));
            end loop;
         end if;
         Q := W*P;
         Alpha := Rho/(Rt*Q);
         for J in B'Range loop
            R(J) := R(J)-Alpha*Q(J);
         end loop;
         T := W*R;
         Gamma := (T*R)/(T*T);
         Err := Zero;
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J)-Gamma*R(J);
            R(J) := R(J)-Gamma*T(J);
            Err := Err+Abs(R(J));
         end loop;
         exit when I=Vsize or Err<Eps;
         I := I+1;
      end loop;
      ShortPut(Current_Output,I);
      Put(" iterations, Err =");
      Put(Current_Output,Err,5,4);
      New_Line;
   end BCGstab;


   procedure BCGstab(W: in Sparse; Bf,B: in Vector; X: in out Vector; Eabs: in Rep := 1.0E-8; Erel: in Rep := 1.0E-1) is
      I,Vsize: Integer;
      Eps,Rho,Rhold,Alpha,Beta,Gamma,S,Err: Rep;
      R,Rt,P,Q,T: Vector(B'Range);
   begin
      Put("Linear_Ops.BCGstab: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Rt := R;
      I := 1;
      Eps := Min(Eabs,Erel*IProd(Bf,R,R));
      Rho := Zero;
      Alpha := Zero;
      Gamma := Zero;
      loop
         Rhold := Rho;
         Rho := IProd(Bf,Rt,R);
         if I=1 then
            P := R;
         else
            Beta := (Rho/Rhold)*(Alpha/Gamma);
            for J in B'Range loop
               P(J) := R(J)+Beta*(P(J)-Gamma*Q(J));
            end loop;
         end if;
         Q := W*P;
         Alpha := Rho/IProd(Bf,Rt,Q);
         for J in B'Range loop
            R(J) := R(J)-Alpha*Q(J);
         end loop;
         T := W*R;
         Gamma := Iprod(Bf,T,R)/IProd(Bf,T,T);
         Err := Zero;
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J)-Gamma*R(J);
            R(J) := R(J)-Gamma*T(J);
            S := Bf(J)*R(J);
            Err := Err+S*S;
         end loop;
         Err := Sqrt(Err);
         exit when I=Vsize or Err<Eps;
         I := I+1;
      end loop;
      ShortPut(Current_Output,I);
      Put(" iterations, Err =");
      Put(Current_Output,Err,5,4);
      New_Line;
   end BCGstab;


   procedure Invert(A: in out Matrix) is
      C: Rep;
      Pvt: Pivot(A'Range(1));
      B: Vector(A'Range(1));
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Decomp(Aux,C,Pvt,B);
      for J in A'Range(2) loop
         SetZero(B);
         B(J) := One;
         Solve(Aux,B,Pvt);
         SetColumn(J,B,A);
      end loop;
   end Invert;


   procedure ElmHes(A: in out Matrix) is
      -- translated from numerical recipes
      N0,N,L: Integer;
      X,Y: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);

      if N-N0>1 then
         for M in N0+1 .. N-1 loop
            X := Zero;
            L := M;
            for J in M .. N loop
               if Abs(A(J,M-1)) > Abs(X) then
                  X := A(J,M-1);
                  L := J;
               end if;
            end loop;
            if L /= M then
               for J in M-1 .. N loop
                  Y := A(L,J);
                  A(L,J) := A(M,J);
                  A(M,J) := Y;
               end loop;
               for J in N0 .. N loop
                  Y := A(J,L);
                  A(J,L) := A(J,M);
                  A(J,M) := Y;
               end loop;
            end if;
            if X /= Zero then
               for I  in M+1 .. N loop
                  Y := A(I,M-1);
                  if Y /= Zero then
                     Y := Y/X;
                     A(I,M-1) := Y;
                     for J in M .. N loop A(I,J) := A(I,J)-Y*A(M,J); end loop;
                     for J in N0 .. N loop A(J,M) := A(J,M)+Y*A(J,I); end loop;
                  end if;
               end loop;
            end if;
         end loop;
      end if;
   end ElmHes;


   procedure Hqr(A: in out Matrix; Wr,Wi: out Vector) is
      -- translated from numerical recipes

      function Sign(R,S: Rep) return Rep is
      begin
         if S<Zero then
            return Abs(R);
         else
            return -Abs(R);
         end if;
      end Sign;
      C1: constant Rep := 0.75;
      C2: constant Rep := -0.4375;
      FoundRoot: Boolean;
      N0,N,Nn,M,L,Its: Integer;
      Z,Y,X,W,V,U,T,S,R,Q,P,Anorm: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Anorm := Abs(A(N0,N0));
      for I in N0+1 .. N loop
         for J in I-1 .. N loop
            Anorm := Anorm+Abs(A(I,J));
         end loop;
      end loop;
      Nn := N;
      T := Zero;
      while Nn>=N0 loop
         FoundRoot := False;
         Its := 0;
         while not FoundRoot loop
            L := Nn;
            while L>N0 loop
               S := Abs(A(L-1,L-1))+Abs(A(L,L));
               if S=Zero then
                  S := Anorm;
               end if;
               exit when (Abs(A(L,L-1))+S) = S;
               L := L-1;
            end loop;
            X := A(Nn,Nn);
            if L=Nn then
               FoundRoot := True;
               Wr(Nn) := X+T;
               Wi(Nn) := Zero;
               Nn := Nn-1;
            else
               Y := A(Nn-1,Nn-1);
               W := A(Nn,Nn-1)*A(Nn-1,Nn);
               if L = Nn-1 then
                  FoundRoot := True;
                  P := Half*(Y-X);
                  Q := P*P+W;
                  Z := Sqrt(Abs(Q));
                  X := X+T;
                  if Q >= Zero then
                     Z := P+Sign(Z,P);
                     Wr(Nn) := X+Z;
                     Wr(Nn-1) := Wr(Nn);
                     if Z /= Zero then Wr(Nn) := X-W/Z; end if;
                     Wi(Nn) := Zero;
                     Wi(Nn-1) := Zero;
                  else
                     Wr(Nn) := X+P;
                     Wr(Nn-1) := Wr(Nn);
                     Wi(Nn) := Z;
                     Wi(Nn-1) := -Z;
                  end if;
                  Nn := Nn-2;
               else
                  if Its=10 or Its=20 then
                     T := T+X;
                     for I in N0 .. Nn loop
                        A(I,I) := A(I,I)-X;
                     end loop;
                     S := Abs(A(Nn,Nn-1))+Abs(A(Nn-1,Nn-2));
                     X := C1*S;
                     Y := X;
                     W := C2*S*S;
                  elsif Its =20000000 then
                     Put("Too many iterations in Hqr");
                     raise Getting_Tired;
                  end if;
                  Its := Its+1;
                  M := Nn-2;
                  if M>=L then
                     loop
                        Z := A(M,M);
                        R := X-Z;
                        S := Y-Z;
                        P := (R*S-W)/A(M+1,M)+A(M,M+1);
                        Q := A(M+1,M+1)-Z-R-S;
                        R := A(M+2,M+1);
                        S := Abs(P)+Abs(Q)+Abs(R);
                        P := P/S;
                        Q := Q/S;
                        R := R/S;
                        exit when M=L;
                        U := Abs(A(M,M-1))*(Abs(Q)+Abs(R));
                        V := Abs(P)*(Abs(A(M-1,M-1))+Abs(Z)+Abs(A(M+1,M+1)));
                        exit when U+V=V;
                        M := M-1;
                     end loop;
                  end if;
                  for I in M+2 .. Nn loop
                     A(I,I-2) := Zero;
                     if  I /= (M+2) then A(I,I-3) := Zero; end if;
                  end loop;
                  for K in M .. Nn-1 loop
                     if K /= M then
                        P := A(K,K-1);
                        Q := A(K+1,K-1);
                        R := Zero;
                        if K /= (Nn-1) then
                           R := A(K+2,K-1);
                        end if;
                        X := Abs(P)+Abs(Q)+Abs(R);
                        if X /= Zero then
                           P := P/X;
                           Q := Q/X;
                           R := R/X;
                        end if;
                     end if;
                     S := Sign(Sqrt(P*P+Q*Q+R*R),P);
                     if S /= Zero then
                        if K=M then
                           if L /= M then A(K,K-1) := -A(K,K-1); end if;
                        else
                           A(K,K-1) := -S*X;
                        end if;
                        P := P+S;
                        X := P/S;
                        Y := Q/S;
                        Z := R/S;
                        Q := Q/P;
                        R := R/P;
                        for J in K .. Nn loop
                           P := A(K,J)+Q*A(K+1,J);
                           if K /= (Nn-1) then
                              P := P+R*A(K+2,J);
                              A(K+2,J) := A(K+2,J)-P*Z;
                           end if;
                           A(K+1,J) := A(K+1,J)-P*Y;
                           A(K,J) := A(K,J)-P*X;
                        end loop;
                        for I in L .. Integer'Min(Nn,K+3) loop
                           P := X*A(I,K)+Y*A(I,K+1);
                           if K /= (Nn-1) then
                              P := P+Z*A(I,K+2);
                              A(I,K+2) := A(I,K+2)-P*R;
                           end if;
                           A(I,K+1) := A(I,K+1)-P*Q;
                           A(I,K) := A(I,K)-P;
                        end loop;
                     end if;
                  end loop;
               end if;
            end if;
         end loop;
      end loop;
   end Hqr;


   procedure Eigenvalues(A: in Matrix; Br,Bi: out Vector) is
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Balanc(Aux);
      ElmHes(Aux);
      Hqr(Aux,Br,Bi);
   end Eigenvalues;


   procedure Eigenvector(A: in Matrix; R: in Rep; B: out Vector; Eps: in Rep := 1.0E-8) is
      Iter: constant Integer := 20000000;
      C: Rep;
      Pvt: Pivot(B'Range);
      Bt: Vector(B'Range);
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      C := One/Sqrt(Rep(B'Last-B'First+1));
      for I in B'Range loop
         Aux(I,I) := Aux(I,I)-R;
         B(I) := C;
      end loop;
      Decomp(Aux,C,Pvt,Bt);
      for I in 1 .. Iter loop
         Bt := B;
         Solve(Aux,B,Pvt);
         C := One/Norm2(B);
         if B*Bt<Zero then C := -C; end if;
         Mult(C,B);
         Sub(B,Bt);
         if Norm2(Bt)<Eps then return; end if;
      end loop;
      Put(Current_Output,Iter,3,10);
      Put(" iterations in Eigenvector");
      raise Getting_Tired;
   end Eigenvector;


   function PlusMinusDet(A: Matrix) return Rep is
      C: Rep;
      Pvt: Pivot(A'Range(1));
      B: Vector(A'Range(1));
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Decomp(Aux,C,Pvt,B);
      C := One;
      for J in A'Range(2) loop
         C := C*Aux(J,J);
      end loop;
      return C;
   end PlusMinusDet;


   function CompareIJ(U,V: Trip) return Integer is
   begin
      if U.I<V.I then return  1; end if;
      if U.I>V.I then return -1; end if;
      if U.J<V.J then return  1; end if;
      if U.J>V.J then return -1; end if;
      return 0;
   end CompareIJ;


   function MatchIJ(U: Trip; W: Sparse; Kmax: Integer; Kmin: Integer := 1) return Integer is
      -- assumes W has been sorted using CompareIJ
      Kmid,IO: Integer;
   begin
      Kmid := (Kmin+Kmax)/2;
      IO := CompareIJ(U,W(Kmid));
      if IO=0 then return Kmid; end if;
      if Kmin<Kmid then
         if IO<0 then return MatchIJ(U,W,Kmax,Kmid+1); end if;
         if IO>0 then return MatchIJ(U,W,Kmid-1,Kmin); end if;
      end if;
      if Kmin<Kmax and then CompareIJ(U,W(Kmax))=0 then return Kmax; end if;
      return 0;
   end MatchIJ;


   procedure Sort(W: in out Sparse; Compare: in TripCompare) is
      -- translated from numerical recipes
      L,J,Ir,I: Integer;
      Rw: Trip;
   begin
      Ir := Length(W);
      L := Ir/2+1;
      loop
         if L>1 then
            L := L-1;
            Rw := W(L);
         else
            Rw := W(Ir);
            W(Ir) := W(1);
            Ir := Ir-1;
            if Ir=1 then
               W(1) := Rw;
               return;
            end if;
         end if;
         I := L;
         J := L+L;
         while J<=Ir loop
            if J<Ir and then Compare(W(J),W(J+1))>0 then
               J := J+1;
            end if;
            if Compare(Rw,W(J))>0 then
               W(I) := W(J);
               I := J;
               J := J+J;
            else
               J := Ir+1;
            end if;
         end loop;
         W(I) := Rw;
      end loop;
   end Sort;


   procedure Merge(P1,P2: in Pivot; L1,L2: in Integer; P: out Pivot; L: out Integer) is
      N1,N2: Integer;
   begin
      if L1=0 then
         P := P2;
         L := L2;
      elsif L2=0 then
         P := P1;
         L := L1;
      else
         N1 := 1;
         N2 := 1;
         L := 1;
         P(1) := IMin(P1(1),P2(1));
         while P(L)< IMax(P1(L1),P2(L2)) loop
            while N1<=L1 and then P1(N1)<=P(L) loop N1 := N1+1; end loop;
            while N2<=L2 and then P2(N2)<=P(L) loop N2 := N2+1; end loop;
            if P1(N1)>P(L) and P2(N2)>P(L) then
               L := L+1;
               P(L) := IMin(P1(N1),P2(N2));
            else
               L := L+1;
               P(L) := IMax(P1(N1),P2(N2));
            end if;
         end loop;
      end if;
   end Merge;

begin
   null;

exception
   when Matrix_Singular =>
      New_Line;
   when Getting_Tired =>
      New_Line;
   when Matrix_Full =>
      New_Line;

end Linear_Ops;

