with Ada.Integer_Text_IO, Messages,Ada.Float_Text_IO;
use Ada.Integer_Text_IO,Messages,Ada.Float_Text_IO;

package body RG2_Ops is


   ScalZero: constant Scalar := Scal(Zero);
   ScalOne: constant Scalar := Scal(One);
   ScalTwo: constant Scalar := Scal(Two);
   ScalHalf: constant Scalar := Scal(Half);
   InvTheta: constant Scalar := Theta-ScalOne;
   ThetaSq: constant Scalar := Theta+ScalOne;
   InvThetaSq: constant Scalar := InvTheta*InvTheta;
   InvThetaCb:  constant Scalar := InvTheta*InvTheta*InvTheta;
   InvTheta12: constant Scalar:=InvThetaSq*InvThetaSq*InvThetaSq*InvThetaSq*InvThetaSq*InvThetaSq;
   Theta12:  constant Scalar:=ThetaSq*ThetaSq*ThetaSq*ThetaSq*ThetaSq*ThetaSq;
   NumZero: constant Numeric := Numeric(Zero);
   NumOne: constant Numeric := Numeric(One);
   InvGolden: constant Numeric := Golden-NumOne;
   Trunc: constant Boolean := IsNumeric(ScalZero);

   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;


   function FixTrivial1(R: Weights) return Hamilton is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      H: Hamilton;
   begin
      SetZero(R,H);
      SetComponent(0,0,0,3,Scal(Cpp),H);
      return H;
   end FixTrivial1;


   function FixTrivial2(R: Weights) return Hamilton is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      H: Hamilton;

   begin
      SetZero(R,H);
      SetComponent(0,0,0,3,-Scal(Cpp),H);
      return H;
   end FixTrivial2;


   procedure THoUmat2(Err: in Radius; P: in ProjParam; Vf: in Vector; H,Hpsi: in Polynom; Hr: out Polynom; W: out Matrix) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Jin,J,It,Jt: Integer;
      S: Numeric;
      R,Ri: Weights;
      D,DD: Degrees;
      Ht1,Ht2,Ht2c,Ht2s,Ht3,Ht3c,Ht3s,Ht4,Ht4c,Ht4s,Hc1,Hc2,Hzp,Hzp1,Hzp2: Fourier2;
   begin
      Ri:=Min(Rho(H),Rho(Hpsi));
      SetZero(Ri,Ht1);
      SetZero(Ri,Ht2);
      SetZero(Ri,Ht2c);
      SetZero(Ri,Ht2s);
      SetZero(Ri,Ht3);
      SetZero(Ri,Ht3c);
      SetZero(Ri,Ht3s);
      SetZero(Ri,Ht4);
      SetZero(Ri,Ht4c);
      SetZero(Ri,Ht4s);
      SetZero(Ri,Hc1);
      SetZero(Ri,Hc2);
      SetZero(Ri,Hzp);
      SetZero(Ri,Hzp1);
      SetZero(Ri,Hzp2);

      R := Rho(Hpsi);                      --- only determines norms in iteration

      -- compute Hzp, Hzp(q,z)=z'
      Gig(P.L,Hpsi,Ht1); --fine ; Ht1 is  D1 on Psi
      NumFixComp3(Err,Ht1,Hzp); --fine ; z'=z+Hzp, solves Hzp=-Gig of Psi(.,.+Hzp)
      AddComponent(0,0,0,1,NumOne,Hzp); --fine adds z to Hzp, gives z'

      -- compute Hr, Hr=HoU_Hpsi, Hr is h^(q,z)
      Dig(P.L,Hpsi,Ht1); --fine , Ht1 is Dig on Psi
      CompOmega12(R,H,Ht1,Ht3);  --FINE  could be combined with next CompOmega12, see THoUmat3
      Diff(Ht3,Hpsi,Hc1); --fine, Hc1 is h inv hat (q,z')
      Comp3(Hc1,Hzp,Hr); --fine, h^(q,z)

      -- determine Hc2 which will multiply Hzp^{k-1} in loop
      NumDqDot(R,NumOne,-InvGolden,H,Ht3);  --fine
      CompOmega12(R,Ht3,Ht1,Ht4);
      Comp3(Ht4,Hzp,Hc2); -- the third term in the expr. for k^ (q,z)

      -- determine Hc1 which will multiply Hzp^k in loop
      Gig(P.L,Hpsi,Ht2);
      Der3(R(3),Ht2,Ht1);
      AddComponent(0,0,0,0,NumOne,Ht1);
      NumInv(Err,Ht1,Ht3);
      Neg(Ht3,Ht2); -- negative inverse part in dzeta
      Der3(R(3),Hc1,Ht1); -- Ht1 is D2 on h inv hat (q,z')
      Prod(R,Ht1,Ht2,Ht4); -- Ht4 is is the second term in the expr. for k^
      Comp3(Ht4,Hzp,Hc1); -- Hc1 - same depending on (q,z), Hc1=D2(h^)/(1+D2 D1 psi)

      --COS LOOP

      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            D(M,N) :=IMin(P.L(M,N),Dmax(M,N));
         end loop;
      end loop;

      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            DD(M,N) :=IMin(P.L(M,N),Dmax(M,N));
         end loop;
      end loop;

      -- loop over all Cos(M*x+N*y)*Hzp^K,  Hzp1=Hzp^K, Hzp2=Hzp^{K-1}
      SetZero(R,Hzp1);
      SetComponent(0,0,0,0,NumOne,Hzp1);

      for K in Index3 loop
         if K > 0 then
            Copy(Hzp1,Hzp2);
            Prod(R,Hzp,Hzp2,Hzp1);
         end if;
         J := 0;
         for M in Freq1 loop
            for N in Nmin(M,0) .. Nmax(M) loop
               if D(M,N) >= K then
                  S := (Numeric(M)-InvGolden*Numeric(N))/(InvGolden*Numeric(M)+Numeric(N));
                  Prod(S,Hc1,Ht1);
                  AddComponent(0,0,0,0,-NumOne,Ht1);--  -1 +Second Term in expression for k
                  Prod(Mon(0,M,N,0,NumOne),Ht1,Ht3c);
                  Prod(R,Ht3c,Hzp1,Ht4c,D);

                  if K > 0 then
                     S := Numeric(K)/(InvGolden*Numeric(M)+Numeric(N));
                     Prod(Mon(1,M,N,0,S),Hc2,Ht2c);
                     Prod(R,Ht2c,Hzp2,Ht3c,D);
                     Add(Ht3c,Ht4c);
                  end if;

                  Jt := J+K+1;
                  It := 0;

                  for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                     for Nt in Nmin(Mt,0) .. Nmax(Mt) loop
                        for Kt in 0 .. D(Mt,Nt) loop
                           It := It+1;
                           W(It,Jt) := Center(Component(0,Mt,Nt,Kt,Ht4c))*Vf(It)/Vf(Jt);
                        end loop;
                     end loop;
                  end loop;

                  for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                     for Nt in Nmin(Mt,1) .. Nmax(Mt) loop
                        if (Mt/=0 or Nt/=0) then
                           for Kt in 0 .. DD(Mt,Nt) loop
                              It:=It+1;
                              W(It,Jt) := Center(Component(1,Mt,Nt,Kt,Ht4c))*Vf(It)/Vf(Jt);
                           end loop;
                        end if;
                     end loop;
                  end loop;
               end if;
               J := J+D(M,N)+1; --there are D(M,N)+1 powers of z at cos
            end loop; --N loop
         end loop; --M loop
      end loop; --K loop

      Jin:=J;

      --SIN LOOP

      -- loop over all SIN(M*x+N*y)*Hzp^K,  Hzp1=Hzp^K, Hzp2=Hzp^{K-1}
      SetZero(R,Hzp1);
      SetComponent(0,0,0,0,NumOne,Hzp1);

      for K in Index3 loop
         if K > 0 then
            Copy(Hzp1,Hzp2);
            Prod(R,Hzp,Hzp2,Hzp1);
         end if;
         J:=Jin;
         for M in Freq1 loop
            for N in Nmin(M,1) .. Nmax(M) loop
               if DD(M,N) >= K then
                  S := (Numeric(M)-InvGolden*Numeric(N))/(InvGolden*Numeric(M)+Numeric(N));
                  Prod(S,Hc1,Ht1);
                  AddComponent(0,0,0,0,-NumOne,Ht1);
                  Prod(Mon(1,M,N,0,NumOne),Ht1,Ht3s);
                  Prod(R,Ht3s,Hzp1,Ht4s,D);
                  if K > 0 then
                     S := Numeric(K)/(InvGolden*Numeric(M)+Numeric(N));
                     Prod(Mon(0,M,N,0,-S),Hc2,Ht2s);
                     Prod(R,Ht2s,Hzp2,Ht3s,D);
                     Add(Ht3s,Ht4s);
                  end if;


                  if (M/=0 or  N/=0) then
                     Jt := J+K+1;
                     It := 0;
                     for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                        for Nt in Nmin(Mt,0) .. Nmax(Mt) loop
                           for Kt in 0 .. D(Mt,Nt) loop
                              It := It+1;
                              W(It,Jt) := Center(Component(0,Mt,Nt,Kt,Ht4s))*Vf(It)/Vf(Jt);
                           end loop;
                        end loop;
                     end loop;

                     for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                        for Nt in Nmin(Mt,1) .. Nmax(Mt) loop
                           if (Mt/=0 or Nt/=0) then
                              for Kt in 0 .. DD(Mt,Nt) loop
                                 It:=It+1;
                                 W(It,Jt) := Center(Component(1,Mt,Nt,Kt,Ht4s))*Vf(It)/Vf(Jt);
                              end loop;
                           end if;
                        end loop;
                     end loop;
                  end if;
               end if;

               if  (M/=0 or  N/=0) then
                  J := J+DD(M,N)+1; --there are D(M,N)+1 powers of z at sin
               end if;

            end loop; --N loop
         end loop; --M loop
      end loop; --K loop
   end THoUmat2;


   procedure THoToUmat2(Err: in Radius; P: in ProjParam; Vf: in Vector; H,Hpsi: in Polynom; Hr: out Polynom; W: out Matrix) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Jin,J,It,Jt: Integer;
      S: Numeric;
      R,Ri: Weights;
      D,DD: Degrees;
      Ht1,Ht2,Ht2c,Ht2s,Ht3,Ht3c,Ht3s,Ht4,Ht4c,Ht4s,Hc1,Hc2,Hzp,Hzp1,Hzp2: Fourier2;
   begin
      Ri:=Min(Rho(H),Rho(Hpsi));
      SetZero(Ri,Ht1);
      SetZero(Ri,Ht2);
      SetZero(Ri,Ht2c);
      SetZero(Ri,Ht2s);
      SetZero(Ri,Ht3);
      SetZero(Ri,Ht3c);
      SetZero(Ri,Ht3s);
      SetZero(Ri,Ht4);
      SetZero(Ri,Ht4c);
      SetZero(Ri,Ht4s);
      SetZero(Ri,Hc1);
      SetZero(Ri,Hc2);
      SetZero(Ri,Hzp);
      SetZero(Ri,Hzp1);
      SetZero(Ri,Hzp2);

      R := Rho(Hpsi);                      -- only determines norms in iteration

      -- compute Hzp, Hzp(q,z)=z'
      Gig(P.L,Hpsi,Ht1); --fine ; Ht1 is  D1 on Psi
      NumFixComp3(Err,Ht1,Hzp); --fine ; z'=z+Hzp, solves Hzp=-Gig of Psi(.,.+Hzp)
      AddComponent(0,0,0,1,NumOne,Hzp); --fine adds z to Hzp, gives z'

      -- compute Hr, Hr=HoU_Hpsi, Hr is h^(q,z)
      Dig(P.L,Hpsi,Ht1); --fine , Ht1 is Dig on Psi
      CompOmegaTmu0(R,H,Ht1,Ht3);  --FINE
      Diff(Ht3,Hpsi,Hc1); --fine, Hc1 is h inv hat (q,z')
      Comp3(Hc1,Hzp,Hr); --fine, h^(q,z)

      -- determine Hc2 which will multiply Hzp^{k-1} in loop
      NumDqDot(R,NumOne,-InvGolden,H,Ht3);  --fine
      CompOmegaTmu0(R,Ht3,Ht1,Ht4);
      Comp3(Ht4,Hzp,Hc2); -- the third term in the expr. for k^ (q,z)

      -- determine Hc1 which will multiply Hzp^k in loop
      Gig(P.L,Hpsi,Ht2);
      Der3(R(3),Ht2,Ht1);
      AddComponent(0,0,0,0,NumOne,Ht1);
      NumInv(Err,Ht1,Ht3);
      Neg(Ht3,Ht2); -- negative inverse part in dzeta
      Der3(R(3),Hc1,Ht1); -- Ht1 is D2 on h inv hat (q,z')
      Prod(R,Ht1,Ht2,Ht4); -- Ht4 is is the second term in the expr. for k^
      Comp3(Ht4,Hzp,Hc1); -- Hc1 - same depending on (q,z), Hc1=D2(h^)/(1+D2 D1 psi)

      --COS LOOP

      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            D(M,N) :=IMin(P.L(M,N),Dmax(M,N));
         end loop;
      end loop;

      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            DD(M,N) :=IMin(P.L(M,N),Dmax(M,N));
         end loop;
      end loop;

      -- loop over all Cos(M*x+N*y)*Hzp^K,  Hzp1=Hzp^K, Hzp2=Hzp^{K-1}
      SetZero(R,Hzp1);
      SetComponent(0,0,0,0,NumOne,Hzp1);

      for K in Index3 loop
         if K > 0 then
            Copy(Hzp1,Hzp2);
            Prod(R,Hzp,Hzp2,Hzp1);
         end if;
         J := 0;
         for M in Freq1 loop
            for N in Nmin(M,0) .. Nmax(M) loop
               if D(M,N) >= K then
                  S := (Numeric(M)-InvGolden*Numeric(N))/(Numeric(M)+Golden*Numeric(N));
                  Prod(S,Hc1,Ht1);
                  AddComponent(0,0,0,0,-NumOne,Ht1);--  -1 +Second Term in expression for k
                  Prod(Mon(0,M,N,0,NumOne),Ht1,Ht3c);
                  Prod(R,Ht3c,Hzp1,Ht4c,D);

                  if K > 0 then
                     S := Numeric(K)/(Numeric(M)+Golden*Numeric(N));
                     Prod(Mon(1,M,N,0,S),Hc2,Ht2c);
                     Prod(R,Ht2c,Hzp2,Ht3c,D);
                     Add(Ht3c,Ht4c);
                  end if;

                  Jt := J+K+1;
                  It := 0;

                  for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                     for Nt in Nmin(Mt,0) .. Nmax(Mt) loop
                        for Kt in 0 .. D(Mt,Nt) loop
                           It := It+1;
                           W(It,Jt) := Center(Component(0,Mt,Nt,Kt,Ht4c))*Vf(It)/Vf(Jt);
                        end loop;
                     end loop;
                  end loop;

                  for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                     for Nt in Nmin(Mt,1) .. Nmax(Mt) loop
                        if (Mt/=0 or Nt/=0) then
                           for Kt in 0 .. DD(Mt,Nt) loop
                              It:=It+1;
                              W(It,Jt) := Center(Component(1,Mt,Nt,Kt,Ht4c))*Vf(It)/Vf(Jt);
                           end loop;
                        end if;
                     end loop;
                  end loop;
               end if;
               J := J+D(M,N)+1; --there are D(M,N)+1 powers of z at cos
            end loop; --N loop
         end loop; --M loop
      end loop; --K loop
      Jin:=J;

      --SIN LOOP

      -- loop over all SIN(M*x+N*y)*Hzp^K,  Hzp1=Hzp^K, Hzp2=Hzp^{K-1}
      SetZero(R,Hzp1);
      SetComponent(0,0,0,0,NumOne,Hzp1);

      for K in Index3 loop
         if K > 0 then
            Copy(Hzp1,Hzp2);
            Prod(R,Hzp,Hzp2,Hzp1);
         end if;
         J:=Jin;
         for M in Freq1 loop
            for N in Nmin(M,1) .. Nmax(M) loop
               if DD(M,N) >= K then
                  S := (Numeric(M)-InvGolden*Numeric(N))/(Numeric(M)+Golden*Numeric(N));
                  Prod(S,Hc1,Ht1);
                  AddComponent(0,0,0,0,-NumOne,Ht1);
                  Prod(Mon(1,M,N,0,NumOne),Ht1,Ht3s);
                  Prod(R,Ht3s,Hzp1,Ht4s,D);
                  if K > 0 then
                     S := Numeric(K)/(Numeric(M)+Golden*Numeric(N));
                     Prod(Mon(0,M,N,0,-S),Hc2,Ht2s);
                     Prod(R,Ht2s,Hzp2,Ht3s,D);
                     Add(Ht3s,Ht4s);
                  end if;

                  if (M/=0 or  N/=0) then
                     Jt := J+K+1;
                     It := 0;
                     for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                        for Nt in Nmin(Mt,0) .. Nmax(Mt) loop
                           for Kt in 0 .. D(Mt,Nt) loop
                              It := It+1;
                              W(It,Jt) := Center(Component(0,Mt,Nt,Kt,Ht4s))*Vf(It)/Vf(Jt);
                           end loop;
                        end loop;
                     end loop;

                     for Mt in Freq1 loop                                      -- pack Ht4 into W(.,Jt)
                        for Nt in Nmin(Mt,1) .. Nmax(Mt) loop
                           if (Mt/=0 or Nt/=0) then
                              for Kt in 0 .. DD(Mt,Nt) loop
                                 It:=It+1;
                                 W(It,Jt) := Center(Component(1,Mt,Nt,Kt,Ht4s))*Vf(It)/Vf(Jt);
                              end loop;
                           end if;
                        end loop;
                     end loop;
                  end if;
               end if;

               if  (M/=0 or  N/=0) then
                  J := J+DD(M,N)+1; --there are D(M,N)+1 powers of z at sin
               end if;

            end loop; --N loop
         end loop; --M loop
      end loop; --K loop
   end THoToUmat2;


   procedure THoUmat3(Err: in Radius; P: in ProjParam; Vf: in Vector; H,Hpsi: in Polynom; Hr: out Polynom; W: out Sparse) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Least: constant Integer := 10;              -- least number of entries in each row of W
      CutH: constant Rep := 1.00E-20;              -- determines cutoff for certain Hamiltonians
      CutW: constant Rep := 4.00E-15;              -- determines cutoff for W(i,j)
      J,Jt: Integer;
      S: Numeric;
      R: Weights;
      D: Degrees;
      Ht1,Ht2,Ht2c,Ht2s,Ht3,Ht3c,Ht3s,Ht4,Ht4c,Ht4s,Hc1,Hc2,Hzp,Hzp1,Hzp2: Fourier2;
   begin
      R := Rho(Hpsi);
      -- compute Hzp, Hzp(q,z)=z'
      Gig(P.L,Hpsi,Ht2); --fine
      NumFixComp3(Err,Ht2,Hzp); --fine
      AddComponent(0,0,0,1,NumOne,Hzp); --fine

      -- prepare computation of Hr and Hc2 below
      Dig(P.L,Hpsi,Ht2); --fine
      NumDqDot(R,NumOne,-InvGolden,H,Ht1); --fine
      CompOmega12(R,H,Ht1,Ht2,Ht3,Ht4); --fine                 -- two for the price of one

      -- compute Hr, Hr=HoU_Hpsi
      Diff(Ht3,Hpsi,Hc1); --fine
      Comp3(Hc1,Hzp,Hr); --fine

      -- determine Hc2 which will multiply Hzp^{k-1} in loop
      Comp3(Ht4,Hzp,Hc2); --fine

      -- determine Hc1 which will multiply Hzp^k in loop
      Gig(P.L,Hpsi,Ht2); --fine
      Der3(R(3),Ht2,Ht1); --fine
      AddComponent(0,0,0,0,NumOne,Ht1);
      NumInv(Err,Ht1,Ht3);
      Neg(Ht3,Ht2);
      Der3(R(3),Hc1,Ht1);
      Prod(R,Ht1,Ht2,Ht4);
      Comp3(Ht4,Hzp,Hc1);

      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            D(M,N) := IMin(P.L(M,N),Dmax(M,N));
         end loop;
      end loop;

      -- multiplication is faster when H's are cut
      --     CutDeg(R,CutH,Hc1);
      --     CutDeg(R,CutH,Hc2);
      --     CutDeg(R,CutH,Hzp);

      -- loop over all Cos(M*x+N*y)*Hzp^K,  Hzp1=Hzp^K, Hzp2=Hzp^{K-1}
      SetZero(R,Hzp1);
      SetComponent(0,0,0,0,NumOne,Hzp1);
      SetZero(W);
      for K in Index3 loop
         if K > 0 then
            Copy(Hzp1,Hzp2);
            Prod(R,Hzp,Hzp2,Hzp1);
            --           CutDeg(R,CutH,Hzp1);
         end if;
         J := 0;
         for M in Freq1 loop
            for N in Nmin(M,0) .. Nmax(M) loop
               if D(M,N) >= K then
                  S := (Numeric(M)-InvGolden*Numeric(N))/(InvGolden*Numeric(M)+Numeric(N));
                  Prod(S,Hc1,Ht1);
                  AddComponent(0,0,0,0,-NumOne,Ht1);
                  Prod(Mon(0,M,N,0,NumOne),Ht1,Ht3c);
                  Prod(Mon(1,M,N,0,NumOne),Ht1,Ht3s);
                  Prod(R,Hzp1,Ht3c,Ht4c,D);
                  Prod(R,Hzp1,Ht3s,Ht4s,D);
                  if K > 0 then
                     S := Numeric(K)/(InvGolden*Numeric(M)+Numeric(N));
                     Prod(Mon(1,M,N,0,S),Hc2,Ht2c);
                     Prod(Mon(0,M,N,0,-S),Hc2,Ht2s);
                     Prod(R,Hzp2,Ht2c,Ht3c,D);
                     Prod(R,Hzp2,Ht2s,Ht3s,D);
                     Add(Ht3c,Ht4c);
                     Add(Ht3s,Ht4s);
                  end if;

                  if (M=0) and (N=0)
                  then
                     Jt := J+K+1;
                  else
                     Jt := J+2*K+1;
                  end if;
                  AddColumn(Jt,CutPack(Least,CutW,Vf,P,Ht4c),W);         -- cut and pack Ht4 into W
                  if (M/=0 or N/=0) then
                     Jt:=Jt+1;
                     AddColumn(Jt,CutPack(Least,CutW,Vf,P,Ht4s),W);         -- cut and pack Ht4 into W
                  end if;
               end if;
               if (M=0) and (N=0) then  J := J+D(M,N)+1;
               else
                  J := J+2*D(M,N)+2;
               end if;

            end loop;
         end loop;
      end loop;
   end THoUmat3;


   function CutPack(Least: Integer; Cut: Radius; Vf: Vector; P: ProjParam; H: Polynom) return Vector is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      N,Nh : Integer;
      Rf: Rep;
      V,Vh: Vector(1 .. P.Count);
   begin
      Pack(H,P,V);
      Rf := Cut;
      Cutoff(Rf,Vf,V,N,Nh,Vh);                    -- cutoff as specified
      if N<=Least then return Vh; end if;
      for I in 1 .. 10 loop                       -- decrease cutoff such that ...
         if Nh>=Least then return Vh; end if;      -- packed H has >=Least nonzero coefficients
         Rf := Half*Rf;
         Cutoff(Rf,Vf,V,N,Nh,Vh);
      end loop;
      return V;
   end CutPack;


   procedure  Pack(H: in Polynom; P: in ProjParam; VH: out Vector) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Ind: Integer;
      D: Integer;
   begin
      Ind:=0;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            D:=IMin(P.L(M,N),Dmax(M,N));
            for K in 0 .. D loop
               Ind:=Ind+1;
               VH(Ind):=Center(Component(0,M,N,K,H));
            end loop;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            D:=IMin(P.L(M,N),Dmax(M,N));
            for K in 0 .. D loop
               if (M/=0 or N/=0) then
                  Ind:=Ind+1;
                  VH(Ind):=Center(Component(1,M,N,K,H));
               end if;
            end loop;
         end loop;
      end loop;
      if Ind /= P.Count then
         Message("RG2_Ops.Pack",Index_Error);
      end if;
   end Pack;


   procedure  JoinPack(H: in Hamilton; V: out Vector) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Ind: Integer;
   begin
      SetZero(V);
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            for K in 0 .. Deg(0,M,N,H) loop
               Ind:=Join(0,M,N,K);
               V(Ind):=Center(Component(0,M,N,K,H));
            end loop;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            for K in 0 .. Deg(1,M,N,H) loop
               Ind:=Join(1,M,N,K);
               V(Ind):=Center(Component(1,M,N,K,H));
            end loop;
         end loop;
      end loop;
   end JoinPack;


   procedure  UnjoinUnpack(V: in Vector; Rfix: in Weights; H: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      I,M,N,K: Integer;
   begin
      SetZero(Rfix,H);
      for L in V'Range loop
         if (V(L)/=Rep(0.0)) then
            UnJoin(L,I,M,N,K);
            SetComponent(I,M,N,K,Scal(V(L)),H);
         end if;
      end loop;
   end UnjoinUnpack;


   procedure UnPack(R: in Weights; P: in ProjParam; VH: in Vector; H: out Polynom) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Ind: Integer;
      D: Integer;
   begin
      SetZero(R,H);
      Ind:=0;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            D:=IMin(P.L(M,N),Dmax(M,N));
            for K in 0 .. D loop
               Ind:=Ind+1;
               if VH(Ind)/=Rep(0.0) then
                  SetComponent(0,M,N,K,Scal(VH(Ind)),H);
               end if;
            end loop;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            D:=IMin(P.L(M,N),Dmax(M,N));
            for K in 0 .. D loop
               if (M/=0 or N/=0) then
                  Ind:=Ind+1;
                  if VH(Ind)/=Rep(0.0) then
                     SetComponent(1,M,N,K,Scal(VH(Ind)),H);
                  end if;
               end if;
            end loop;
         end loop;
      end loop;
      if Ind /= P.Count then
         Message("RG2_Ops.UnPack",Index_Error);
      end if;
   end UnPack;


   procedure NormFactors(P: in ProjParam; R: in Weights; Vf: out Vector) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      J: Integer;
      F3: Rep;
      F: Scalar_Array;
   begin
      NormFactors(R,F);
      J := 0;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            F3 := One;
            for K in 0 .. IMin(P.L(M,N),Dmax(M,N)) loop
               J := J+1;
               Vf(J) := Center(F(M,N))*F3;
               F3 := F3*R(3);
            end loop;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            if (M/=0 or N/=0) then
               F3 := One;
               for K in 0 .. IMin(P.L(M,N),Dmax(M,N)) loop
                  J:= J+1;
                  Vf(J) := Center(F(M,N))*F3;
                  F3 := F3*R(3);
               end loop;
            end if;
         end loop;
      end loop;
      if J /= P.Count then
         Message("RG2_Ops.NormFactors",Index_Error);
      end if;
   end NormFactors;


   procedure ShowParam(Sigma,Kappa: in Rep) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
   begin
      Put("   Size_1 ="); Put(Current_Output,Size_1); New_Line;
      Put("   Size_2 ="); Put(Current_Output,Size_2); New_Line;
      Put("    Q_Deg ="); Put(Current_Output,Q_Deg);  New_Line;
      Put("    P_Deg ="); Put(Current_Output,P_Deg);  New_Line;
      Put("    QPdeg ="); Put(Current_Output,QPdeg);  New_Line;
      Put("DmaxCount ="); Put(Current_Output,DmaxCount); New_Line;
      Put("PPi.Count ="); Put(Current_Output,IminusCount(Sigma,Kappa)); New_Line;
      Put("      Cpp ="); Put(Current_Output,Cpp);    New_Line;
      Put("    Sigma ="); Put(Current_Output,Sigma);  New_Line;
      Put("    Kappa ="); Put(Current_Output,Kappa);  New_Line;
      Put("      Tau ="); Put(Current_Output,Sigma/Kappa); New_Line;
      New_Line;
   end ShowParam;


   function IminusCountNew(Sigma,Kappa: Rep) return Integer is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Count: Integer;
   begin
      Count := 0;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            Count := Count+IMin(IminusDegNew(Sigma,Kappa,M,N),Dmax(M,N))+1;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            if (N/=0 or M/=0) then
               Count := Count+IMin(IminusDegNew(Sigma,Kappa,M,N),Dmax(M,N))+1;
            end if;
         end loop;
      end loop;
      return Count;
   end IminusCountNew;


   function IminusCount(Sigma,Kappa: Rep) return Integer is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Count: Integer;
   begin
      Count := 0;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            Count := Count+IMin(IminusDeg(Sigma,Kappa,M,N),Dmax(M,N))+1;
         end loop;
      end loop;

      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            if (N/=0 or M/=0) then
               Count := Count+IMin(IminusDeg(Sigma,Kappa,M,N),Dmax(M,N))+1;
            end if;
         end loop;
      end loop;
      return Count;
   end IminusCount;


   function IminusDeg(Sigma,Kappa: Rep; M,N: Integer; Check: in Boolean := True) return Integer is
      -- Iminus: |omega*nu|>Sigma*|Omega*nu|+Kappa*k
      D: Integer;
      Nudotw,Nudotbigw,R: Scalar;
   begin
      if M=0 and then N=0 then
         return -1;
      end if;
      Nudotw := Abs(InvTheta*Scal(M)+Scal(N));
      Nudotbigw := Abs(Scal(M)-InvTheta*Scal(N));
      R := (Nudotw-Scal(Sigma)*Nudotbigw)/Scal(Kappa);
      D := Integer(Rep'Floor(Inf(R)));
      if Check and then D<Integer(Rep'Floor(Sup(R))) then
         Message("RG2_Ops.IminusDeg",Parameter_Error);
      end if;
      return IMax(D,-1);
   end IminusDeg;


   function IminusDegNew(Sigma,Kappa: Rep; M,N: Integer; Check: in Boolean := True) return Integer is
      -- Iminus: |omega*nu|>Sigma*|Omega*nu| and |omega*nu|>Kappa*k
      D: Integer;
      Nudotw,SigNudotbigw,R: Scalar;
   begin
      if M=0 and then N=0 then
         return -1;
      end if;
      Nudotw := Abs(Scal(M)*InvTheta+Scal(N));
      SigNudotbigw := Scal(Sigma)*Abs(Scal(M)-Scal(N)*InvTheta);
      if Inf(Nudotw) <= Sup(SigNudotbigw) then
         if Check and then Sup(Nudotw) > Inf(SigNudotbigw) then
            Message("RG2_Ops.IminusDegNew",Parameter_Error);
         end if;
         return -1;
      end if;
      R := Nudotw/Scal(Kappa);
      D := Integer(Rep'Floor(Inf(R)));
      if Check and then D<Integer(Rep'Floor(Sup(R))) then
         Message("RG2_Ops.IminusDegNew",Parameter_Error);
      end if;
      return D;
   end IminusDegNew;


   procedure ShowProj(P: in ProjParam) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
   begin
      Put("  P.Sigma =");
      Put(Current_Output,P.Sigma); New_Line;
      Put("  P.Kappa =");
      Put(Current_Output,P.Kappa); New_Line;
      Put("  P.V     =");
      Put(Current_Output,P.V);     New_Line;
      Put("  P.Count =");
      Put(Current_Output,P.Count); New_Line;
      Put("H.COS:");
      New_Line;
      Put("  \ M");
      for M in Freq1 loop
         Put(Current_Output,M,3,10);
      end loop;
      New_Line;
      Put("  N \_");
      for M in Freq1 loop
         Put("___");
      end loop;
      New_Line;
      for N in Freq2 loop
         Put(Current_Output,N,3,10);
         Put(" |");
         for M in Freq1 loop
            if Nmin(M,0)>N or Nmax(M)<N then
               Put("   ");
            elsif P.L(M,N)<0 then
               Put("  +");
            else
               Put(Current_Output,P.L(M,N),3,10);
            end if;
         end loop;
         New_Line;
      end loop;
      New_Line;
      Put("H.SIN:");
      New_Line;
      Put("  \ M");
      for M in Freq1 loop
         Put(Current_Output,M,3,10);
      end loop;
      New_Line;
      Put("  N \_");
      for M in Freq1 loop
         Put("___");
      end loop;
      New_Line;
      for N in Freq2 loop
         Put(Current_Output,N,3,10);
         Put(" |");
         for M in Freq1 loop
            if Nmin(M,1)>N or Nmax(M)<N then
               Put("   ");
            elsif P.L(M,N)<0 then
               Put("  +");
            else
               Put(Current_Output,P.L(M,N),3,10);
            end if;
         end loop;
         New_Line;
      end loop;
      New_Line;
   end ShowProj;


   procedure IminusDef(Sigma,Kappa: in Rep; P: out ProjParam) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
   begin
      P.Sigma := Sigma;
      P.Kappa := Kappa;
      P.V := Zero;
      P.Count := 0;
      for M in Freq1 loop
         for N in IMin(Nmin(M,0),Nmin(M,1)) .. Nmax(M) loop
            P.L(M,N) := IminusDeg(Sigma,Kappa,M,N);
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            P.Count := P.Count+IMin(P.L(M,N),Dmax(M,N))+1;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            if (N/=0 or M/=0) then
               P.Count := P.Count+IMin(P.L(M,N),Dmax(M,N))+1;
            end if;
         end loop;
      end loop;
   end IminusDef;


   procedure IminusDefNew(Sigma,Kappa: in Rep; P: out ProjParam) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
   begin
      P.Sigma := Sigma;
      P.Kappa := Kappa;
      P.V := Zero;
      P.Count := 0;
      for M in Freq1 loop
         for N in IMin(Nmin(M,0),Nmin(M,1)) .. Nmax(M) loop
            P.L(M,N) := IminusDegNew(Sigma,Kappa,M,N);
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            P.Count := P.Count+IMin(P.L(M,N),Dmax(M,N))+1;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            if (N/=0 or M/=0) then
               P.Count := P.Count+IMin(P.L(M,N),Dmax(M,N))+1;
            end if;
         end loop;
      end loop;
   end IminusDefNew;


   function Vguess(PPi: ProjParam; SigHat,KapHat: Rep) return Rep is
      use Scalar_Fou1_Ops;
      Flip,Simple: Boolean;
      Mt,Nt,Kt: Integer;
      V: Rep;
      Nudotbigw: Scalar;
   begin
      V := 1.0E+2;
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            Normalize(N,M+N,Mt,Nt,Flip,Simple);
            if not Simple then
               Kt := PPi.L(M,N)+1;
               Nudotbigw := Abs(Scal(M)-InvTheta*Scal(N));
               V := Min(V,Sup(Scal(SigHat)*Nudotbigw+Scal(KapHat)*Scal(Kt)));
            end if;
         end loop;
      end loop;
      return V;
   end Vguess;


   procedure SKCheck(R,Rp: in Weights; Sigma,Kappa,SigHat1,KapHat1: in Rep; SigHat2,KapHat2: out Rep) is
      -- uses global variable MuMax
      use Scalar_Fou1_Ops;
      S: Scalar;
   begin
      if R(1) /= Rp(1) then
         Message("RG2_Ops.SKCheck",Not_Implemented);
      end if;
      S := Scal(Rp(2))-InvTheta*(Scal(R(2))+Scal(Sigma)*Scal(R(1)));
      SigHat2 := Inf(S);
      Put("IminusDef Rp(3):=");Put(Current_Output,Rp(3)); New_Line;
      Put("IminusDef R(3):=");Put(Current_Output,R(3)); New_Line;
      Put("Theta=");Put(Current_Output,Theta);New_Line;
      Put("MuMax=");Put(Current_Output,MuMax);New_Line;
      S := Log(Scal(Rp(3))/(Scal(R(3))*Theta*Scal(MuMax)))-InvTheta*Scal(Kappa)*Scal(R(1));
      KapHat2 := Inf(S);
      if SigHat2<SigHat1 or KapHat2<KapHat1 then
         Message("RG2_Ops.SKCheck",Domain_Violation);
      end if;
   end SKCheck;


   procedure VHatCheck(Sigma,Kappa,SigHat,KapHat,VHat1: in Rep; VHat2: out Rep) is
      use Scalar_Fou1_Ops;
      Flip,Simple: Boolean;
      Mtop,Ntop,Ktop,Mt,Nt,Msav,Nsav,Ksav: Integer;
      V: Rep;
      Ssh,Skh,Svh,Ci,B,B0,B1,A,A0,A1,Nudotbigw: Scalar;
   begin
      Ssh := Scal(SigHat);
      Skh := Scal(KapHat);
      Svh := Scal(VHat1);
      Ci := Inv(ScalOne+InvTheta*InvTheta);
      B0 := Svh/Ssh;
      B1 := -Skh/Ssh;
      A0 := Scal(Sigma)*B0;
      A1 := Scal(Sigma)*B1+Scal(Kappa);
      Ktop := Integer(Rep'Ceiling(Sup(Svh/Skh)));                 -- bound on K in Iplus intersect Jminus
      VHat2 := VHat1;
      for K in 0 .. Ktop loop
         A := A0+A1*Scal(K);
         B := B0+B1*Scal(K);
         Mtop := Integer(Rep'Ceiling(Sup(Ci*(InvTheta*A+B))));     -- bound on M in Iplus intersect Jminus (given K)
         Ntop := Integer(Rep'Ceiling(Sup(Ci*(A+InvTheta*B))));     -- bound on N in Iplus intersect Jminus (given K)
         for M in -Mtop .. Mtop loop
            for N in Nmin(M,0) .. Ntop loop
               Simple := False;
               if M in Freq1 and then N<=Nmax(M) then
                  Normalize(N,M+N,Mt,Nt,Flip,Simple);
               end if;
               if not Simple and then IminusDeg(Sigma,Kappa,M,N,False)<K then
                  Nudotbigw := Abs(Scal(M)-InvTheta*Scal(N));
                  V := Inf(Ssh*Nudotbigw+Skh*Scal(K));
                  if V<VHat2 then
                     VHat2 := V;
                     Msav := M;
                     Nsav := N;
                     Ksav := K;
                  end if;
               end if;
            end loop;
         end loop;
      end loop;
      if VHat2<Vhat1 then
         Put(" MuMax  = "); Put(Current_Output,MuMax);  New_Line;
         Put(" SigHat = "); Put(Current_Output,SigHat); New_Line;
         Put(" KapHat = "); Put(Current_Output,KapHat); New_Line;
         Put(" VHat   = "); Put(Current_Output,VHat2);  Put("   (");
         Put(Current_Output,Msav,3,10); Put(",");
         Put(Current_Output,Nsav,2,10); Put(",");
         Put(Current_Output,Ksav,2,10); Put_Line(")");
         New_Line;
      end if;
   end VhatCheck;


   function RoughPsi(P: ProjParam; H: Hamilton; R: Weights) return Polynom is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      H2: Polynom;
      H1: Hamilton;
   begin
      Shift(R,P,P,H,H1);
      H2:=Iminus(P.L,PutNumeric(H1));
      return H2;
   end RoughPsi;


   function RoughPsiNoShift(P: ProjParam; H: Hamilton; R: Weights) return Polynom is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      H2: Polynom;
   begin
      H2:=Iminus(P.L,PutNumeric(H));
      return H2;
   end RoughPsiNoShift;


   procedure PsiFind_Sparse(Err: in Radius; R: in Weights; P: in ProjParam; H: in Hamilton; Hpsi: in out Polynom; W: in out Sparse) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Eabs: constant Rep := 1.0E-15;                 --- minimal precision in Solve
      Erel: constant Rep := 1.0E-5;                 --- minimal improvement in Solve
      Foo: constant Rep := 1.0E-15;                  --- re-use W only if ||Hdpsi||<Foo
      Rf: constant Weights := (0.0,0.4,0.6);        --- weights for range of W, used also to cut matrix W
      E: Error;
      Vf,V1,V2: Vector(1 .. P.Count);
      Hhat,Hdpsi: Fourier2;
   begin
      NormFactors(P,Rf,Vf); --fine
      for K in 1 .. MaxI loop
         if Length(W)=0 then
            THoUmat3(Err,P,Vf,PutNumeric(H),Hpsi,Hhat,W);
            DiagFirst(W);
            Put("Length(W) =");
            Put(Current_Output,Length(W)); New_Line;
         else
            HouNum(Err,P,PutNumeric(H),Hpsi,Hhat);
         end if;
         ---- solve W*Hdpsi=Hhat
         Pack(Hhat,P,V1);
         V2 := -V1;
         BCGstab(W,Vf,V1,V2,Eabs,Erel);
         Unpack(Rho(Hhat),P,V2,Hdpsi);
         Sub(Hdpsi,Hpsi);
         E := Norm(R,Hdpsi);
         Put("PsiFind: ");
         Put(Current_Output,Sup(E),3,3);
         if Sup(E)<Err then
            SetRho(R,Hpsi);
            return;
         end if;
         if Sup(E)>Foo then
            SetZero(W);
         end if;
      end loop;
      Put("PsiFind is not converging well");
      New_Line;
   end PsiFind_Sparse;


   procedure PsiFind(Err: in Radius; R: in Weights; P: in ProjParam; H: in Hamilton; Hpsi: in out Polynom) is
      pragma Suppress(Storage_Check);
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      subtype Vindex is Integer range 1 .. P.Count;
      E: Rep;
      Cond: Rep;
      Hhat,Hdpsi,Hppsi: Fourier2;
      Pvt: Pivot(Vindex);
      Vf,V,VV: Vector(Vindex);
      W: Matrix(Vindex,Vindex);
      F: Scalar_Array;
   begin
      SetZero(R,Hhat);
      SetZero(R,Hppsi);
      Hpsi:=Hppsi+Hpsi;
      NormFactors(P,R,Vf);            -- used to condition W in THoUmat
      NormFactors(R,F);
      for I in 1..MaxI loop
         SetZero(R,Hdpsi);
         THoUmat2(Err,P,Vf,PutNumeric(H),Hpsi,Hhat,W);
         Decomp(W,Cond,Pvt,VV);
         ---- solve W*Hdpsi=Hhat
         Pack(Hhat,P,V);
         DMult(Vf,V);
         Solve(W,V,Pvt);
         InvDmult(Vf,V);
         Unpack(Rho(Hhat),P,V,Hdpsi);
         Sub(Hdpsi,Hpsi);
         E := Sup(Norm(F,R(3),Hdpsi));
         if E > Rep(1.0E+10) then
            EX:=1;
            Message("RG2_Ops.PsiFind",Giving_Up);
            SetRho(R,Hpsi);
            return;
         end if;
         if E < Rep(Err) then
            Iterr:=I;
            Exc:=E;
            SetRho(R,Hpsi);
            return;
         end if;
      end loop;
      Iterr:=MaxI;
      Put("PsiFind is not converging well");
      New_Line;
      Exc:=E;
      SetRho(R,Hpsi);
   end PsiFind;


   procedure PsiFind2(Err: in Radius; R: in Weights; P: in ProjParam; H: in Hamilton; Hpsi: in out Polynom) is
      pragma Suppress(Storage_Check);
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      subtype Vindex is Integer range 1 .. P.Count;
      E: Rep;
      Cond: Rep;
      Hhat,Hdpsi,Hppsi: Fourier2;
      Pvt: Pivot(Vindex);
      Vf,V,VV: Vector(Vindex);
      W: Matrix(Vindex,Vindex);
      F: Scalar_Array;
   begin
      SetZero(R,Hhat);
      SetZero(R,Hppsi);
      Hpsi:=Hppsi+Hpsi;
      NormFactors(P,R,Vf);            -- FINE, used to condition W in THoUmat
      NormFactors(R,F);
      for I in 1..MaxI loop
         SetZero(R,Hdpsi);
         THoToUmat2(Err,P,Vf,PutNumeric(H),Hpsi,Hhat,W); --FINE??
         Decomp(W,Cond,Pvt,VV);--FINE

         ---- solve W*Hdpsi=Hhat
         Pack(Hhat,P,V); --FINE
         DMult(Vf,V); --FINE
         Solve(W,V,Pvt); --FINE
         InvDmult(Vf,V); --FINE
         Unpack(Rho(Hhat),P,V,Hdpsi); --FINE
         Sub(Hdpsi,Hpsi); --FINE
         E := Sup(Norm(F,R(3),Hdpsi)); --FINE
         if E > Rep(1.0E+10) then
            EX:=1;
            Message("RG_Ops.PsiFind",Giving_Up);
            SetRho(R,Hpsi);
            return;
         end if;
         if E < Rep(Err) then
            Iterr:=I;
            Exc:=E;
            SetRho(R,Hpsi);
            return;
         end if;
      end loop;
      Iterr:=MaxI;
      Put("PsiFind2 is not converging well");
      New_Line;
      Exc:=E;
      SetRho(R,Hpsi);
   end PsiFind2;


   procedure HoU(Err: in Radius; P: in ProjParam; R: in Weights; H,Hpsi: in Hamilton; Hr: out Hamilton) is
      -- Hpsi is 1/(omega1*D_1+omega2*D2) applied to the generating function
      -- Hr(q,z)=-Hpsi(q,z')+H(q+(Dig Hpsi)(q,z')*Omega,z') where z'=z+Z(q,z)
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Rt: Weights;
      H1,Hzp: Fourier2;
   begin
      if R(3)>Rho(H)(3) then
         Message("RG2_Ops.HoU",Domain_Violation);
      end if;

      -- compute Hzp, Hzp(q,z)=z'
      Gig(P.L,Hpsi,H1); --fine
      FixComp3(Err,R(3),H1,Hzp); --fine
      AddComponent(0,0,0,1,ScalOne,Hzp); --fine

      -- compute Hr, Hr=HoU_Hpsi
      Dig(P.L,Hpsi,H1); --fine
      Rt := R;
      Rt(3) := Min(Rho(H)(3),Rho(Hpsi)(3));
      CompOmega12(Rt,H,H1,Hr); --fine
      Diff(Hr,Hpsi,H1); --fine
      Comp3(H1,Hzp,Hr); --fine
      SetComponent(0,0,0,0,ScalZero,Hr);
      if Rho(Hr) /= R then
         Put_Line("RG2_Ops.HoU, Warning: Domain changed");
         ShowPair("Rpri2",Rho(Hr)(2));
         ShowPair("Rpri3",Rho(Hr)(3));
      end if;
   end HoU;


   procedure HoToU(Err: in Radius; P: in ProjParam; R: in Weights; H,Hpsi: in Hamilton; Hr: out Hamilton) is
      -- Hpsi is (omega1*D_1+omega2*D2) applied to the generating function
      -- Hr(q,z)=-Hpsi(q,z')+H(q+(Dig Hpsi)(q,z')*Omega,z') where z'=z+Z(q,z)
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Rt: Weights;
      H1,Hzp: Fourier2;
   begin
      if R(3)>Rho(H)(3) then
         Message("RG2_Ops.HoToU",Domain_Violation);
      end if;
      -- compute Hzp, Hzp(q,z)=z'
      Gig(P.L,Hpsi,H1);
      FixComp3(Err,R(3),H1,Hzp);
      AddComponent(0,0,0,1,ScalOne,Hzp);

      -- compute Hr, Hr=HoU_Hpsi
      Dig(P.L,Hpsi,H1);
      Rt := R;
      Rt(3) := Min(Rho(H)(3),Rho(Hpsi)(3));
      CompOmegaTmu0(Rt,H,H1,Hr); --fine
      Diff(Hr,Hpsi,H1); --fine
      Comp3(H1,Hzp,Hr); --fine
      SetComponent(0,0,0,0,ScalZero,Hr);
      if Rho(Hr) /= R then
         Put_Line("RG2_Ops.HoToU, Warning: Domain changed");
         ShowPair("Rpri2",Rho(Hr)(2)); --?
         ShowPair("Rpri3",Rho(Hr)(3)); --?
      end if;
   end HoToU;


   procedure HoUnum(Err: in Radius; P: in ProjParam; H,Hpsi: in Polynom; Hr: out Polynom) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Ht1,Ht2: Polynom;
   begin
      -- compute Ht2, Ht2(q,z)=z'
      Gig(P.L,Hpsi,Ht1);
      NumFixComp3(Err,Ht1,Ht2);
      AddComponent(0,0,0,1,NumOne,Ht2);
      -- compute Hr=Iminus(H o U_Psi)
      Dig(P.L,Hpsi,Hr);
      CompOmega12(Rho(Hpsi),H,Hr,Ht1);
      Sub(Hpsi,Ht1);
      Comp3(Ht1,Ht2,Hr);
   end HoUnum;


   procedure Shift(R: in Weights; PPi,PPj: ProjParam; H1: in Hamilton; H2: out Hamilton) is
      -- note: scaling with MuMax is part of the shift
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      use Scalar_Taylor_Ops;
      Flip,Simple: Boolean;
      Mt,Nt: Integer;
      Rs,Rk:  Rep;
      Rp: Weights;
      RmuMax,S: Scalar;
      P1,P2: Scalar_Taylor_Ops.Taylor;
      H: Hamilton;
   begin
      Rp := Rho(H1);
      SetZero(R,H2);
      --    Iplus(PPi.L,H1,H,False);  --fine                                         -- H := H1, assuming type Iplus
      Copy(H1,H);
      RmuMax := Scal(MuMax);
      for M in Freq1 loop
         for N in Nmin(M,0) .. Nmax(M) loop
            Normalize(N,M+N,Mt,Nt,Flip,Simple);
            if Simple then
               ScalePlus(PPi.L(M,N),Theta/RmuMax,-Theta*RmuMax,Rp(3),R(3),Coeff(0,M,N,H),P1);
               if not Trunc then
                  ResetErr(P1,PPj.L(M,N)+1,P_Deg);                           -- consider only errors in Jminus
                  SetComponent(0,M,N,ResetCenter(Component(0,M,N,H)),H);     -- remove stuff from H that we ...
               end if;
               SetComponent(0,Mt,Nt,DegConvert(R(3),Dmax(Mt,Nt),P1),H2);      -- shift explicitly into H2
            end if;
         end loop;
      end loop;
      for M in Freq1 loop
         for N in Nmin(M,1) .. Nmax(M) loop
            Normalize(N,M+N,Mt,Nt,Flip,Simple);
            if Simple then
               ScalePlus(PPi.L(M,N),Theta/RmuMax,-Theta*RmuMax,Rp(3),R(3),Coeff(1,M,N,H),P2);
               if not Trunc then
                  ResetErr(P2,PPj.L(M,N)+1,P_Deg);
                  SetComponent(1,M,N,ResetCenter(Component(1,M,N,H)),H);
               end if;
               SetComponent(1,Mt,Nt,DegConvert(R(3),Dmax(Mt,Nt),P2),H2);
            end if;
         end loop;
      end loop;
      if not Trunc then
         SKCheck(R,Rp,PPi.Sigma,PPi.Kappa,PPj.Sigma,PPj.Kappa,Rs,Rk); -- mainly to check Rho(H1)
         S := Scal(Norm(H));                                          -- now errors of type Iplus intersect Jplus
         S := S*Ubound(R,Rp)*Exp(Scal(-PPj.V))*Theta/RmuMax;          -- which should contract
         AddComponent(0,0,0,0,S,H2);
      end if;
   end Shift;


   function Ubound(R,Rp: Weights) return Scalar is
   begin
      return Scal(102)/Scal(100)+(Theta*Scal(Rp(2))/Scal(R(2))-Scal(1))/Scal(5);
   end Ubound;


   procedure Normalize(H: in Hamilton; Hn: out Hamilton) is
      --NON-TWIST NORMALIZE
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      C2,R: Scalar;
   begin
      C2 := Coeff(0,0,0,3,H);
      if not Trunc and then Inf(C2)<Cpp then
         Put("C2 ="); Show(C2);
         Message("RG2_Ops.Normalize",Domain_Violation);
      end if;
      if  abs(Center(C2))<1.0E-250 then
         Message("RG2_Ops.Normalize",Giving_Up);New_Line;
      end if;
      R := Sqrt(abs(Scal(Cpp)/C2));
      Scale(Inv(R),R,H,Hn);
      Rvh0 := Component(0,0,0,0,Hn);
      SetComponent(0,0,0,0,ResetCenter(Rvh0),Hn);
      Rmu:= R*Scal(MuMax);
   end Normalize;


   procedure ShowRR(R: in Weights) is
   begin
      Put(Current_Output," R=(");
      Put(Current_Output,R(1),5,3);
      Put(Current_Output,R(2),5,3);
      Put(Current_Output,R(3),5,3);
      Put_Line(" )");
   end ShowRR;


   procedure StarStuff(Err: in Radius; PPi: in ProjParam; Mu: in Scalar; Hpsi: in Polynom; Zstar,Lambda2: out Scalar) is
      use Scalar_Taylor_Ops, Scalar_Fou1_Ops, Scalar_Fou2_Ops;
      P1: Scalar;
      P: Scalar_Taylor_Ops.Taylor;
      H: Fourier2;
   begin
      GetNumeric(Hpsi,H);
      Gig(PPi.L,H);
      NumValQ0(H,P);
      P1 := Component(1,P);
      AddComponent(1,ScalOne+Mu*Theta,P);
      Zstar := ScalZero;
      FindRoot(Err,P,ScalZero,Zstar);
      SetComponent(1,P1,P);
      Lambda2 := -InvTheta*(ScalOne+DerVal(One,Zstar,P));
   end StarStuff;


   procedure DS(Err: in Radius; PPi: in ProjParam; Rt: in Weights; Hpsi: in Polynom; Q1,Q2,Z: in Scalar; DS: out DS_Matrix) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      One,Hpsi1,Hzp,H1,Bop,OmegaPQ: Fourier2;
   begin
      SetZero(Rt,One);
      SetComponent(0,0,0,1,ScalOne,One);
      Hpsi1:=GetNumeric(Hpsi);

      -- compute BoP=Omega*P
      Gig(PPi.L,Hpsi1,H1); --even part of H1 has odd Z-powers, odd - even
      NumFixComp3(Err,H1,BoP); -- even part of BoP has odd Z-powers, odd - even

      -- compute OmegaPQ=Omega'*Q
      SetZero(Rt,Hzp);
      Dig(PPi.L,Hpsi1,H1); -- H1 := (D_3/(omega*Grad))Hpsi1=D_3(phi)
      Copy(BoP,Hzp);
      AddComponent(0,0,0,1,ScalOne,Hzp);
      Comp3(H1,Hzp,OmegaPQ);

      DS(1,1):=ScalOne;
      DS(1,2):=ScalZero;
      DS(1,3):=ScalZero;
      DS(1,4):=ScalZero;

      DS(2,1):=NumVal(Q1,Q2,Z,NumDqDot(Rt,InvTheta,ScalOne,OmegaPQ));
      DS(2,2):=ScalOne+NumVal(Q1,Q2,Z,NumDqDot(Rt,ScalOne,-InvTheta,OmegaPQ));
      DS(2,3):=ScalZero;
      DS(2,4):=NumVal(Q1,Q2,Z,Der3(Rt(3),OmegaPQ));

      DS(3,1):=ScalZero;
      DS(3,2):=ScalZero;
      DS(3,3):=ScalOne;
      DS(3,4):=ScalZero;

      DS(4,1):=NumVal(Q1,Q2,Z,NumDqDot(Rt,InvTheta,ScalOne,BoP));
      DS(4,2):=NumVal(Q1,Q2,Z,NumDqDot(Rt,ScalOne,-InvTheta,BoP));
      DS(4,3):=ScalZero;
      DS(4,4):=ScalOne+NumVal(Q1,Q2,Z,Der3(Rt(3),BoP));
   end DS;


   procedure DS(Err: in Radius; PPi: in ProjParam; Rt: in Weights; DSIname,MuIname: in String; Q1,Q2,Z: in Scalar; DS: out DS_Matrix) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Bop,OmegaPQ: Fourier2;
      Mu: Mu_Array;
      F: File_Type;
      Mu12: Scalar;
   begin
      Read(DSIname & "Y",Rt,OmegaPQ);
      Read(DSIname & "Z",Rt,BoP);
      Open(F,In_File,MuIname);

      Mu12:=ScalOne;
      for K in 0..5 loop
         Get(F,Mu(K)); Mu(K+6):=Mu(K);
         Mu12:=Mu(K)*Mu(K+6)*Mu12;
      end loop;
      Close(F);

      DS(1,1):=Theta12;
      DS(1,2):=ScalZero;
      DS(1,3):=ScalZero;
      DS(1,4):=ScalZero;

      DS(2,1):=NumVal(Q1,Q2,Z,NumDqDot(Rt,InvTheta,ScalOne,OmegaPQ));
      DS(2,2):=InvTheta12+NumVal(Q1,Q2,Z,NumDqDot(Rt,ScalOne,-InvTheta,OmegaPQ));
      DS(2,3):=ScalZero;
      DS(2,4):=NumVal(Q1,Q2,Z,Der3(Rt(3),OmegaPQ));

      DS(3,1):=ScalZero;
      DS(3,2):=ScalZero;
      DS(3,3):=Mu12*InvTheta12;
      DS(3,4):=ScalZero;

      DS(4,1):=NumVal(Q1,Q2,Z,NumDqDot(Rt,InvTheta,ScalOne,BoP));
      DS(4,2):=NumVal(Q1,Q2,Z,NumDqDot(Rt,ScalOne,-InvTheta,BoP));
      DS(4,3):=ScalZero;
      DS(4,4):=Mu12*Theta12+NumVal(Q1,Q2,Z,Der3(Rt(3),BoP));
   end DS;


   procedure ShowDS(A: in DS_Matrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Put(Current_Output,A(I,J));Put(" ");
         end loop;
         New_Line;
      end loop;
      New_Line;
   end ShowDS;


   procedure Mult(A1: in out DS_Matrix; A2: in DS_Matrix) is
      R: Scalar;
      B: DS_Vector;
      C: DS_Matrix;
   begin
      for J in B'Range loop
         for K in B'Range loop
            B(K) := A2(K,J);
         end loop;
         for I in B'Range loop
            R := ScalZero;
            for K in B'Range loop
               R := R+A1(I,K)*B(K);
            end loop;
            C(I,J) := R;
         end loop;
      end loop;
      A1:=C;
   end Mult;


   procedure RG1(Err: in Radius; PPi: in ProjParam; Rt,Rf: in Weights; H1: in Hamilton; Hpsi: in out Polynom; W: in out Sparse; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Rpsi: Weights;
   begin
      if not Trunc then
         Message("RG2_Ops.RG",Numeric_Only);
      end if;
      Rpsi := (Rt(1),Rt(2),Rho(H1)(3));
      Numeric_Fou2_Ops.Iminus(PPi.L,Hpsi);                 -- just in case
      PsiFind_Sparse(Err,Rpsi,PPi,H1,Hpsi,W);
      HoU(Err,PPi,Rt,H1,GetNumeric(Hpsi),H2);                 -- Rt is proposed domain of HoU
      declare
         Ht: Hamilton;
      begin
         Iminus(PPi.L,H2,Ht);
         Shift(Rf,PPi,PPi,H2,Ht);                                -- includes Iplus and scaling with MuMax
         Normalize(Ht,H2);
      end;
   end RG1;


   procedure RG2(Err: in Radius; PPi: in ProjParam; Rt,Rf: in Weights; H1: in Hamilton; Hpsi: in out Polynom; H2: out Hamilton; N: in Integer:=0) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Ht: Hamilton;
      H11: Hamilton;
   begin
      SetZero(Rf,H2);
      SetZero(Rf,Ht);
      if not Trunc then
         Message("RG2_Ops.RG",Numeric_Only);
      end if;
      --      Numeric_Fou2_Ops.Iminus(PPi.L,Hpsi);                 -- just in case
      H11:=H1;
      PsiFind2(Err,Rt,PPi,H11,Hpsi);
      HoToU(Err,PPi,Rt,H11,GetNumeric(Hpsi),Ht);
      Normalize(Ht,H2);
   exception
      when Messages.Leaving =>
         Message("RG2_Ops.RG",Giving_Up);
         raise;
         return;
   end RG2;


   procedure RG3(Err: in Radius; PPi: in ProjParam; Rt,Rf: in Weights; H1: in Hamilton; Hpsi: in out Polynom; H2: out Hamilton; N: in Integer:=0) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      Htt,Ht,Hm: Hamilton;
      HHpsi: Polynom;
   begin
      SetZero(Rf,H2);
      SetZero(Rf,Ht);
      if not Trunc then
         Message("RG2_Ops.RG3",Numeric_Only);
      end if;
      HoToU(Err,PPi,Rt,H1,GetNumeric(Hpsi),Ht);
      Iminus(PPi.L,Ht,Hm);
      Copy(Hpsi,HHpsi);
      Copy(Ht,Htt);
      SetZero(Rt,Hpsi);
      PsiFind(Err,Rt,PPi,Htt,Hpsi);
      HoU(Err,PPi,Rt,Htt,GetNumeric(Hpsi),Ht);
      PsiOfUoU(Err,PPi,HHpsi,Hpsi);
      Normalize(Ht,H2);
      Iminus(PPi.L,H2,Hm);
   exception
      when Messages.Leaving =>
         Message("RG2_Ops.RG3",Giving_Up);
         raise;
         return;
   end RG3;


   procedure PsiOfUoU(Err: in Radius; P: in ProjParam; Psi1: in Polynom; Psi2: in out Polynom) is
      use Numeric_Fou2_Ops;
      use Numeric_Fou1_Ops;
      E: Error;
      Hq,Hz,Ht,Ht1,Ht2: Fourier2;
   begin
      Dig(P.L,Psi1,Ht1);
      Gig(P.L,Psi2,Ht2);
      Copy(Ht1,Hq);
      for K in 1 .. 20 loop
         CompOmega12(Rho(Ht2),Ht2,Hq,Hz);
         AddComponent(0,0,0,1,NumOne,Hz);
         Copy(Hq,Ht);
         Comp3(Ht1,Hz,Hq);
         E := Norm(Hq-Ht);
         exit when Sup(E)<Err;
      end loop;
      if Sup(E)>Err then
         Put_Line("RG2_Ops.PsiOfUoU: Got tired");
      end if;
      CompOmega12(Rho(Psi2),Psi2,Hq,Ht2);
      Comp3(Psi1,Hz,Ht1);
      Sum(Ht1,Ht2,Psi2);
      ZeroAverage(Psi2);
   end PsiOfUoU;


   procedure Hinc(T: in Scalar; H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      Copy(H1,H2);
      AddComponent(0,0,0,1,T,H2);
   end Hinc;


   procedure Hinc2(T: in Scalar; H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      Copy(H1,H2);
      AddComponent(0,0,0,2,T,H2);
   end Hinc2;


   procedure Hinx(T: in Svec2; He1,He2,H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      Copy(H1,H2);
      MultAdd(T(1),He1,H2);
      MultAdd(T(2),He2,H2);
   end Hinx;


   procedure Hinx(T: in Scalar; He1,H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      Copy(H1,H2);
      MultAdd(T,He1,H2);
   end Hinx;


   procedure Hinx(T: in Svec2; He1,He2: in Hamilton; H2: in out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      MultAdd(T(1),He1,H2);
      MultAdd(T(2),He2,H2);
   end Hinx;


   procedure Hinx(T: in Scalar; He1: in Hamilton; H2: in out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      MultAdd(T,He1,H2);
   end Hinx;


   function Nfunc(H1,H2: Hamilton) return Scalar is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      return Component(0,0,0,1,H2)-Component(0,0,0,1,H1);
   end Nfunc;


   function Nfunc2(H1,H2: Hamilton) return Scalar is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      return Component(0,0,0,2,H2)-Component(0,0,0,2,H1);
   end Nfunc2;


   function Nfunx(R: Weights; He1,He2,H1,H2: Hamilton) return Svec2 is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      return Components(R,He1,He2,H2-H1);
   end Nfunx;


   function Nfunx(R: Weights; He1,H1,H2: Hamilton) return Scalar is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      return NumIprod(R,H2-H1,He1)/NumIprod(R,He1,He1);
   end Nfunx;


   procedure RGC(Iname,Oname: in String;Pw,P: in Integer; Eps: in Rep; Err: in Radius; PPi: in ProjParam; Rf,Rpsi: in Weights; H1: in  Hamilton; H2: out Hamilton; Read_Psi,Write_Psi: in Boolean) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      AS: Boolean;
      D,Eps1: Rep;
      G0,Gplus,Gminus,Dt,Aeps,Beps,Ceps: Scalar;
      HH,Htt,Ht,Ht2,Htt2: Hamilton;
      T: Rep;
      S: Scalar:=ScalZero;
   begin
      Eps1:=Rep(1.0E-3);
      if not Trunc then
         Message("RG2_Ops.RGC",Numeric_Only);
      end if;
      AuxSave := False;
      AS := AuxSave;
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H1,H2,Read_Psi,False);
      G0 := Nfunc(H1,H2); --Difference of z terms of H2 and H1

      Dt := Scal(Eps1);
      Hinc(Dt,H1,Ht);
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Ht,Ht2,Read_Psi,False);
      Gplus := Nfunc(Ht,Ht2); --Difference of z for a right shifted Hamiltonian

      Dt := Scal(-Eps1);
      Hinc(Dt,H1,Htt); --Ht=H1+Dt*Z
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Htt,Htt2,Read_Psi,False);
      Gminus := Nfunc(Htt,Htt2); --Difference of z for a left shifted Hamiltonian

      Ceps := G0*Scal(Eps1);
      Beps := (Gplus-Gminus)/ScalTwo; -- first derivative
      Aeps := ((Gplus+Gminus)/ScalTwo-G0)/Scal(Eps1);

      if abs(Center(Aeps))<1.0E-300 then
         if abs(Center(Beps))<1.0E-300 then
            Message("RG2_Ops.RGC",Giving_Up);
         end if;
         T := -Center(Ceps/Beps); --Shift in z-terms divided by the derivative of the shift
      else
         T := -Center(Beps/(ScalTwo*Aeps));
         D := T*T-Center(Ceps/Aeps);
         if abs(D)>Rep(1.0E+300) then
            Message("RG2_Ops.RGC",Giving_Up);
         end if;
         if D<=Zero then
            null;
         elsif T>Zero then
            T := T-Sqrt(D);
         else
            T := T+Sqrt(D);
         end if;
      end if;

      Hinc(Scal(T),H1,Ht);
      Copy(Ht,HH);
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,HH,H2,Read_Psi,Write_Psi);

   exception
      when Messages.Leaving =>
         EX:=1;
         SetRho(Rho(H1),H2);
         return;
         raise;
   end RGC;


   procedure Rotate_Scale(Rt: in Weights; S1, S2, S3: in Scalar; H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou2_Ops;
      HH,H,Hq:  Hamilton;
   begin
      SetZero(Rt,Hq);
      SetComponent(0,0,0,0,ScalOne,Hq);
      Rotation(Rt,S1,S2,H1,Hq,H);
      Eliminate(H,HH);
      H2:=Scale(ScalOne,S3,HH);
   end Rotate_Scale;


   procedure RG_Pw(Iname,Oname: in String; Pw,P: in Integer; Err: in Radius; PPi: in ProjParam; Rf,Rpsi: in Weights; H1: in Hamilton; H2: out Hamilton; Read_Psi,Write_Psi: in Boolean) is
      --Pw-th power of RG2, starts from P-th Hamiltonian in the Pw-periodic cycle
      --Iname & K are the files from which the generating functions are read
      --Oname & K are the files to which the generating functions are written
      use Scalar_Fou2_Ops;
      use Scalar_Fou2_Ops;
      Hr,Hrr: Hamilton;
      Hpsi: Polynom;
      HHpsi: Hamilton;
      KK: String:="00";
      F: File_Type;
      Tc: Scalar:=ScalZero;
   begin
      Hr:=H1;
      for K in P..(P+(Pw-1)) loop
         case (K mod 12) is
            when  0 => KK:="00";
            when  1 => KK:="01";
            when  2 => KK:="02";
            when  3 => KK:="03";
            when  4 => KK:="04";
            when  5 => KK:="05";
            when  6 => KK:="06";
            when  7 => KK:="07";
            when  8 => KK:="08";
            when  9 => KK:="09";
            when  10 => KK:="10";
            when  11 => KK:="11";
            when others =>  Message("RG2_Ops.RG_Pw",Parameter_Error);
         end case;

         if Read_Psi then
            SetZero(Rf,HHpsi);
            Read(Iname & KK,Rf,HHpsi);
            Hpsi:=PutNumeric(HHpsi);
         else
            Hpsi:=RoughPsi(PPi,Hr,Rf);
         end if;

         RG2(Err,PPi,Rpsi,Rf,Hr,Hpsi,Hrr,0);

         if Write_Psi then
            Create(F,Out_File,Oname & KK);
            Write(F,GetNumeric(Hpsi));
            Close(F);
         end if;
         Hr:=Hrr;
      end loop;

      if (P mod 3=0) then
         PiTranslateX(Hr,H2);
      end if;

      if (P mod 3=1) then
         PiTranslateX(Hr,Hrr);
         PiTranslateY(Hrr,H2);
      end if;

      if (P mod 3=2) then
         PiTranslateY(Hr,H2);
      end if;

   exception
      when Messages.Leaving =>
         Message("RG2_Ops.RG_Pw ",Giving_Up);
         raise;
         return;
   end RG_Pw;


   function Nfuncc(H1,H2: Hamilton) return Scalar is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      if not Trunc then
         Message("RG2_Ops.Nfuncc",Numeric_Only);
      end if;
      return NumVal(ScalZero,ScalZero,Scal(0.2),H2)-NumVal(ScalZero,ScalZero,Scal(0.2),H1);
   end Nfuncc;


   procedure Hincc(T: in Scalar; P: in ProjParam; H1: in Hamilton; H2: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
   begin
      Copy(H1,H2);
      if P.L(1,0)<0 then
         AddComponent(0,1,0,0,T,H2);
      else
         AddComponent(0,-1,1,0,T,H2);     --- original, where (1,0,0) is of type Iminus
      end if;
   end Hincc;


   procedure FindEV(Iname,Oname: in String; Pw,P: in Integer; Err: in Radius; PPi: in ProjParam; Rf,Rpsi: in Weights; H: in Hamilton; H1,H2: in out Hamilton;Read_Psi: in Boolean; Text_Output: in Boolean:=True) is
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      Err1,Err2,Eps1,Eps2: Radius;
      Ei1,Ei2,E1,E2,Et1,Et2,Es1,Es2: Scalar;
      V1,V2,A1,A2: Svec2;
      A: Smat2;
      Hi,Hi1,Hi2: Hamilton;

      procedure Eigen2(A: in Smat2; E1,E2: out Scalar; V1,V2: out Svec2) is
         R1,R2,X: Scalar;
      begin
         R1 := (A(1,1)+A(2,2))/ScalTwo;
         R2 := R1*R1-(A(1,1)*A(2,2)-A(1,2)*A(2,1));
         if Center(R2)>=Zero then
            if Center(R1)>=Zero then
               E1 := R1+Sqrt(R2);
               E2 := R1-Sqrt(R2);
            else
               E1 := R1-Sqrt(R2);
               E2 := R1+Sqrt(R2);
            end if;
            X := (A(1,2)+A(1,1)-E1)/(A(1,2)-A(1,1)+E1);
            V1(1) := ScalOne+X;
            v1(2) := ScalOne-X;
            X := (A(2,2)+A(2,1)-E2)/(A(2,2)-A(2,1)-E2);
            V2(1) := ScalOne+X;
            v2(2) := ScalOne-X;
         else
            Message("FindEV.Eigen2, complex eigenvalues",Not_Implemented);
            E1 := R1;
            E2 := R2;
            X := (A(1,2)+A(1,1)-E1)/(A(1,2)-A(1,1)+E1);
            V1(1) := X-ScalOne;
            V1(2) := X+ScalOne;
            V2(1) := ScalOne+X;
            V2(2) := ScalOne-X;
         end if;
      end Eigen2;

   begin
      Err1 := 1.0E-8;                      -- precision for e1
      Err2 := 1.0E-8;                      -- precision for e2
      Eps1  := 1.0E-5;                     -- increment for difference quotient
      Eps2  := 1.0E-5;
      if Text_Output then
         New_Line;
         Put("Renormalization precision in FindEV:"); Put(Current_Output,Err);
         New_Line;
         Put("Precision for eigenvectors:"); Put(Current_Output,Err1);
         New_Line;
         Put("Increment in FindEV for the first eigenvector:"); Put(Current_Output,Eps1);
         New_Line;
         Put("Increment in FindEV for the first eigenvector:"); Put(Current_Output,Eps2);
         New_Line;
      end if;
      E1 := Scal(999);
      E2 := Scal(999);
      Et1 := -E1;
      Et2 := -E2;
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H,Hi,Read_Psi,False);
      for Iter in 1 .. 100  loop
         if Text_Output then
            New_Line;
            New_Line;
            Put("Iteration # "); ShortPut(Iter);
            New_Line;
         end if;
         NumInormalize(Rf,H1);
         RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H+Scal(Eps1)*H1,Hi1,Read_Psi,False);
         Hi1 := Inv(Scal(Eps1))*(Hi1-Hi);

         NumInormalize(Rf,H2);
         RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H+Scal(Eps2)*H2,Hi2,Read_Psi,False);
         Hi2 := Inv(Scal(Eps2))*(Hi2-Hi);

         A1 := Components(Rf,Hi1,Hi2,H1); --Components of H1 along Hi1 and Hi2
         A2 := Components(Rf,Hi1,Hi2,H2); --Components of H2 along Hi1 and Hi2
         A(1,1) := A1(1); --Component of H1 along Hi1
         A(2,1) := A1(2);--Component of H1 along Hi2
         A(1,2) := A2(1); --Component of H2 along Hi1
         A(2,2) := A2(2);--Component of H2 along Hi1
         Eigen2(A,Ei2,Ei1,V2,V1);

         H1 := V1(1)*Hi1+V1(2)*Hi2;
         H2 := V2(1)*Hi1+V2(2)*Hi2;

         Es1 := Et1;
         Es2 := Et2;
         Et1 := E1;
         Et2 := E2;
         E1 := Inv(Ei1);
         E2 := Inv(Ei2);
         NumInormalize(Rf,H1);
         NumInormalize(Rf,H2);
         if Text_Output then
            Put("e1 = "); Show(E1);
            Put("e2 = "); Show(E2);
            Put("Error 1.1:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E1-Et1)/(E1+Et1)));
            Put("   Error 1.2:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E1-Es1)/(E1+Es1)));
            New_Line;
            Put("Error 2.1:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E2-Et2)/(E2+Et2)));
            Put("   Error 2.2:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E2-Es2)/(E2+Es2)));
            New_Line;
            Put("<He1,He1>:="); Put(Current_Output,Center(NumIprod(Rf,H1,H1)));
            New_Line;
            Put("<He2,He2>:="); Put(Current_Output,Center(NumIprod(Rf,H2,H2)));
            New_Line;
         end if;
         if SupAbs(Scal(2.0)*(E1-Et1)/(E1+Et1))<Err1 and SupAbs(Scal(2.0)*(E2-Et2)/(E2+Et2))<Err2 then
            if Text_Output then
               New_Line;
               New_Line;
               Put("e1 = "); Show(E1);
               Put("e2 = "); Show(E2);
               New_Line;
               Put("Error 1:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E1-Et1)/(E1+Et1)));
               New_Line;
               Put("Error 2:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E2-Et2)/(E2+Et2)));
               New_Line;
               Put_Line("FindEV: Done");
            end if;
            return;
         end if;
         if SupAbs(Scal(2.0)*(E1-Es1)/(E1+Es1))<Err1 and SupAbs(Scal(2.0)*(E2-Es2)/(E2+Es2))<Err2 then
            if Text_Output then
               New_Line;
               New_Line;
               Put("e1 = "); Show(E1);
               Put("e2 = "); Show(E2);
               New_Line;
               Put("Error 1:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E1-Es1)/(E1+Es1)));
               New_Line;
               Put("Error 2:= "); Put(Current_Output,SupAbs(Scal(2.0)*(E2-Es2)/(E2+Es2)));
               New_Line;
               Put_Line("FindEV: Done");
            end if;
            return;
         end if;
      end loop;
      <<EN>> New_Line;
      Message("RG2_Ops.FindEV ",Giving_Up);
   exception

      when Messages.Leaving =>
         EX:=1;
         return;
         raise;
   end FindEV;


   procedure RGX(Iname,Oname: in String;Pw,P: in Integer; Err: in Radius; PPi: in ProjParam; Rf,Rpsi: in Weights; He1,He2: in out Hamilton; H1: in out Hamilton; H2: out Hamilton; Read_Psi,Write_Psi: in Boolean) is
      --Iname & K are the files from which the generating functions are read
      --Oname & K are the files to which the generating functions are written
      --the L-loop included for better accuracy. If the approximate Hamiltonian is a good approximation, L-loop can be shortened/removed.
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      Eps1,Eps2: Rep;
      Det: Scalar;
      TT,DT,G0,G1,G2,G1p,G2p: Svec2;
      Ht1,Ht2,Htt2,D1,D2: Hamilton;                          -- one Ht suffices if G0 is computed after G1 and G2
   begin
      for L in 1..4 loop
         if not Trunc then
            Message("RG2_Ops.RGX2",Numeric_Only);
         end if;
         Eps1 := 1.0E-4;
         Eps2 := 1.0E-4;

         NumInormalize(Rf,He1);
         NumInormalize(Rf,He2);
         RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H1,H2,Read_Psi,False);
         G0 := Nfunx(Rf,He1,He2,H1,H2);

         DT(1) := Scal(Eps1);
         DT(2) := ScalZero;
         Hinx(DT,He1,He2,H1,Ht1); --Ht1=He1*Eps+H1

         RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Ht1,Ht2,Read_Psi,False);
         G1p := Nfunx(Rf,He1,He2,Ht1,Ht2); --G1(1)=Comp(Ht2-Ht1) along He1

         DT(1) := ScalZero;
         DT(2) := Scal(Eps2);
         Hinx(DT,He1,He2,H1,Ht1);

         RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Ht1,Htt2,Read_Psi,False);
         G2p := Nfunx(Rf,He1,He2,Ht1,Htt2);

         G1(1) := (G1p(1)-G0(1))/Scal(Eps1); --derivative of the H1 component of the difference
         G1(2) := (G1p(2)-G0(2))/Scal(Eps1); --derivative of the H2 component of the difference
         G2(1) := (G2p(1)-G0(1))/Scal(Eps2); --derivative of the H1 component of the difference
         G2(2) := (G2p(2)-G0(2))/Scal(Eps2); --derivative of the H2 component of the difference
         Det := G1(2)*G2(1)-G1(1)*G2(2);
         TT(1) := (G0(1)*G2(2)-G0(2)*G2(1))/Det;
         TT(2) := (-G0(1)*G1(2)+G0(2)*G1(1))/Det;

         -- a possible way to avoid the 4-th iteration

         --D1:=(ScalOne/Scal(Eps1))*(Ht2-H2);
         --D2:=(ScalOne/Scal(Eps2))*(Htt2-H2);
         --H2:=H2+TT(1)*D1+TT(2)*D2;

         --otherwise...
         Hinx(TT,He1,He2,H1);
      end loop;
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H1,H2,Read_Psi,Write_Psi);
   end RGX;


   function Join(I,M,N,K: Integer) return Integer is
   begin
      if I=0 then
         return ((2*1000+(100+M))*100+N)*100+K;
      else
         return ((1*1000+(100+M))*100+N)*100+K;
      end if;
   end Join;


   procedure UnJoin(MNK: in Integer; I,M,N,K: out Integer) is
      L: Integer;
   begin

      I :=MNK/10000000;
      L:= MNK -I*10000000;
      I:=I mod 2;

      M := L/10000;
      L := L-10000*M;
      M := M-100;

      N := L/100;
      K := L-100*N;
   end UnJoin;


   procedure NumDRG(Iname,Oname: in String; Pw,P: in Integer; Mmin,Mmax: in Integer; Err,CutOff: in Rep; PPi: in ProjParam; Rf,Rpsi: in Weights; H1,H2: in Hamilton; Read_Psi: in Boolean;  Text_Output: in Boolean:=True) is
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      Eps: Rep;
      FF,FFF: File_Type;
      Ht1,Ht2,Htest: Hamilton;
      II: String:=" ";
      MM: String:="   ";
      NN: String:="  ";
      KK: String:="  ";
      Rfix: Weights;
      T: Scalar:=ScalZero;
   begin
      Eps:=Rep(1.0E-4);
      Rfix := (0.00, 0.85, 0.15);
      if not Trunc then
         Message("RG2_Ops.NumDRG",Numeric_Only);
      end if;
      for Mj in Mmin..Mmax loop
         for Nj in Nmin(Mj,0) .. Nmax(Mj) loop
            for Kj in 0 .. Deg(0,Mj,Nj,H2) loop
               NameForm(0,Mj,Nj,Kj,II,MM,NN,KK);
               if abs(Center(Component(0,Mj,Nj,Kj,H2)))>CutOff then
                  Copy(H1,Ht1);
                  XE:=0;
                  Read(Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK,Rfix,Htest); --sets XE=1 if the file does not exist
                  if XE=1 then
                     AddComponent(0,Mj,Nj,Kj,Scal(Eps),Ht1);
                     RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Ht1,Ht2,Read_Psi,False);
                     Sub(H2,Ht2);
                     Mult(Scal(One/Eps),Ht2);
                     if Text_Output then
                        Put(Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK &": Computing");
                        New_Line;
                     end if;
                     Create(FF,Out_File,Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK);
                     Write(FF,Ht2);
                     Close(FF);
                  else
                     if Text_Output then
                        Put(Oname & "_" & II & "_" & MM &"_"& NN & "_" & KK & ": File exists" );
                        New_Line;
                     end if;
                  end if;
               else
                  if Text_Output then
                     Put(Oname & "_" & II & "_" & MM & "_" & NN & "_" &  KK & ": Component too small");
                     New_Line;
                  end if;
               end if;
            end loop;
         end loop;

         for Nj in Nmin(Mj,1) .. Nmax(Mj) loop
            for Kj in 0 .. Deg(1,Mj,Nj,H2) loop
               NameForm(1,Mj,Nj,Kj,II,MM,NN,KK);
               if abs(Center(Component(1,Mj,Nj,Kj,H2)))>CutOff then
                  Copy(H1,Ht1);
                  XE:=0;
                  Read(Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK,Rfix,Htest); --sets XE=1 if the file does not exist
                  if XE=1 then
                     if Text_Output then
                        Put(Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK & ": Computing");
                        New_Line;
                     end if;
                     AddComponent(1,Mj,Nj,Kj,Scal(Eps),Ht1);
                     RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,Ht1,Ht2,Read_Psi,False);
                     Sub(H2,Ht2);
                     Mult(Scal(One/Eps),Ht2);

                     Create(FF,Out_File,Oname & "_" & II & "_" & MM & "_" & NN & "_" & KK);
                     Write(FF,Ht2);
                     Close(FF);
                  else
                     if Text_Output then
                        Put(Oname & "_" & II & "_" & MM & "_" & NN & "_"  & KK & ": File exists");
                        New_Line;
                     end if;
                  end if;
               else
                  if Text_Output then
                     Put(Oname & "_" & II & "_" & MM & "_" & NN & "_" &  KK & ": Component too small");
                     New_Line;
                  end if;
               end if;
            end loop;
         end loop;
      end loop;
   end NumDRG;


   procedure ReadDRG(Rfix: in Weights; Iname: in String; DR: out Matrix; Iz,Jz: out Pivot; Text_Output: in Boolean:=True ) is
      pragma Suppress(Storage_Check);
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      J,I,Imax,M,N,K: Integer;
      H: Hamilton;
      II: String:=" ";
      MM: String:="   ";
      NN: String:="  ";
      KK: String:="  ";
   begin
      SetZero(DR);
      J:=0;
      Imax:=0;
      for Mj in Freq1 loop
         for Nj in Nmin(Mj,0) .. Nmax(Mj) loop
            for Kj in 0 .. P_Deg loop
               if Mj+Nj<=Q_Deg and Mj+Nj+Kj<=QPdeg and Nj mod 2=0 and Kj mod 2 =1 then
                  J:=J+1;
                  Jz(J):=Join(0,Mj,Nj,Kj);
                  NameForm(0,Mj,Nj,Kj,II,MM,NN,KK);
                  XE:=0;
                  Read(Iname & "_" & II & "_" & MM & "_" & NN & "_" & KK,Rfix,H);
                  if XE=1 then
                     if Text_Output then
                        New_Line;
                        Put("No column_" & II & "_" & MM & "_" & NN & "_" & KK);
                     end if;
                     DR(J,J):=Rep(-1.0);
                     goto JUMP1;
                  end if;
                  I:=0;
                  if Text_Output then
                     New_Line;
                     Put("Reading column_" & II & "_" & MM & "_" & NN & "_" & KK);
                  end if;
                  for Mi in Freq1 loop
                     for Ni in Nmin(Mi,0) .. Nmax(Mi) loop
                        for Ki in 0 .. P_Deg loop
                           if Mi+Ni<=Q_Deg  and Mi+Ni+Ki<=QPdeg and Ni mod 2=0 and Ki mod 2 =1 then
                              I :=I+1;
                              Iz(I):=Join(0,Mi,Ni,Ki);
                              if I/=J then
                                 DR(I,J):=Center(Component(0,Mi,Ni,Ki,H));
                              else
                                 DR(I,J):=Center(Component(0,Mi,Ni,Ki,H))-Rep(1.0);
                              end if;
                           end if;
                        end loop;
                     end loop;


                     for Ni in Nmin(Mi,1) .. Nmax(Mi) loop
                        for Ki in 0 .. P_Deg loop
                           if Mi+Ni<=Q_Deg and Mi+Ni+Ki<=QPdeg and Ni mod 2=1 and Ki mod 2 =0 then
                              I :=I+1;
                              Iz(I):=Join(1,Mi,Ni,Ki);
                              if I/=J then
                                 DR(I,J):=Center(Component(1,Mi,Ni,Ki,H));
                              else
                                 DR(I,J):=Center(Component(1,Mi,Ni,Ki,H))-Rep(1.0);
                              end if;
                           end if;
                        end loop;
                     end loop;
                  end loop;
                  <<JUMP1>> null;
               end if;
            end loop;
         end loop;

         for Nj in Nmin(Mj,1) .. Nmax(Mj) loop
            for Kj in 0 .. P_Deg loop
               if Mj+Nj <=Q_Deg and Mj+Nj+Kj<=QPdeg and  Nj mod 2=1 and Kj mod 2 =0 then
                  J:=J+1;
                  Jz(J):=Join(1,Mj,Nj,Kj);
                  NameForm(1,Mj,Nj,Kj,II,MM,NN,KK);
                  XE:=0;
                  Read(Iname & "_" & II & "_" & MM & "_" & NN & "_" & KK,Rfix,H);
                  if XE=1 then
                     if Text_Output then
                        New_Line;
                        Put("No column_" & II & "_" & MM & "_" & NN & "_" & KK);
                     end if;
                     DR(J,J):=Rep(-1.0);
                     goto JUMP2;
                  end if;
                  if Text_Output then
                     New_Line;
                     Put("Reading column_" & II & "_" & MM & "_" & NN & "_" & KK);
                  end if;
                  I:=0;
                  for Mi in Freq1 loop
                     for Ni in Nmin(Mi,0) .. Nmax(Mi) loop
                        for Ki in 0 .. P_Deg loop
                           if  Mi+Ni <=Q_Deg and Mi+Ni+Ki<=QPdeg and  Ni mod 2=0 and Ki mod 2 =1 then
                              I :=I+1;
                              Iz(I):=Join(0,Mi,Ni,Ki);
                              if I/=J then
                                 DR(I,J):=Center(Component(0,Mi,Ni,Ki,H));
                              else
                                 DR(I,J):=Center(Component(0,Mi,Ni,Ki,H))-Rep(1.0);
                              end if;
                           end if;
                        end loop;
                     end loop;

                     for Ni in Nmin(Mi,1) .. Nmax(Mi) loop
                        for Ki in 0 .. P_Deg loop
                           if  Mi+Ni <=Q_Deg and Mi+Ni+Ki<=QPdeg and Ni mod 2=1 and Ki mod 2=0 then
                              I:=I+1;
                              Iz(I):=Join(1,Mi,Ni,Ki);
                              if I/=J then
                                 DR(I,J):=Center(Component(1,Mi,Ni,Ki,H));
                              else
                                 DR(I,J):=Center(Component(1,Mi,Ni,Ki,H))-Rep(1.0);
                              end if;
                           end if;
                        end loop;
                     end loop;
                  end loop;
                  <<JUMP2>> null;
               end if;
            end loop;
         end loop;
      end loop;
      if Text_Output then
         New_Line;
         New_Line;
         Put("Dimension:");Put(J);
         New_Line;
      end if;

      Invert(DR);

      if Text_Output then
         for L in DR'Range(1) loop
            for S in DR'Range(2) loop
               UnJoin(Iz(L),I,M,N,K); NameForm(I,M,N,K,II,MM,NN,KK);
               New_Line;
               ShortPut(L);Put(" "); ShortPut(Iz(L));Put(" "); ShortPut(S);Put(" "); ShortPut(Jz(S)); Put(" " & II & "_" & MM & "_" & NN & "_" &  KK & "  and ");
               UnJoin(Jz(S),I,M,N,K); NameForm(I,M,N,K,II,MM,NN,KK);
               Put(II & "_" & MM & "_" & NN & "_" &  KK & " ");
               Put(Current_Output,DR(L,S));
            end loop;
         end loop;
      end if;
   end ReadDRG;


   procedure MultAdd(W: in Sparse; H1: in Hamilton; H2: in out Hamilton) is
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      Mi,Ni,Ki,Mj,Nj,Kj,Ii,Ij: Integer;
   begin
      if not Trunc then
         Message("RG2_Ops.MultAdd",Numeric_Only);
      end if;
      for L in 1 .. Length(W) loop
         UnJoin(W(L).I,Ii,Mi,Ni,Ki);
         UnJoin(W(L).J,Ij,Mj,Nj,Kj);
         AddComponent(Ii,Mi,Ni,Ki,Scal(W(L).R)*Component(Ij,Mj,Nj,Kj,H1),H2);
      end loop;
   end MultAdd;


   procedure MultAdd(W: in Matrix; Iz,Jz: in Pivot; H1: in Hamilton; H2: in out Hamilton) is
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      Mi,Ni,Ki,Mj,Nj,Kj,Ii,Ij: Integer;
   begin
      if not Trunc then
         Message("RG2_Ops.MultAdd",Numeric_Only);
      end if;
      for I in W'Range(1) loop
         UnJoin(Iz(I),Ii,Mi,Ni,Ki);
         for J in W'Range(2) loop
            UnJoin(Jz(J),Ij,Mj,Nj,Kj);
            AddComponent(Ii,Mi,Ni,Ki,Scal(W(I,J))*Component(Ij,Mj,Nj,Kj,H1),H2);
         end loop;
      end loop;
   end MultAdd;


   procedure RGY(Iname,Oname: in String; Pw,P: in Integer; Err: in Radius; PPi: in ProjParam; Rf,Rpsi: in Weights; H1: in out Hamilton; Y: in Matrix;  Iz,Jz: in Pivot; H2: out Hamilton;  Read_Psi,Write_Psi: in Boolean) is
      --Iname & K are the files from which the generating functions are read
      --Oname & K are the files to which the generating functions are written
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      H: Hamilton;
   begin
      RG_Pw(Iname,Oname,Pw,P,Err,PPi,Rf,Rpsi,H1,H2,Read_Psi,Write_Psi);
      Copy(H1,H);
      MultAdd(Y,Iz,Jz,H-H2,H);
      Copy(H,H2);
   end RGY;


   procedure NameForm(P,M,N,K: in Integer; II,MM,NN,KK: out String) is
   begin

      if P=0 then II:="0";
      else II:="1";
      end if;

      if M=-20 then MM:="-20"; end if;
      if M=-19 then MM:="-19"; end if;
      if M=-18 then MM:="-18"; end if;
      if M=-17 then MM:="-17"; end if;
      if M=-16 then MM:="-16"; end if;
      if M=-15 then MM:="-15"; end if;
      if M=-14 then MM:="-14"; end if;
      if M=-13 then MM:="-13"; end if;
      if M=-12 then MM:="-12"; end if;
      if M=-11 then MM:="-11"; end if;
      if M=-10 then MM:="-10"; end if;
      if M=-9 then MM:="-09"; end if;
      if M=-8 then MM:="-08"; end if;
      if M=-7 then MM:="-07"; end if;
      if M=-6 then MM:="-06"; end if;
      if M=-5 then MM:="-05"; end if;
      if M=-4 then MM:="-04"; end if;
      if M=-3 then MM:="-03"; end if;
      if M=-2 then MM:="-02"; end if;
      if M=-1 then MM:="-01"; end if;
      if M=0 then MM:="+00"; end if;
      if M=1 then MM:="+01"; end if;
      if M=2 then MM:="+02"; end if;
      if M=3 then MM:="+03"; end if;
      if M=4 then MM:="+04"; end if;
      if M=5 then MM:="+05"; end if;
      if M=6 then MM:="+06"; end if;
      if M=7 then MM:="+07"; end if;
      if M=8 then MM:="+08"; end if;
      if M=9 then MM:="+09"; end if;
      if M=10 then MM:="+10"; end if;
      if M=11 then MM:="+11"; end if;
      if M=12 then MM:="+12"; end if;
      if M=13 then MM:="+13"; end if;
      if M=14 then MM:="+14"; end if;
      if M=15 then MM:="+15"; end if;
      if M=16 then MM:="+16"; end if;
      if M=17 then MM:="+17"; end if;
      if M=18 then MM:="+18"; end if;
      if M=19 then MM:="+19"; end if;
      if M=20 then MM:="+20"; end if;

      if N=0 then NN:="00"; end if;
      if N=1 then NN:="01"; end if;
      if N=2 then NN:="02"; end if;
      if N=3 then NN:="03"; end if;
      if N=4 then NN:="04"; end if;
      if N=5 then NN:="05"; end if;
      if N=6 then NN:="06"; end if;
      if N=7 then NN:="07"; end if;
      if N=8 then NN:="08"; end if;
      if N=9 then NN:="09"; end if;
      if N=10 then NN:="10"; end if;
      if N=11 then NN:="11"; end if;
      if N=12 then NN:="12"; end if;
      if N=13 then NN:="13"; end if;
      if N=14 then NN:="14"; end if;
      if N=15 then NN:="15"; end if;
      if N=16 then NN:="16"; end if;
      if N=17 then NN:="17"; end if;
      if N=18 then NN:="18"; end if;
      if N=19 then NN:="19"; end if;
      if N=20 then NN:="20"; end if;

      if K=0 then KK:="00"; end if;
      if K=1 then KK:="01"; end if;
      if K=2 then KK:="02"; end if;
      if K=3 then KK:="03"; end if;
      if K=4 then KK:="04"; end if;
      if K=5 then KK:="05"; end if;
      if K=6 then KK:="06"; end if;
      if K=7 then KK:="07"; end if;
      if K=8 then KK:="08"; end if;
      if K=9 then KK:="09"; end if;
      if K=10 then KK:="10"; end if;
      if K=11 then KK:="11"; end if;
      if K=12 then KK:="12"; end if;
      if K=13 then KK:="13"; end if;
      if K=14 then KK:="14"; end if;
      if K=15 then KK:="15"; end if;
      if K=16 then KK:="16"; end if;
      if K=17 then KK:="17"; end if;
      if K=18 then KK:="18"; end if;
      if K=19 then KK:="19"; end if;
      if K=20 then KK:="20"; end if;
   end NameForm;


   function Components(R: Weights; H1,H2,H: Hamilton) return Svec2 is
      use Scalar_Fou1_Ops;
      use Scalar_Fou2_Ops;
      A,B,D,Det,W1,W2: Scalar;
      W: Svec2;
   begin
      if not Trunc then
         Message("RG2_Ops.Components",Numeric_Only);
      end if;
      A := NumIprod(R,H1,H1);
      B := NumIprod(R,H1,H2);
      D := NumIprod(R,H2,H2);
      Det := A*D-B*B;
      W1 := NumIprod(R,H,H1); --component of H along H1
      W2 := NumIprod(R,H,H2); --component of H along H2
      W(1) := (D*W1-B*W2)/Det;
      W(2) := (A*W2-B*W1)/Det;
      return W;
   end Components;


   procedure K_Term(K: in Integer; Err: in Radius; P: in ProjParam; R: in Weights; Hpsi: in Hamiltons; Mu: in Mu_Array; Y,Z: out Hamilton) is
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      H1,Hzp1: Fourier2;
      YY,ZZ: Hamilton;
      Pw: Integer;
   begin
      Pw:=11;
      -- compute Hzp1, Hzp1(q,z)=z'
      Gig(P.L,Hpsi(K),H1); --fine
      FixComp3(Err,R(3),H1,Hzp1);
      Z:=Hzp1;
      AddComponent(0,0,0,1,ScalOne,Hzp1);
      -- compute Hr, Hr=HoU_Hpsi
      Dig(P.L,Hpsi(K),H1);
      Comp3(H1,Hzp1,Y);
      --Composing with T_mu o U
      for N in K+1..Pw loop
         Scale(InvThetaCb,ThetaSq*Mu(N),Y,YY);
         Scale(InvThetaCb,ThetaSq*Mu(N),Z,ZZ);
         HoToU(Err,P,R,YY,Hpsi(N),Y);
         HoToU(Err,P,R,ZZ,Hpsi(N),Z);
      end loop;
      for N in 0..K loop
         Y:=-InvTheta*Y;
         Z:=-Mu(N)*Theta*Z;
      end loop;
   end K_Term;


   procedure ST(Switch: in Integer; ScalName,PsiName,MuName: in String; Err: in Radius; P: in ProjParam; R: in Weights; Q1,Q2,Z: in out Scalar) is
      --ScalName & X or & Y is the name of the file to/from which the components of the scaling transformation are written/read.
      --PsiName & K is the name of the file to/from which the generating functions are written/read.
      --MuName is the name of the file from which mu's are read
      use Scalar_Fou2_Ops;
      use Scalar_Fou1_Ops;
      KK:  String:="00";
      Hpsi: Hamiltons;
      HamY,HamYY,HamZ,HamZZ,HHpsi: Hamilton;
      Mu: Mu_Array;
      Q11,Q22,Q111,Y,ZZ: Scalar;
      F,FF: File_Type;
      Pw: Integer;
      TwoPi: Rep:=Rep(2.0)*Rep(3.14159265358979323846);
   begin
      Pw:=11;
      if Switch=0 then
         for K in 9..14 loop --change, if need a different compositon of scaling transfromations
            if (K mod 12)=0 then KK:="00"; end if;
            if (K mod 12)=1 then KK:="01"; end if;
            if (K mod 12)=2 then KK:="02"; end if;
            if (K mod 12)=3 then KK:="03"; end if;
            if (K mod 12)=4 then KK:="04"; end if;
            if (K mod 12)=5 then KK:="05"; end if;
            if (K mod 12)=6 then KK:="06"; end if;
            if (K mod 12)=6 then KK:="07"; end if;
            if (K mod 12)=8 then KK:="08"; end if;
            if (K mod 12)=9 then KK:="09"; end if;
            if (K mod 12)=10 then KK:="10"; end if;
            if (K mod 12)=11 then KK:="11"; end if;
            Read(PsiName &  KK,R,Hpsi(K mod 12));
         end loop;
         for K in 3..8 loop --change, if need a different compositon of scaling transfromations
            if (K mod 3=0) then
               PiTranslateX(Hpsi((K+6) mod 12),Hpsi(K));
            end if;
            if (K mod 3=1) then
               PiTranslateX(Hpsi((K+6) mod 12),HHpsi);
               PiTranslateY(HHpsi,Hpsi(K));
            end if;
            if (K mod 3=2) then
               PiTranslateY(Hpsi((K+6) mod 12),Hpsi(K));
            end if;
         end loop;
         Open(F,In_File,MuName);
         for K in 0..5 loop --change, if need a different compositon of scaling transfromations
            Get(F,Mu(K)); Mu(K+6):=Mu(K);
         end loop;
         Close(F);
         SetZero(R,HamY);
         SetZero(R,HamZ);
         New_Line;
         for K in 0..Pw loop
            Put("Step "); ShortPut(K);     New_Line;
            K_Term(K mod 12,Err,P,R,Hpsi,Mu,HamYY,HamZZ);
            HamY:=HamY+HamYY;
            HamZ:=HamZ+HamZZ;
         end loop;
         --Writing
         Create(FF,Out_File,ScalName & "Y");
         Write(FF,HamY);
         Close(FF);
         Create(FF,Out_File,ScalName & "Z");
         Write(FF,HamZ);
         Close(FF);
         return;
      else
         Read(ScalName & "Y",R,HamY);
         Read(ScalName & "Z",R,HamZ);
      end if;

      --Evaluating
      Q11:=Q1;
      Q22:=Q2;
      ZZ:=Z;
      Y:=NumVal(Q1,Q2,ZZ,HamY);
      Z:=NumVal(Q1,Q2,ZZ,HamZ);

      --Adding T_Mu_0 O...O T_Mu_11
      for N in 0..Pw loop
         Q111:=Q11;
         Q11:=Q22;
         Q22:=Q111+Q22;
         Q11:=Scal((Sup(Q11)/TwoPi-Rep'Floor(Sup(Q11)/TwoPi))*TwoPi);
         Q22:=Scal((Sup(Q22)/TwoPi-Rep'Floor(Sup(Q22)/TwoPi))*TwoPi);
         ZZ:=-Mu(N mod 12)*Theta*ZZ;
      end loop;

      --Modding
      Q1:=Y+Q11;
      Q2:=-InvTheta*Y+Q22;
      Z:=Z+ZZ;
      Q1:=Scal((Sup(Q1)/TwoPi-Rep'Floor(Sup(Q1)/TwoPi))*TwoPi);
      Q2:=Scal((Sup(Q2)/TwoPi-Rep'Floor(Sup(Q2)/TwoPi))*TwoPi);
   end ST;
end RG2_Ops;





































