--A sample program that implements RG with contraction in two unstable directions;
-- the program looks for eigenvectors at each renormalization step
with Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops, Fourier_Init,Messages;   --*
use Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops, Fourier_Init,Messages;    --*
with RG2_Ops;

procedure Rgx is

   pragma Suppress( Storage_Check);

   subtype Scalar is Numeric;
   Zero: constant Scalar := Scal(0);

   MaxDegQ1: constant Positive := 7;
   MaxDegQ2: constant Positive := 7;
   MaxDegQ:  constant Positive := 7;
   MaxDegP:  constant Positive := 7;
   MaxDegQP: constant Positive := 9;
   Np:       constant Positive := 128;

   package Run_RG2_Ops is
      new RG2_Ops (Size_1 => MaxDegQ1, Size_2 => MaxDegQ2, Q_Deg => MaxDegQ,
                   P_Deg  => MaxDegP,   QPdeg =>  MaxDegQP , Scalar => Scalar);

   use Run_RG2_Ops,Run_RG2_Ops.Scalar_Fou2_Ops,Run_RG2_Ops.Scalar_Taylor_Ops, Run_RG2_Ops.Scalar_Fou1_Ops,Run_RG2_Ops.Numeric_Fou2_Ops,Run_RG2_Ops.Numeric_Fou1_Ops;

   Sigma: constant Rep := 1.10;
   Tau:   constant Rep := 1.50;
   IK: constant Integer:=9;
   Pw: constant Integer:=6;
   Eren:  constant Radius := 1.0E-9;
   Accuracy: constant Radius := 1.0E-9;
   InvGolden: constant Scalar:=Golden-Scal(One);

   DirOut: constant String :="7_7_7_7_9/";
   DirInPsi: constant String :="7_7_7_7_9/";
   DirInH: constant String :="7_7_7_7_9/";
   DirInEv: constant String :="7_7_7_7_9/";
   SigTa: constant String:= "1.10_1.50/";
   SigTaIn: constant String:= "1.10_1.50/";
   HOname: constant String:=DirOut & SigTa & "H/h_x";
   HIname: constant String:=DirInH & SigTaIn & "H/h_x";
   EigenOname: constant String:=DirOut & SigTa & "Eigenvectors/"  &  "ev_x";
   EigenIname: constant String:=DirInEv & SigTaIn & "Eigenvectors/"  &  "ev_x";
   PsiIname: constant String:=DirInPsi & SigTaIn & "Psi/psi_x";
   PsiOname: constant String:=DirOut & SigTa & "Psi/psi_x";
   KK:  constant String:="_09";

   Kappa: Rep;
   PPi: ProjParam;
   Rfix,Rpri,Rpsi: Weights;
   F,FF: File_Type;
   H,Hr,Hrr,He1,He2: Hamilton;
begin
   MaxI:=500;
   MuMax:=Rep(InvGolden)*Rep(InvGolden);

   Rfix := (0.0, 0.85, 0.15);      -- initial and final R for H
   Rpri := (0.0, 0.85, 0.15);      -- R for HoU  i.e. R'
   Rpsi := (0.0, 0.85, 0.15);      -- R for psi

   SetZero(Rfix,H);
   SetZero(Rfix,Hr);
   SetZero(Rfix,Hrr);
   SetZero(Rfix,He1);
   SetZero(Rfix,He2);

   Put("INPUT H FILE:"); Put(HIname);
   New_Line;
   Put("OUTPUT H FILES:"); Put(HOname);
   New_Line;
   Put("INPUT PSI FILES:"); Put(PsiIname);Put(" and ");Put(PsiIname);
   New_Line;
   Put("OUTPUT PSI FILES:"); Put(PsiOname);Put(" and ");Put(PsiOname);
   New_Line;
   Put("# FIXED POINT (IN PERIOD 12):"); Put(Current_Output,IK);
   New_Line;
   Put("ERROR IN RENORMALIZATION:="); Put(Current_Output,Eren);
   New_Line;
   Put("Rfix:"); ShowRR(Rfix);
   Put("Rpri:"); ShowRR(Rpri);
   Put("Rpsi:"); ShowRR(Rpsi);

   Kappa := Sigma/Tau;

   New_Line;
   Put("PROJECTION PARAMETERS FOR H:");
   New_Line;
   ShowParam(Sigma,Kappa);

   New_Line;
   IminusDef(Sigma,Kappa,PPi);
   Put_Line("Iminus:");
   ShowProj(PPi);
   New_Line;

   Read(HIname & KK,Rfix,H);
   Read(EigenIname & "_1" & KK,Rfix,He1);
   Read(EigenIname & "_2" & KK,Rfix,He2);

   --addjusting the symmetries of the eigenvectors for additional safety
   if (IK mod 3)=0 then Nproj(H);  Nproj(He1);  Nproj(He2); end if;
   if (IK mod 3)=1 then Mproj(H);  Mproj(He1);  Mproj(He2); end if;
   if (IK mod 3)=2 then MNproj(H); MNproj(He1);  MNproj(He2); end if;

   SetZero(Rfix,Hrr);
   while Sup(Norm(Hrr-H))>Accuracy loop
      New_Line;
      Put("'MOST CRITICAL' HAMILTONIAN #" & KK);
      New_Line;
      Put("******************************************************************************************************");
      New_Line;
      Show(H);
      Put("******************************************************************************************************");
      New_Line;

      --EIGENVECTORS
      New_Line;
      Put("LOOKING FOR EGENVECTORS...");
      New_Line;
      FindEV(PsiIname,PsiOname,Pw,IK,Eren,PPi,Rfix,Rpsi,H,He1,He2,True,True); --comment if using accurate eigenvectors
      if (IK mod 3)=0 then Nproj(He1);  Nproj(He2); end if;
      if (IK mod 3)=1 then Mproj(He1);  Mproj(He2); end if;
      if (IK mod 3)=2 then MNproj(He1);  MNproj(He2); end if;
      if EX=1 then
         New_Line; Put("FINDEV DID NOT CONVERGE...");
         return;
      end if;

      --RENORMALIZATION
      New_Line;
      Put("RENORMALIZING...");
      New_Line;

      Copy(H,Hrr);
      RGX(PsiIname,PsiOname,Pw,IK,Eren,PPi,Rfix,Rpsi,He1,He2,H,Hr,True,True);
      New_Line;
      Put("||R(H)-H||:=");Put(Current_Output,Sup(Norm(Hr-Hrr)));

      --WRITING Hr
      Create(F,Out_File,HOname & KK);
      Write(F,Hr);
      Close(F);

      --WRITING He's
      Create(F,Out_File,EigenOname & "_1" & KK);
      Create(FF,Out_File,EigenOname & "_2" & KK);
      Write(F,He1);
      Write(FF,He2);
      Close(F);
      Close(FF);

      Copy(Hr,H);
   end loop;
end RGX;
























































