--A sample program that implements RG with contraction in two unstable directions;
with Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops,
  Fourier_Init,Messages,Ada.Characters.Latin_1;   --*
use Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops, Fourier_Init,Messages, Ada.Characters.Latin_1;    --*
with RG2_Ops;

procedure Rgy is

   pragma Suppress( Storage_Check);

   subtype Scalar is Numeric;
   Zero: constant Scalar := Scal(0);

   MaxDegQ1: constant Positive := 7;
   MaxDegQ2: constant Positive := 7;
   MaxDegQ:  constant Positive := 7;
   MaxDegP:  constant Positive := 7;
   MaxDegQP: constant Positive := 9;
   Np:       constant Positive := 128;

   package Run_RG2_Ops is
      new RG2_Ops (Size_1 => MaxDegQ1, Size_2 => MaxDegQ2, Q_Deg => MaxDegQ,
                   P_Deg  => MaxDegP,   QPdeg =>  MaxDegQP , Scalar => Scalar);

   use Run_RG2_Ops;
   package Run_Num2_Ops renames Numeric_Fou2_Ops;
   package Run_Num1_Ops  renames Numeric_Fou1_Ops;
   package Run_Scalar_Taylor_Ops renames Scalar_Taylor_Ops;
   package Run_Numeric_Taylor_Ops renames Numeric_Taylor_Ops;
   package Run_Scalar2_Ops renames  Scalar_Fou2_Ops;
   package Run_Scalar1_Ops renames  Scalar_Fou1_Ops;
   use  Run_Num2_Ops, Run_Num1_Ops, Run_Scalar_Taylor_Ops, Run_Numeric_Taylor_Ops, Run_Scalar2_Ops, Run_Scalar1_Ops;

   Sigma: constant Rep := 1.10;
   Tau:   constant Rep := 1.50;
   IK: constant Integer:=9;
   Pw: constant Integer:=6;
   Eren:  constant Radius := 1.0E-9;
   Accuracy: constant Radius := 1.0E-10;
   InvGolden: constant Scalar:=Golden-Scal(One);

   DirOut: constant String :="7_7_7_7_9/";
   DirInPsi: constant String :="7_7_7_7_9/";
   DirInH: constant String :="7_7_7_7_9/";
   DirInDRG: constant String :="7_7_7_7_9/";
   SigTa: constant String:= "1.10_1.50/";
   PsiOname: constant String:=DirOut & SigTa & "Psi/psi_x";
   PsiIname: constant String:=DirInPsi & SigTa & "Psi/psi_y";
   HOname: constant String:=DirOut & SigTa & "H/h_y";
   HIname: constant String:=DirInH & SigTa & "H/h_x";
   DRGIname: constant String:=DirInDRG & SigTa & "DRG/col";
   KK: constant String:= "_09";

   Kappa: Rep;
   PPi: ProjParam;
   Rfix,Rpri,Rpsi: Weights;
   F: File_Type;
   Hpsi: Polynom;
   H,Hr,Hrr: Hamilton;
   J,S: Integer;
   DR: Matrix(1..174,1..174);
   Iz,Jz: Pivot(1..174);
begin
   MaxI:=500;
   MuMax:=Rep(InvGolden)*Rep(InvGolden);

   Rfix := (0.0, 0.85, 0.15);      -- initial and final R for H
   Rpri := (0.0, 0.85, 0.15);      -- R for HoU  i.e. R'
   Rpsi := (0.0, 0.85, 0.15);      -- R for psi

   Put("INPUT H FILE:"); Put(HIname);
   New_Line;
   Put("OUTPUT H FILE:"); Put(HOname);
   New_Line;
   Put("INPUT PSI FILES:"); Put(PsiIname);Put(" and ");Put(PsiIname);
   New_Line;
   Put("OUTPUT PSI FILES:"); Put(PsiOname);Put(" and ");Put(PsiOname);
   New_Line;
   Put("# FIXED POINT (IN PERIOD 12):"); Put(Current_Output,IK);
   New_Line;
   Put("ERROR IN RENORMALIZATION:="); Put(Current_Output,Eren);
   New_Line;
   Put("Rfix:"); ShowRR(Rfix);
   Put("Rpri:"); ShowRR(Rpri);
   Put("Rpsi:"); ShowRR(Rpsi);

   Kappa := Sigma/Tau;

   New_Line;
   Put("PROJECTION PARAMETERS FOR H:");
   New_Line;
   ShowParam(Sigma,Kappa);

   New_Line;
   IminusDef(Sigma,Kappa,PPi);
   Put_Line("Iminus:");
   ShowProj(PPi);
   New_Line;

   Read(HIname & KK,Rfix,H);
   ReadDRG(Rfix,DirInDRG & SigTa & "DRG/col",DR,Iz,Jz,False);

   SetZero(Rfix,Hrr);
   New_Line;
   while Sup(Norm(Hrr-H))>Accuracy loop
      New_Line;
      Put("'MOST CRITICAL' HAMILTONIAN #" & KK);
      New_Line;
      Put("******************************************************************************************************");
      New_Line;
      Show(H);
      Put("******************************************************************************************************");
      New_Line;

      Copy(H,Hrr);
      RGY(PsiIname,PsiOname,Pw,IK,Eren,PPi,Rfix,Rpsi,H,DR,Iz,Jz,Hr,True,True);
      New_Line;
      Put("||R(H)-H||:=");Put(Current_Output,Sup(Norm(Hr-Hrr)));

      --WRITING Hr
      Create(F,Out_File,HOname & KK);
      Write(F,Hr);
      Close(F);

      Copy(Hr,H);
   end loop;
end Rgy;
























































