
with Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops, Fourier_Init,Messages;   --*
use Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops, Error_Ops, Numeric_Ops, Interval_Ops, Linear_Ops, Fourier_Init,Messages;    --*
with RG2_Ops;

procedure Scaling is

  pragma Suppress( Storage_Check);

  subtype Scalar is Numeric;
  Zero: constant Scalar := Scal(0);

  MaxDegQ1: constant Positive := 7;
  MaxDegQ2: constant Positive := 7;
  MaxDegQ:  constant Positive := 7;
  MaxDegP:  constant Positive := 7;
  MaxDegQP: constant Positive := 9;
  Np:       constant Positive := 120;

  package Run_RG2_Ops is
     new RG2_Ops (Size_1 => MaxDegQ1, Size_2 => MaxDegQ2, Q_Deg => MaxDegQ,
                  P_Deg  => MaxDegP,   QPdeg =>  MaxDegQP , Scalar => Scalar);

  use Run_RG2_Ops;

  package Run_Num2_Ops renames Numeric_Fou2_Ops;
  package Run_Num1_Ops  renames Numeric_Fou1_Ops;
  package Run_Scalar_Taylor_Ops renames Scalar_Taylor_Ops;
  package Run_Numeric_Taylor_Ops renames Numeric_Taylor_Ops;
  package Run_Scalar2_Ops renames  Scalar_Fou2_Ops;
  package Run_Scalar1_Ops renames  Scalar_Fou1_Ops;
  use  Run_Num2_Ops, Run_Num1_Ops, Run_Scalar_Taylor_Ops, Run_Numeric_Taylor_Ops, Run_Scalar2_Ops, Run_Scalar1_Ops;


  Sigma: constant Rep := 1.10;
  Tau:   constant Rep := 1.50;
  IK: constant Integer:=9;
  Pw: constant Integer:=6;
  Eren:  constant Radius := 1.0E-9;
  InvGolden: constant Scalar:=Golden-Scal(One);
  InvTheta: constant Scalar := Theta-Scal(One);

  SigTa: constant String:= "1.10_1.50/";
  SigTaIn: constant String:= "1.10_1.50/";
  DirOut: constant String :="7_7_7_7_9/";
  DirIn: constant String :="7_7_7_7_9/";
  DegName: constant String :="7_7_7_7_9";
  PsiIname: constant String:=DirIn & SigTa & "Psi/psi_x";
  ScalName: constant String:=DirOut & SigTa & "Scaling_Data/scaling_matrix";
  EigenName: constant String:=DirOut & SigTa & "Scaling_Data/eigenvalues";
  MuIname: constant String:=DirIn & SigTa & "Scaling_Data/mu";

  Mu: Mu_Array;
  Pi: Scalar:=Scal(3.14159265358979323846);
  PiOverTwo: Scalar:=Pi/Scal(2.0);
  ScalZero: Scalar:=Scalar(0.0);
  Mu12: Scalar;
  Kappa: Rep;
  PPi: ProjParam;
  Rds,Rfix,Rpri,Rpsi: Weights;
  Hpsi,HHpsi,HHHpsi: Polynom;
  F: File_Type;
  D,C,B,A: DS_Matrix;
  V,Bi,Br: Vector(1..4);
  ScalM,ScalMM: Matrix(1..4,1..4);
  Det: Rep;
  II: String:="0";
  KK:  String:="_09";
begin
   MaxI:=100;
   Kappa := Sigma/Tau;
   MuMax:=Rep(InvGolden)*Rep(InvGolden);

   Rfix := (0.0, 0.85, 0.15);      -- initial and final R for H
   Rpri := (0.0, 0.85, 0.15);      -- R for HoU  i.e. R'
   Rpsi := (0.0, 0.85, 0.15);      -- R for psi
   Rds:=(0.0,0.85,0.15);

   Put("INPUT PSI FILES:"); Put(PsiIname);
   New_Line;
   Put("INPUT MU FILE:"); Put(MuIname);
   New_Line;
   Put("ERROR IN RENORMALIZATION:="); Put(Current_Output,Eren);
   New_Line;
   Put("Rfix:"); ShowRR(Rfix);
   Put("Rpri:"); ShowRR(Rpri);
   Put("Rpsi:"); ShowRR(Rpsi);
   New_Line;

   Put("PROJECTION PARAMETERS:");
   New_Line;
   ShowParam(Sigma,Kappa);
   New_Line;
   New_Line;
   IminusDef(Sigma,Kappa,PPi);
   Put_Line("Iminus:");
   ShowProj(PPi);


   Mu12:=Scal(1.0);
   Open(F,In_File,MuIname);
   for K in IK..(IK+5) loop
      Get(F,Mu(K mod 12)); Mu((K+6) mod 12):=Mu(K mod 12);
      Mu12:=Mu(K mod 12)*Mu((K+6) mod 12)*Mu12;
   end loop;
   Close(F);

   B(1,1):=Scal(One);
   B(2,1):=ScalZero;
   B(3,1):=ScalZero;
   B(4,1):=ScalZero;

   B(1,2):=ScalZero;
   B(2,2):=Scal(One);
   B(3,2):=ScalZero;
   B(4,2):=ScalZero;

   B(1,3):=ScalZero;
   B(2,3):=ScalZero;
   B(3,3):=Scal(One);
   B(4,3):=ScalZero;

   B(1,4):=ScalZero;
   B(2,4):=ScalZero;
   B(3,4):=ScalZero;
   B(4,4):=Scal(One);


   D(1,1):=Scal(One);
   D(2,1):=ScalZero;
   D(3,1):=ScalZero;
   D(4,1):=ScalZero;

   D(1,2):=ScalZero;
   D(2,2):=Scal(One);
   D(3,2):=ScalZero;
   D(4,2):=ScalZero;

   D(1,3):=ScalZero;
   D(2,3):=ScalZero;
   D(3,3):=Scal(One);
   D(4,3):=ScalZero;

   D(1,4):=ScalZero;
   D(2,4):=ScalZero;
   D(3,4):=ScalZero;
   D(4,4):=Scal(One);

   New_Line;
   Put("COMPUTING THE SCALING MATRIX...");
   for K in IK..(IK+5) loop
      if (K mod 12 =0) then KK:="_00"; end if;
      if (K mod 12 =1) then KK:="_01"; end if;
      if (K mod 12 =2) then KK:="_02"; end if;
      if (K mod 12 =3) then KK:="_03"; end if;
      if (K mod 12 =4) then KK:="_04"; end if;
      if (K mod 12 =5) then KK:="_05"; end if;
      if (K mod 12 =6) then KK:="_06"; end if;
      if (K mod 12 =7) then KK:="_07"; end if;
      if (K mod 12 =8) then KK:="_08"; end if;
      if (K mod 12 =9) then KK:="_09"; end if;
      if (K mod 12 =10) then KK:="_10"; end if;
      if (K mod 12 =11) then KK:="_11"; end if;
      New_Line;
      Put("STEP" & KK);

      SetZero(Rpsi,Hpsi);
      Read(PsiIname & KK,Rpsi,Hpsi);
      New_Line;
      Put("Mu=");Put(Current_Output,Mu(K mod 12));
      New_Line;
      Put("Generating Function #");Put(Current_Output,K mod 12);
      New_Line;
      Show(Hpsi);
      New_Line;

      DS(Eren,PPi,Rds,Hpsi,ScalZero,ScalZero,ScalZero,A);

      for I in 1..4 loop
         for J in 1..4 loop
            ScalM(I,J):=Rep(A(I,J));
         end loop;
      end loop;
      New_Line;
      Put("Det of A:="); Put(Current_Output,PlusMinusDet(ScalM));

      if (K mod 3=0) then
         PiTranslateX(Hpsi,HHpsi);
      end if;

      if (K mod 3=1) then
         PiTranslateX(Hpsi,HHHpsi);
         PiTranslateY(HHHpsi,HHpsi);
      end if;

      if (K mod 3=2) then
         PiTranslateY(Hpsi,HHpsi);
      end if;

      DS(Eren,PPi,Rds,HHpsi,ScalZero,ScalZero,ScalZero,C);
      for I in 1..4 loop
         for J in 1..4 loop
            ScalM(I,J):=Rep(C(I,J));
         end loop;
      end loop;
      New_Line;
      Put("Det of C:="); Put(Current_Output,PlusMinusDet(ScalM));

      A(1,1):=Theta*A(1,1);
      A(2,1):=Theta*A(2,1);
      A(3,1):=Theta*A(3,1);
      A(4,1):=Theta*A(4,1);

      A(1,2):=-InvTheta*A(1,2);
      A(2,2):=-InvTheta*A(2,2);
      A(3,2):=-InvTheta*A(3,2);
      A(4,2):=-InvTheta*A(4,2);

      A(1,3):=Mu(K mod 12 )*InvTheta*A(1,3);
      A(2,3):=Mu(K mod 12)*InvTheta*A(2,3);
      A(3,3):=Mu(K mod 12)*InvTheta*A(3,3);
      A(4,3):=Mu(K mod 12)*InvTheta*A(4,3);

      A(1,4):=-Mu(K mod 12)*Theta*A(1,4);
      A(2,4):=-Mu(K mod 12)*Theta*A(2,4);
      A(3,4):=-Mu(K mod 12)*Theta*A(3,4);
      A(4,4):=-Mu(K mod 12)*Theta*A(4,4);

      C(1,1):=Theta*C(1,1);
      C(2,1):=Theta*C(2,1);
      C(3,1):=Theta*C(3,1);
      C(4,1):=Theta*C(4,1);

      C(1,2):=-InvTheta*C(1,2);
      C(2,2):=-InvTheta*C(2,2);
      C(3,2):=-InvTheta*C(3,2);
      C(4,2):=-InvTheta*C(4,2);

      C(1,3):=Mu(K mod 12)*InvTheta*C(1,3);
      C(2,3):=Mu(K mod 12)*InvTheta*C(2,3);
      C(3,3):=Mu(K mod 12)*InvTheta*C(3,3);
      C(4,3):=Mu(K mod 12)*InvTheta*C(4,3);

      C(1,4):=-Mu(K mod 12)*Theta*C(1,4);
      C(2,4):=-Mu(K mod 12)*Theta*C(2,4);
      C(3,4):=-Mu(K mod 12)*Theta*C(3,4);
      C(4,4):=-Mu(K mod 12)*Theta*C(4,4);

      Mult(B,A);
      Mult(D,C);
   end loop;
   Mult(B,D);

   Create(F,Out_File,ScalName);
   for I in 1..4 loop
      for J in 1..4 loop
         Put(F,Center(B(I,J)));Put(" ");
         ScalM(I,J):=Rep(B(I,J));
      end loop;
      New_Line(F);
   end loop;
   Close(F);

   Br(1):=ScalM(1,1); Bi(1):=Rep(0.0);
   Br(3):=ScalM(3,3); Bi(3):=Rep(0.0);
   Det:=(ScalM(4,4)-ScalM(2,2))*(ScalM(4,4)-ScalM(2,2))+Rep(4.0)*ScalM(4,2)*ScalM(2,4);

   if Det<Rep(0.0) then
      New_Line;
      Put("COMPLEX EIGENVALUES");
      Bi(2):=Half*Sqrt(-Det);
      Bi(4):=Half*(-Sqrt(-Det));
      Br(2):=Half*(ScalM(4,4)+ScalM(2,2));
      Br(4):=Half*(ScalM(4,4)+ScalM(2,2));
   else
      Bi(2):=Rep(0.0);
      Bi(4):=Rep(0.0);
      Br(2):=Half*(ScalM(4,4)+ScalM(2,2)+Sqrt(Det));
      Br(4):=Half*(ScalM(4,4)+ScalM(2,2)-Sqrt(Det));
   end if;
   New_Line;
   New_Line;
   Put("THE EIGENVALUES OF THE SCALING MATRIX IN COORDINATES (X,Y,U,Z) ARE:");
   New_Line;

   Create(F,Out_File,EigenName);
   for I in 1..4 loop
      case I is
         when    1 =>     II:="1";
         when    2 =>     II:="2";
         when    3 =>     II:="3";
         when    4 =>     II:="4";
         when others =>  Message("Scaling",Parameter_Error);
      end case;
      Put("Eigenvalue " & II & ":=");Put(Current_Output,Br(I));Put("+I");Put(Current_Output,Bi(I));
      New_Line;
      Put(F,Br(I)); Put(F,Bi(I));
      New_Line(F);
   end loop;
   Close(F);

end Scaling;



















































