with Ada.Text_IO, Rep_Ops, Error_Ops, Taylor_Init;
use Ada.Text_IO, Rep_Ops, Error_Ops, Taylor_Init;

generic

   Size: in Positive;
   type Scalar is private;

   with function IsNumeric(S: Scalar) return Boolean is <>;
   with function IsInterval(S: Scalar) return Boolean is <>;
   with procedure Get(F: in File_Type; S: out Scalar) is <>;
   with procedure Put(F: in File_Type; S: in Scalar) is <>;
   with procedure Show(S: in Scalar) is <>;
   with function Scal(K: Integer) return Scalar is <>;
   with function Scal(R: Rep) return Scalar is <>;
   with function Scal(E: Error) return Scalar is <>;
   with function Center0(S: Scalar) return Boolean is <>;
   with function Err0(S: Scalar) return Boolean is <>;
   with function Center(S: Scalar) return Rep is <>;
   with function ErrPart(S: Scalar) return Error is <>;
   with procedure ResetCenter(S: in out Scalar) is <>;
   with function ResetCenter(S: Scalar) return Scalar is <>;
   with procedure Split(S: in Scalar; Sr,Se: out Scalar) is <>;
   with function Inf(S: Scalar) return Rep is <>;
   with function Sup(S: Scalar) return Rep is <>;
   with function SupAbs(S: Scalar) return Rep is <>;
   with function Err(S: Scalar) return Error is <>;
   with procedure ResetErr(S: in out Scalar) is <>;
   with function ResetErr(S: Scalar) return Scalar is <>;
   with procedure ErrMult(Factor: in Error; S: in out Scalar) is <>;
   with function "abs"(S: Scalar) return Scalar is <>;
   with function "-"(S: Scalar) return Scalar is <>;
   with function "+"(Left,Right: Scalar) return Scalar is <>;
   with function "-"(Left,Right: Scalar) return Scalar is <>;
   with function "*"(S: Scalar; E: Error) return Error is <>;
   with function "*"(E: Error; S: Scalar) return Error is <>;
   with function "*"(Left,Right: Scalar) return Scalar is <>;
   with function Inv(S: Scalar) return Scalar is <>;
   with function "/"(E: Error; S: Scalar) return Error is <>;
   with function "/"(Left,Right: Scalar) return Scalar is <>;
   with function "**"(X1: Scalar; I: Integer) return Scalar is <>;
   with function Sqrt(S: Scalar) return Scalar is <>;
   with function ExpAbsLog(S: Scalar) return Scalar is <>;
   with function Cosh(S: Scalar) return Scalar is <>;
   with function Sinh(S: Scalar) return Scalar is <>;
   with procedure CoshSinh(S: in Scalar; Sc,Ss: out Scalar) is <>;
   with function Exp(S: Scalar) return Scalar is <>;
   with function Log(S: Scalar) return Scalar is <>;
   with function Golden return Scalar is <>;

package Taylor_Ops is

   subtype Degree is Integer range -1 .. Size;
   type Taylor is private;

   function Deg(P: Taylor) return Degree;                              -- Deg := degree of polynomial part of P
   function EffDeg(P: Taylor) return Degree;                           -- Deg := effective degree ...
   function Component(K: Natural; P: Taylor) return Scalar;            -- Component := P[K]
   procedure SetComponent(D: in Degree; S: in Scalar;
                                        P: in out Taylor);             -- P[D] := S
   procedure AddComponent(D: in Degree; S: in Scalar;
                                        P: in out Taylor);             -- P[D] := P[D] + S
   procedure SetZero(P: out Taylor);                                   -- P := 0
   procedure Get(F: in File_Type; P: out Taylor);                      -- Read P from F               center only
   procedure Put(F: in File_Type; P: in Taylor);                       -- Write P to F                center only
   procedure Read(Name: in String; P: out Taylor);                     -- Read P from file Name       center only
   procedure Write(Name: in String; P: in Taylor);                     -- Write P to file Name        center only
   procedure Show(P: in Taylor);                                       -- Write P to Current_Output
   procedure Cshow(F: in File_Type; P: in Taylor);                     -- Write P as a C-output
   function Coeff(R: Radius; K: Natural; P: Taylor) return Scalar;     -- Coeff := P_K
   function Val(R: Radius; Z: Scalar; P: Taylor) return Scalar;        -- Val := P(Z)
   procedure MaxAbsComponent(P: Taylor; R: in out Rep);                -- R := Max of R and Max_K |P[K]|
   function Norm(R: Radius; P: Taylor) return Error;                   -- Norm := ||P||_R
   function SupNorm(R: Radius; P: Taylor) return Error;                -- Norm := sup_K {||P(K)||_R}
   procedure Copy(P1: in Taylor; P2: out Taylor);                      -- P2 := P1
   procedure OddEliminate(P1: in Taylor; P2: out Taylor);              -- Projection on even K's
   procedure EvenEliminate(P1: in Taylor; P2: out Taylor);             -- Projection on odd K's
   procedure CopyAbs(P1: in Taylor; P2: out Taylor);                   -- P2 := Abs(P1)
   procedure CopySign(P1: in Taylor; P2: in out Taylor);               -- P2 := Sign(P1)*abs(P2)
   procedure Neg(P1: in Taylor; P2: out Taylor);                       -- P2 := -P1
   function "-"(P: Taylor) return Taylor;                              -- Return -P
   procedure Mult(S: in Scalar; P: in out Taylor);                     -- P := S*P
   procedure Prod(S: in Scalar; P1: in Taylor; P2: out Taylor);        -- P2 := S*P1
   function "*"(S: Scalar; P: Taylor) return Taylor;                   -- Return S*P
   procedure Add(P1: in Taylor; P2: in out Taylor);                    -- P2 := P2+P1
   procedure Sum(P1,P2: in Taylor; P3: out Taylor);                    -- P3 := P1+P2
   function "+"(P1,P2: Taylor) return Taylor;                          -- Return P1+P2
   procedure Sub(P1: in Taylor; P2: in out Taylor);                    -- P2 := P2-P1
   procedure Diff(P1,P2: in Taylor; P3: out Taylor);                   -- P3 := P1-P2
   function "-"(P1,P2: Taylor) return Taylor;                          -- Return P1-P2
   procedure MultAdd(S: in Scalar; P1: in Taylor; P2: in out Taylor);  -- P2 := P2+S*P1
   procedure Scale(Z,S: in Scalar; P1: in Taylor; P2: out Taylor);     -- P2 := Z*P1(S.)
   function Scale(Z,S: Scalar; P: Taylor) return Taylor;               -- Scale := Z*P(S.)
   procedure ScalePlus(L: in Cutoff; Z,S: in Scalar; R1,R2: in Radius;
                                                     P1: in Taylor;
                                                     P2: out Taylor);  -- P2 := Z*P1(S.) assuming degree > L
   procedure Der(R1,R2: in Radius; P1: in Taylor; P2: out Taylor);     -- P2 := P1'
   function Der(R1,R2: Radius; P: Taylor) return Taylor;               -- Der := P'
   procedure DerDer(R1,R2: in Radius; P1: in Taylor; P2: out Taylor);  -- P2 := P1''
   function DerDer(R1,R2: Radius; P: Taylor) return Taylor;            -- DerDer := P''
   procedure DerMultPart(S: in Scalar; P: in out Taylor);              -- Part of P -> S*P'
   procedure DerProdPart(S: in Scalar; P1: in Taylor; P2: out Taylor); -- Part of P2 := S*P1'
   function Center0(P: Taylor) return Boolean;                         -- True iff polynomial part is zero
   procedure ResetCenter(P: in out Taylor; Kmin: in Natural := 0;
                                           Kmax: in Cutoff := Size);   -- Reset Center of P.C(Kmin .. Kmax)
   procedure ResetCenter(P: in Taylor; Pr: out Taylor;
                                       Kmin: in Natural := 0;
                                       Kmax: in Cutoff := Size);       -- Pr := P with centers Kmin .. Kmax reset
   function ResetCenter(P: Taylor; Kmin: Natural := 0;
                                   Kmax: Cutoff := Size) return Taylor;-- Return P with centers Kmin .. Kmax reset
   function Err0(P: Taylor) return Boolean;                            -- True iff error part is zero
   procedure ResetErr(P: in out Taylor; Kmin: in Natural := 0;
                                        Kmax: in Cutoff := Size);      -- Reset Error of P.C(Kmin .. Kmax)
   procedure ResetErr(P: in Taylor; Pr: out Taylor;
                                    Kmin: in Natural := 0;
                                    Kmax: in Cutoff := Size);          -- Pr := P with errors Kmin .. Kmax reset
   function ResetErr(P: Taylor; Kmin: in Natural := 0;
                                Kmax: in Cutoff := Size) return Taylor;-- Return P with errors Kmin .. Kmax reset
   procedure Split(P: in Taylor; Pc,Pe: out Taylor);                   -- Split P into polynomial and error
   procedure ErrConvert(P: in Taylor; Pe: out Taylor);                 -- Enlarge P -> Pe centered at 0
   function ErrConvert(P: Taylor) return Taylor;                       -- Enlarge P -> ErrConvert centered at 0
   procedure DegConvert(R: in Radius; D: in Degree; P: in out Taylor); -- Enlarge P s.t. P.D <= D
   procedure DegConvert(R: in Radius; D: in Degree; P: in Taylor;
                                                    Pc: out Taylor);   -- Enlarge P -> Pc s.t. Pc.D <= D
   function DegConvert(R: Radius; D: Degree; P: Taylor) return Taylor; -- Enlarge P -> DegConvert s.t. Deg <= D
   procedure XMultAdd(R: in Radius; K: in Degree; S: in Scalar;
                                                  P1: in Taylor;
                                                  P2: in out Taylor;
                                                  D: in Degree:= Size);-- P2(Z) := P2(Z)+S*(Z^K)*P1(Z) all Z
   procedure Mult(R: in Radius; P1: in Taylor; P2: in out Taylor;
                                               D: in Degree := Size);  -- P2 := P2*P1
   procedure Prod(R: in Radius; P1,P2: in Taylor; P3: out Taylor;
                                                  D: in Degree:= Size);-- P3 := P1*P2
   function Prod(R: in Radius; P1,P2: Taylor;
                               D: Degree := Size) return Taylor;       -- Prod := P1*P2
   procedure MultAdd(R: in Radius; P1,P2: in Taylor;
                                   P3: in out Taylor;
                                   D: in Degree := Size);              -- P3 := P3+P1*P2
   procedure MultSub(R: in Radius; P1,P2: in Taylor;
                                   P3: in out Taylor;
                                   D: in Degree := Size);              -- P3 := P3-P1*P2

   -- the following work only with Numeric=True
   procedure Quot(P1,P2: in Taylor; P3: out Taylor;
                                    D: in Degree := Size);             -- P3 := P1/P2
   function Quot(P1,P2: Taylor; D: Degree := Size) return Taylor;      -- Quot := P1/P2
   function "/"(P1,P2: Taylor) return Taylor;                          -- Return P1/P2
   procedure Comp(P1,P2: in Taylor; P3: out Taylor);                   -- P3 := P1 o P2
   function Comp(P1,P2: Taylor) return Taylor;                         -- Comp := P1 o P2
   procedure InvComp(R,Err: in Radius; P1,P2: in Taylor;
                                       P3: in out Taylor);             -- P3 solves P2(P3)=P1
   function InvComp(R,Err: Radius; P1,P2: Taylor) return Taylor;       -- InvComp solves P2(InvComp)=P1
   function DerVal(R: Radius; Z: Scalar; P: Taylor) return Scalar;     -- Derval := P'(Z)
   procedure MinDeg(P1: in Taylor; P2: in out Taylor);                 -- P2.D := Min(P1.D,P2.D)
   procedure CutDeg(R: in Radius; Cut: in Rep; P: in out Taylor);      -- Cut h.o. terms with Norm<Cut
   function NumIprod(R: Radius; P1,P2: Taylor) return Scalar;          -- NumIprod := <P1,P2>_R
   function NumIprod(P1,P2: Taylor) return Scalar;                     -- NumIprod := <P1,P2>
   procedure FindRoot(Err: in Radius; P: in Taylor; S: in Scalar;
                                                    R: in out Scalar); -- R solves P(R)=S

   -- useful for testing
   procedure NextRandom(I: in out Long_Integer);
   function Exp1 return Taylor;                                        -- Exp1(Z) := Exp(Z)-1  for all Z
   function Ln1 return Taylor;                                         -- Ln1(Z) := Ln(1+Z)  for all Z
   procedure Random(D: in Degree; Seed: in out Long_Integer;
                                  R: in Radius; P: out Taylor);        -- P := (random polynomial)
   procedure RandomOdd(D: in Degree; Seed: in out Long_Integer;
                                     R: in Radius; P: out Taylor);     -- P := (random polynomial)
   procedure RandomEven(D: in Degree; Seed: in out Long_Integer;
                                      R: in Radius; P: out Taylor);    -- P := (random polynomial)
private

   type Scalar_Array is array (Integer range 0 .. Size) of Scalar;

   type Taylor is
      record
         D: Degree;
         C: Scalar_Array;
      end record;

   procedure Fmax1(R: in Scalar; M: out Natural; F: out Error);         -- used by Der
   procedure Fmax2(R: in Scalar; M: out Natural; F: out Error);         -- used by DerDer
   function ErrCoeff(R: Radius; K: Natural; P: Taylor) return Error;    -- ErrCoeff := Err(P_K)

end Taylor_Ops;

